## Filename: UCS.pm
## Modified: Thu Aug 25 02:25:30 2005 (severt)   
##   Author: Stefan Evert
##  Purpose: UCS core library (AM registry, variable names, wildcard patterns, error messages)

package UCS;

use sigtrap qw<die normal-signals>;		# so interrupts during system commands work and clean-up functions are called

use FileHandle;
use Carp qw<cluck>;

# make sure that configuration information is immediately available
# (because of certain problems with circular dependencies between modules)
BEGIN {				
# central version information
$Version = "0.5";
$Copyright = "Copyright (C) 2004-2005 by Stefan Evert";

# UCS and UCS/Perl base directories
$BaseDir = '/usr/src/RPM/BUILD/perl-UCS-0.5/System';	 # in-place edit by Install.perl
$PerlDir = "$BaseDir/perl";

# UCS/Perl configuration information
$PerlCmd = '#!/usr/bin/perl -w -I/usr/src/RPM/BUILD/perl-UCS-0.5/System/perl/lib';	 # in-place edit by Install.perl
$PerlBin = '/usr/bin/perl';	 # in-place edit by Install.perl
@PerlLibs = ('/usr/src/RPM/BUILD/perl-UCS-0.5/System/perl/lib');	 # in-place edit by Install.perl

# full path to R binary (for UCS::R module) 
$Rbin = '/usr/bin/R';	 # in-place edit by Install.perl

# UCS base directory reported to R interpreter (uses Windows path under Cygwin)
$R_BaseDir = '/usr/src/RPM/BUILD/perl-UCS-0.5/System';	 # in-place edit by Install.perl

# version number of installed R interpreter (for certain workarounds)
$R_Version = 2.72;	 # in-place edit by Install.perl

# whether RSPerl interface is available (and has been tested to work)
$RSPerl = 0;	 # in-place edit by Install.perl

# full path to a2ps program (for "-ps" mode in ucs-print program)
$a2ps = '/usr/bin/a2ps';	 # in-place edit by Install.perl

# names of core variables and derived variables
@CoreVars = qw<id l1 l2 f f1 f2 N>;
@DerivedVars = (
		qw<O11 O12 O21 O22 E11 E12 E21 E22 R1 R2 C1 C2 lf lf1 lf2 e b o le lb lo>,
		qw<p p1 p2 lp lp1 lp2>,		# for backward compatibility (and convenience :o)
		);
}

##
##  general utility functions
##

# warnings are suppressed in non-verbose mode
$Verbose = 1;

# really die, with multi-line error message, even if in Tk event loop (may do some clean-up)
sub Die ( @ ) {
  print STDERR "UCS ERROR:\n";
  foreach (@_) {
    print STDERR "$_\n";
  }
  cluck "Stack trace:\n";
  exit 1;
}

# print warning on STDERR (by default, may later be caught by UCS::Tk or recorded in log file)
sub Warn ( @ ) {
  if ($Verbose) {
    print STDERR "UCS WARNING/ERROR:\n";
    foreach (@_) {
      print STDERR "$_\n";
    }
  }
}

# status messages will only be displayed by Tk interfaces
sub Status ( $ ) {
  my $msg = shift;
  # do nothing with status messages (unless overridden by Tk modules)
}
STDOUT->autoflush(1);				# so incomplete messages will be displayed

# cool-looking splash screen ;-) (might be overriden with graphic display by Perl/Tk interfaces)
sub Splash ( ) {
  print "\n";
  print "  UU   UUU  CCCCCCC SSSSSSSSS        		 \n";
  print "  UU   UUU CCC     SSS               		 \n";
  print "  UU   UUU CC       SSSS   UCS Version $Version \n";
  print "  UU   UUU CC         SSS                       \n";
  print "  UU  UUUU CCC        SSS    $Copyright \n";
  print "   UUUUUU   CCCCCC SSSSS                        \n";
  print "\n";
}

# removed duplicate values from list but retain original order
sub Unique ( @ ) {
  my %seen = ();
  return grep { not $seen{$_}++ } @_;
} 


##
##  manipulating variable names
##

# @matches = UCS::Match($pattern, @names);
# match variable names using wildcards ? (single char except .), * (anything except .), and % (anything)
# (pattern must consist of characters A-Z a-z 0-9 . and wildcards ? * % only)
sub Match ( $@ ) {
  my $pattern = shift;
  if (not $pattern =~ /^[A-Za-z0-9\.\?\*\%]+$/) {
    Die "UCS::Match(): invalid characters in pattern '$pattern'.";
  }
  $pattern =~ s/\./\\./g;			# translate pattern into perl regular expression
  $pattern =~ s/\?/[^.]/g;
  $pattern =~ s/\*/[^.]*/g;
  $pattern =~ s/\%/.*/g;
  my $rx = qr/^$pattern$/;			# compile regular expression
  return grep {/$rx/} @_;
}

# $ok = UCS::ValidKey($key);
# check if $key is a valid identifier (e.g. for an AM or a user-defined variable "b.$key")
# (a UCS identifier may only contain the characters A-Z, a-z, 0-9, and .)
sub ValidKey ( $ ) {
  my $key = shift;
  return ($key =~ /^[A-Za-z0-9]+(\.[A-Za-z0-9]+)*$/) ? 1 : 0;
}

# $ok = UCS::ValidName($name);
# check if variable name $name is valid, i.e. either a standard (core or derived) variable,
# an AM value or rank, or a user-defined variables
sub ValidName ( $ ) {
  my $name = shift;
  return 1 if					# user-defined variable
    $name =~ /^[bnfx]\./ and ValidKey($');
  return 1 if					# AM values or ranks
    $name =~ /^(r|am)\./ and ValidKey($');
  return 1 if					# standard (core and derived) variables
    $name =~ /^(id | l[12] | l?f[12]? | N | [RC][12] | [OE][12]{2} | l?[ebo] | l?p[12]?)$/x;
  return 0;					# invalid
}

# $type = UCS::VarType($name);                  # data types: BOOL, INT, DOUBLE, STRING
# determine data type of variable from its name (returns "" if name is invalid)
sub VarType ( $ ) {
  my $name = shift;
  return "" 
    unless ValidName($name);
  return "BOOL"
    if $name =~ /^b\./;
  return "STRING"
    if $name =~ /^f\./ or $name =~ /^l[12]$/;
  return "INT"
    if $name =~ /^[nr]\./ or $name =~ /^(id | f[12]? | N | [RC][12] | O[12]{2})$/x;
  return "DOUBLE";   # everything else =  lf lf1 lf2 e b o le lb lo E?? x.% am.% p p1 p2 lp lp1 lp2
}

# ($spec, $key) = UCS::SplitName($name);
# split variable name into data type specifier $spec (am, r, b, f, n, x) and identifier $key
# (AM key or name of user-defined variable);  returns (undef, $name) if split fails
sub SplitName ( $ ) {
  my $name = shift;
  return (undef, $name)
    unless ValidName($name)
      and $name =~ /^(am|[rbfnx])\.(.+)$/;
  return ($1, $2);
}


##
##  association measure registry
##

use UCS::Expression;				# can't load UCS::Expression earlier because of circular dependencies

%AM = ();     # $AM{$key} = { NAME => $name,  EXP => $expression,  DESCRIPTION => $opt_description }

sub AM_Keys ( ) {
  return sort keys %AM;
}

sub AM ( $ ) {
  my $key = shift;
  return exists $AM{$key};
}

sub AM_Name ( $ ) {
  my $key = shift;
  UCS::Die "UCS::AM_Name(): association measure '$key' not registered"
      unless AM($key);
  return $AM{$key}{NAME};
}

sub AM_Description ( $ ) {
  my $key = shift;
  UCS::Die "UCS::AM_Description(): association measure '$key' not registered"
      unless AM($key);
  return $AM{$key}{DESCRIPTION};
}

sub AM_Expression ( $ ) {
  my $key = shift;
  UCS::Die "UCS::AM_Expression(): association measure '$key' not registered"
      unless AM($key);
  return $AM{$key}{EXP};
}

sub Eval_AM ( $$ ) {
  my $key = shift;
  my $arghash = shift;
  Die "UCS::Eval_AM(): association measure '$key' not registered"
    unless exists $AM{$key};
  return $AM{$key}{EXP}->eval($arghash);
}

sub Load_AM_Package {
  my @package_names = map { m{/([^/]+)\.pm$} and $1 } glob "$PerlDir/lib/UCS/AM/*.pm";
  my %requested = ();				# collect names of requested packages
  foreach my $name (@_) {
    if ($name eq "") {
      $requested{"UCS::AM"}++;
    }
    elsif ($name =~ /^all$/i) {
      return Load_AM_Package(@package_names, "");
    }
    else {
      my @matches = grep { /^$name/i } @package_names;
      if (@matches == 1) {
	my $p = shift @matches;
	$requested{"UCS::AM::$p"}++;
      }
      else {
	Warn "UCS::Load_AM_Package():  no match for requested AM package '$name'.",
	  "[available packages: @package_names]";
      }
    }
  }
  my @modules = sort keys %requested;
  foreach my $mod (@modules) {
    eval "use $mod; 1"
      or Die "UCS:  could not load AM package $mod\n";
  }
  return @modules;
}

sub Register_AM ( $$$;$ ) {
  my $key = shift;
  my $name = shift;
  my $code = shift;
  my $desc = (@_) ? shift : "";

  Die "UCS::Register_AM(): invalid key '$key' for association measure"
    unless ValidKey($key);
  $name =~ s/\s+/ /g;				# normalise whitespace in $name
  $name =~ s/^\s+//;
  $name =~ s/\s+$//;
  # don't normalise full description or UCS expression
  
  my $exp = undef;
  if (ref $code eq "UCS::Expression") {
    $exp = $code->copy;
  }
  else {
    $exp = new UCS::Expression $code;
  }
  return 0 unless defined $exp;			# UCS::Expression constructor should have issued warning(s)

  $AM{$key} = {					# add new measure to registry
	       NAME => $name,
	       EXP => $exp,
	       DESCRIPTION => $desc,
	      };

  return 1;
}



1;


__END__


=head1 NAME

UCS - Core library


=head1 SYNOPSIS

  use UCS;

  $UCS::Version;				# UCS version
  $UCS::Copyright;				# UCS copyright string
  $UCS::BaseDir;				# base directory of UCS system
  $UCS::PerlDir;				# base directory of UCS/Perl
  
  UCS::Die("Msg line 1", "Msg line 2", ...);	# really die (even in Tk loop)
  UCS::Warn("Msg line 1", "Msg line 2", ...);	# warning message (may be caught by Tk)
  UCS::Status("Message");			# display status message in Tk window
  UCS::Splash();				# splash screen (may be shown during start-up)
  $UCS::Verbose = 0;				# suppress warnings 
  @unique_values = UCS::Unique(@list);		# remove duplicates from list

  @vars = (@UCS::CoreVars, @UCS::DerivedVars);  # standard variable names (core and derived)
  @matches = UCS::Match($pattern, @names);	# match variable names
  $ok = UCS::ValidKey($key);			# valid identifier, e.g as AM key
  $ok = UCS::ValidName($name);			# whether variable name is valid
  $type = UCS::VarType($name);			# "BOOL", "INT", "DOUBLE", "STRING"
  ($spec, $key) = UCS::SplitName($name);        # split am.*, r.*, or user-defined variable name

  @registered_AMs = UCS::AM_Keys();		# keys for built-in AMs (when loaded)
  if (UCS::AM($key)) {
    $full_name = UCS::AM_Name($key);		# long descriptive name
    $description = UCS::AM_Description($key);	# optional multi-line text
    $exp = UCS::AM_Expression($key);		# AM equation as compiled UCS expression
    $score = $exp->eval({f=>$f, f1=>$f1, ...});	# use UCS::Expression methods to evaluate AM
  }
  $score = UCS::Eval_AM($key, $arghash);        # convenient but slow

  UCS::Load_AM_Package("HTest", ...);		# load built-in AM packages

  $ok = UCS::Register_AM			# register new association measure
    "tscore",					# AM key (-> variables am.tscore and r.tscore)
    "t-score measure (Church et. al. 1991)",	# long descriptive name
    '(%O11% - %E11%) / sqrt(%O11%)',	        # UCS expression (will be compiled into UCS::Expression)
    $multiline_text;				# optional multi-line description of AM

=head1 DESCRIPTION

This UCS core library maintains a list of B<bulit-in AM>s and Perl subroutines
for computing their B<score>s from a candidate's signatures. Utility functions
perform syntax checks for B<field name>s, determine B<field type>s from the naming
conventions, and B<match pattern>s containing UCS wildcards against field names.

=head1 CONFIGURATION VARIABLES

=over 4

=item $UCS::Version;

The currently installed UCS version.

=item $UCS::Copyright;

A copyright string for the UCS system.  Will be displayed by some
UCS/Perl scripts.

=item $UCS::BaseDir;

The base directory of the UCS System installation. Compiled UCS
B<programs> and links to Perl scripts are installed in
I<$UCS::BaseDir>/bin/, while the components of B<UCS/R> can be found
in I<$UCS::BaseDir>/R/.

=item $UCS::PerlDir;

The base directory of the B<UCS/Perl> installation. The UCS Perl
modules are installed in I<$UCS::PerlDir>/lib/ and its subdirectories,
Perl scripts in I<$UCS::PerlDir>/bin/.

=back


=head1 GENERAL FUNCTIONS

=over 4

=item UCS::Die($message, ...);

"Safe" replacement for Perl's built-in B<die> function, which will
even exit properly from a Perl/Tk loop.  One or more lines of error
messages are printed on STDERR (or shown in some other suitable
manner).

=item UCS::Warn($message, ...);

By default, prints one or more lines of warning/error messages on
STDERR like B<UCS::Die>, but does not exit the script.  The purpose of
this replacement for the built-in B<warn> function is to allow
warnings to be caught and displayed in a Perl/Tk user interface.
Warnings might also be redirected to a log file.

=item UCS::Status($message);

Displays a status message in a Perl/Tk interface.  By default,
I<$message> is appended to any previous messages.  When I<$message>
ends in a newline character (C<\n>), the next call to B<UCS::Status>
will replace the current message; when it ends in a carriage return
(C<\r>), the next call will overwrite the current message from the
start.  (This is the usual effect of B<print>ing such control
characters, and will be simulated in Perl/Tk interfaces).

=item UCS::Splash();

Displays a UCS splash screen with UCS version information and
copyright, e.g. during the start-up phase of a larger UCS/Perl script.

=item $UCS::Verbose = 0;

The variable I<$UCS::Verbose> controls whether status messages and
warnings are printed on STDOUT and STDERR, respectively.  Verbose
output is enabled by default, and can be suppressed by setting
I<$UCS::Verbose> to 0.

=item @unique_values = UCS::Unique(@list);

Removes duplicate values from I<@list> and returns the remaining elements in
the original order.  Useful to avoid repretitions of variable names etc.


=back

=head1 MANIPULATING VARIABLE NAMES

=over 4

=item $std_vars = (@UCS::CoreVars, @UCS::DerivedVars);

Names of B<core> and B<derived variables>.

=item $ok = UCS::ValidKey($key);

Returns true iff I<$key> is a valid UCS identifier, which may be used
as an AM key or in the name of a user-defined variable.

=item $ok = UCS::ValidName($name);

Returns true iff I<$name> is a valid UCS variable name, i.e. either a
standard variable (core or derived) , an association score or ranking,
or a user-defined variable.  See L<ucsfile> for details on the UCS
naming conventions.

=item $type = UCS::VarType($name);

Determines the data type of a variable from its name I<$name>,
according to the UCS naming conventions.  Possible data types are
C<BOOL> (Boolean, 0/1), C<INT> (signed integer), C<DOUBLE>
(double-precision floating-point), and C<STRING> (string value).

=item ($spec, $key) = UCS::SplitName($name);

Splits the variable name I<$name> of an association score, ranking, or
user-defined variable into the specifier I<$spec> and the key
I<$key>. I<$spec> will be one of C<am>, C<r>, C<b>, C<f>, C<n>, or
C<x>. If I<$name> is invalid or the name of a standard variable,
(undef, I<$name>) is returned.

=item @matches = UCS::Match($pattern, @names);

Extract strings from I<@names> that match the UCS B<wildcard pattern>
I<$pattern>.  The pattern may contain literal characters C<A-Z a-z 0-9
.> and the wildcards C<?>, C<*>, and C<%>.

  ?  ...  arbitrary character
  *  ...  arbitrary substring without "."
  %  ...  arbitrary string

Thus, the pattern C<%> selects all field names, C<*> selects the names of 
core and derived fields, C<am.%> all AM scores, etc.  See L<ucsexp> for
more examples.

=back


=head1 ASSOCIATION MEASURE REGISTRY

This B<registry> maintains a list of association measures, which are
automatically available to all UCS/Perl scripts.  Association measures are
identified by their B<key>, which must be a valid UCS identifier.  Association
scores for a measure with the key C<fisher>, for instance, will be stored in
the variable C<am.fisher>, and the corresponding rankings in the variable
C<r.fisher>.  A wide range of predefined association measures can be imported
from the B<UCS::AM> module and several add-on packages (see L<the UCS::AM
manpage|UCS::AM>).

=over 4

=item @registered_AMs = UCS::AM_Keys();

The B<UCS::AM_Keys> function returns the keys of all currently registered
association measures as an unordered list.  (Note that no association measures
are defined unless B<UCS::AM> and/or the add-on packages have been imported.)

=item $ok = UCS::AM($key);

Returns true if an association measure is registered under I<$key>.

=item $full_name = UCS::AM_Name($key);

Returns a long and descriptive name for the association measure
identified by I<$key>.  This name should be suitable for presentation
to the user in a selection dialogue.

=item $description = UCS::AM_Description($key);

An optional lengthy description of the association measure identified
by I<$key>.  I<$description> is a single string but will usually
contain linebreaks (C<\n>), which may need to be removed for automatic
justification (e.g. in a Perl/Tk interface).

=item $exp = UCS::AM_Expression($key);

Returns the equation of the association measure I<$key>, compiled into
a B<UCS::Expression> object.  Call the B<eval> or B<evalloop> method
of I<$exp> to compute association scores (see L<UCS::Expression>).
The sourcecode of this expression can be retrieved with the B<string>
method (which is especially useful for built-in association measures).

=item $score = UCS::Eval_AM($key, $arghash);

The B<UCS::Eval_AM> function is a convenient and shorter alternative,
and is equivalent to:

  $exp = UCS::AM_Expression($key);
  $score = $exp->eval($arghash);

It incurs considerable overhead when association scores are
calculated for multiple pair types (because of the repeated lookup of
I<$key> in the AM registry), and should be avoided in tight loops.
(See L<UCS::Expression> for some comments on efficiency.)

=item @packages = UCS::Load_AM_Package($name, ...);

Load one or more of the built-in AM packages as specified by the function
arguments.  I<$name> must match the last part of the corresponding module
name, e.g. C<'HTest'> to load the B<UCS::AM::HTest> package.  I<$name> is
case-insensitive and may be abbreviated to a unique prefix.  The special name
C<'ALL'> (or C<'all'>) loads all available add-on packages, while the empty
string C<''> loads the basic measures from B<UCS::AM>.
B<UCS::Load_AM_Package> returns a list containing the full names of all loaded
packages (with duplicates removed).  If there is no match for I<$name>, an
empty list is returned.

=item $ok = UCS::Register_AM($key, $name, $equation [, $description]);

The B<UCS::Register_AM> function is used to register a new association
measure, or overwrite an existing one with a new definition.  I<$key> is the
identification key of the new measure, I<$name> a descriptive name,
I<$equation> the measure's equation in the form of an (uncompiled) UCS
expression, and I<$description> an optional multi-line description.
I<$equation> may also be an object of class B<UCS::Expression> (which is
cloned rather than re-compiled), enabling the use of advanced features such as
parametric expressions.

The function call returns true if the new measure has been
successfully registered.  A false return value indicates that
compilation of I<$equation> into an B<UCS::Expression> object failed.
The B<UCS::Register_AM> function will B<die> if I<$key> is not a valid
UCS identifier.

The example below shows the code used to register the B<t-score> measure
(Church I<et. al.> 1991) which has been widely used in English lexicography.

  $ok = UCS::Register_AM  "tscore",
    "t-score measure (Church et. al. 1991)",
    '(%O11% - %E11%) / sqrt(%O11%)',
    "The t-score measure applies Student's t-test to ...";
  die "Syntax error in UCS expression for t-score measure"
    unless $ok;


=back



=head1 SEE ALSO

Type C<ucsdoc ucsintro> for an introduction to UCS/Perl and an overview of
its components (in the MODULES and PROGRAMS sections).


=head1 COPYRIGHT

Copyright 2003 Stefan Evert.

This software is provided AS IS and the author makes no warranty as to
its use and performance. You may use the software, redistribute and
modify it under the same terms as Perl itself.

=cut
