## Filename: UCS/AM.pm
## Modified: Thu Sep  9 21:32:09 2004 (evert)   
##   Author: Stefan Evert
##  Purpose: built-in association measures

package UCS::AM;

use UCS;

UCS::Register_AM "random", 
  "random scores between 0 and 1",
  'rand',
  "Random association scores simulate random selection of pair types."
  or UCS::Die "Error in UCS expression for built-in association measure am.random";

UCS::Register_AM "frequency", 
  "coooccurrence frequency",
  '%f%',
  ""
  or UCS::Die "Error in UCS expression for built-in association measure am.frequency";

UCS::Register_AM "MI", 
  "MLE of pointwise mutual information (MI)",
  'log10(%O11% / %E11%)',
  ""
  or UCS::Die "Error in UCS expression for built-in association measure am.MI";

UCS::Register_AM "relative.risk", 
  "MLE of (log) relative risk coefficient",
  '(%O12% > 0) ? log10((%O11% * %C2%) / (%O12% * %C1%)) : inf()',
  ""
  or UCS::Die "Error in UCS expression for built-in association measure am.relative.risk";

UCS::Register_AM "Dice", 
  "MLE of Dice coefficient",
  '2 * %O11% / (%R1% + %C1%)',
  ""
  or UCS::Die "Error in UCS expression for built-in association measure am.Dice";

UCS::Register_AM "gmean", 
  "MLE of geometric mean coefficient",
  '%O11% / sqrt(%N% * %E11%)',
  ""
  or UCS::Die "Error in UCS expression for built-in association measure am.gmean";

UCS::Register_AM "MS", 
  "MLE of minimum sensitivity (MS) coefficient",
  'min( %O11%/%R1%, %O11%/%C1% )',
  "(Pedersen and Bruce 1996)"
  or UCS::Die "Error in UCS expression for built-in association measure am.MS";

UCS::Register_AM "Jaccard", 
  "MLE of Jaccard coefficient",
  '%O11% / (%O11% + %O12% + %O21%)',
  ""
  or UCS::Die "Error in UCS expression for built-in association measure am.Jaccard";

UCS::Register_AM "odds.ratio", 
  "MLE of logarithmic odds ratio",
  '(%O12% > 0 and %O21% > 0) ? log10( (%O11% * %O22%) / (%O12% * %O21%) ) : inf()',
  ""
  or UCS::Die "Error in UCS expression for built-in association measure am.odds.ratio";

UCS::Register_AM "odds.ratio.disc", 
  "MLE of discounted (log) odds ratio",
  'log10( ((%O11%+0.5) * (%O22%+0.5)) / ((%O12%+0.5) * (%O21%+0.5)))',
  "It is common practice to add 0.5 to each factor in the log odds ratio\n".
  "in order to avoid overflow (where the argument of the logarithm is 0\n".
  "or +Infinity).  I refer to this method as 'discounting'."
  or UCS::Die "Error in UCS expression for built-in association measure am.odds.ratio.disc";

UCS::Register_AM "MI2", 
  "MLE of MI, but with numerator squared",
  'log10(%O11% * %O11% / %E11%)',
  ""
  or UCS::Die "Error in UCS expression for built-in association measure am.MI2";

UCS::Register_AM "MI3", 
  "MLE of MI, but with numerator cubed",
  'log10(%O11% * %O11% * %O11% / %E11%)',
  ""
  or UCS::Die "Error in UCS expression for built-in association measure am.MI3";

UCS::Register_AM "z.score", 
  "z-score for O11 against E11",
  '(%O11% - %E11%) / sqrt(%E11%)',
  "Under the point null hypothesis H0', X11 has a binomial distribution with\n".
  "expectation E11.  This measure approximates the correct binomial distribution\n".
  "with a normal distribution, and computes the corresponding z-score for O11."
  or UCS::Die "Error in UCS expression for built-in association measure am.z.score";

UCS::Register_AM "z.score.corr", 
  "z-score with Yates' correction",
  '(%O11% - %E11% + (%O11% > %E11% ? -0.5 : 0.5)) / sqrt(%E11%)',
  "Yates' continuity correction for a z-score (one-sided probability, upper tail)\n".
  "simply subtracts 0.5 from the observed cooccurrence frequency O11."
  or UCS::Die "Error in UCS expression for built-in association measure am.z.score.corr";

UCS::Register_AM "t.score", 
  "t-test for comparison of O11 against E11",
  '(%O11% - %E11%) / sqrt(%O11%)',
  "Church et. al. (1991) describe the highly suspicious application of Student's\n".
  "t-test to compare the observed values of O11 and E11.  With some simplifications,\n".
  "they arrive at this formula, which is strikingly similar to the z-score measure."
  or UCS::Die "Error in UCS expression for built-in association measure am.t.score";

# store formula for chi-squared statistic (homogeneity version) in (local) variable
my $chi_squared = '%N% * (%O11% * %O22% - %O12% * %O21%) ** 2 / (%R1% * %R2% * %C1% * %C2%)';
my $chi_squared_yates = '%N% * (abs(%O11% * %O22% - %O12% * %O21%) - %N% / 2) ** 2 / (%R1% * %R2% * %C1% * %C2%)';

UCS::Register_AM "chi.squared", 
  "Pearson's chi-squared statistic (one-sided)",
  'my $chi2 = '.$chi_squared.';
   (%O11% >= %E11%) ? $chi2 : -$chi2',
  "In order to obtain a one-sided version of Pearson's chi-squared test, the resulting\n".
  "test statistic is negated iff O11 < E11.  Thus, a positive score indicates positive\n".
  "association, and a negative score negative association."
  or UCS::Die "Error in UCS expression for built-in association measure am.chi.squared";

UCS::Register_AM "chi.squared.corr",
  "Pearson's chi-squared statistic (one-sided, Yates' correction)",
  'my $chi2 = '.$chi_squared_yates.';
   (%O11% >= %E11%) ? $chi2 : -$chi2',
  "One-sided chi-squared test with Yates' continuity correction applied."
  or UCS::Die "Error in UCS expression for built-in association measure am.chi.squared.corr";

my $log_likelihood = 
  '2*( ((%O11% > 0) ? %O11% * log(%O11% / %E11%) : 0) +
       ((%O12% > 0) ? %O12% * log(%O12% / %E12%) : 0) +
       ((%O21% > 0) ? %O21% * log(%O21% / %E21%) : 0) +
       ((%O22% > 0) ? %O22% * log(%O22% / %E22%) : 0)
     )';

UCS::Register_AM "log.likelihood", 
  "Dunning's log-likelihood ratio (one-sided)",
  'my $g2 = '.$log_likelihood.';
   (%O11% >= %E11%) ? $g2 : -$g2',
  "In order to obtain a one-sided version of Dunning's log-likelihood ratio, the resulting\n".
  "test statistic is negated iff O11 < E11.  Thus, a positive score indicates positive\n".
  "association, and a negative score negative association."
  or UCS::Die "Error in UCS expression for built-in association measure am.log.likelihood";

UCS::Register_AM "Poisson.Stirling", 
  "Poisson likelihood simplified with Stirling's formula",
  '%O11% * (log(%O11%) - log(%E11%) - 1) / log(10)',
  "Quasthoff and Wolff (2002) use Stirling's formula to simplify the expression for\n".
  "the (negative logarithm of the) Poisson likelihood.  Their formula is based on natural\n".
  "logarithms and is rescaled to base 10 logarithms in this implementation."
  or UCS::Die "Error in UCS expression for built-in association measure am.Poisson.Stirling";

UCS::Register_AM "Poisson.pv", 
  "Poisson test for null distribution of O11 (one-sided, p-value)",
  '(%O11% > 0) ? lppois(%O11%-1, %E11%, "upper") : 0',
  ""
  or UCS::Die "Error in UCS expression for built-in association measure am.Poisson.pv";

UCS::Register_AM "Fisher.pv", 
  "Fisher's exact test (one-sided, p-value)",
  'lphyper(%O11%-1, %R1%, %R2%, %C1%, %C2%, "upper")',
  ""
  or UCS::Die "Error in UCS expression for built-in association measure am.Fisher.pv";

UCS::Register_AM "average.MI",
  "average mutual information between indicator variables Y and Z",
  '((%O11% > 0) ? %O11% * log10(%O11% / %E11%) : 0) +
   ((%O12% > 0) ? %O12% * log10(%O12% / %E12%) : 0) +
   ((%O21% > 0) ? %O21% * log10(%O21% / %E21%) : 0) +
   ((%O22% > 0) ? %O22% * log10(%O22% / %E22%) : 0)',
  "This information-theoretically sound measure of the mutual information between\n".
  "the occurrences of two word types u and v happens to be identical to Dunning's\n".
  "log-likelihood measure (except for a scaling factor).  While it is customary to\n".
  "express mutual information in bits (using base 2 logarithms), this implementation\n".
  "follows the UCS convention of using base 10 logarithms and multiplies with the\n".
  "size N, so that the computed scores cannot be interpreted as information units."
  or UCS::Die "Error in UCS expression for built-in association measure am.global.MI";

UCS::Register_AM "local.MI", 
  "contribution to average mutual information of full cooccurrence data",
  '%O11% * log10(%O11% / %E11%)',
  "Local mutual information measures the contribution of the given pair type to the\n".
  "global mutual information between the components of cooccurring pairs (described\n".
  "by the random variables U and V).  Note that this measure uses base 10 logarithms\n".
  "and multiplies with the sample size N, so that the resulting scores cannot be\n".
  "interpreted as 'bits of information'."
  or UCS::Die "Error in UCS expression for built-in association measure am.local.MI";


# TODO: should record some version of MI.fvg (mutual information variant for FVG) 
# in some script or so, and convert it to the conventional decimal units.
# equation: -|MI - 5.75| (for MI measured in bits)

# TODO: perhaps entropy measure from previous work, measuring local contribution
# to global cross-entropy between joint distribution (U, V) and the distribution
# under the null hypothesis that U and V are independent (multiplied with N * log N).
# Equation: - %O11% * log10( %E11% )

return 1;


UCS::Register_AM "", 
  "",
  '',
  ""
  or UCS::Die "Error in UCS expression for built-in association measure am.";



1;


__END__


=head1 NAME

UCS::AM - Built-in association measures


=head1 SYNOPSIS

  use UCS;
  use UCS::AM;

  @builtin_AMs = UCS::AM_Keys();

  # random
  # frequency
  # z.score
  # z.score.corr
  # t.score
  # chi.squared
  # chi.squared.corr
  # log.likelihood
  # Poisson.Stirling
  # Poisson.pv
  # Fisher.pv
  # MI
  # MI2
  # MI3
  # relative.risk
  # odds.ratio
  # odds.ratio.disc
  # Dice
  # gmean
  # MS
  # Jaccard
  # average.MI
  # local.MI



=head1 DESCRIPTION

This module contains definitions for a wide range of B<association
measures>.  When the B<UCS::AM> module is imported, the built-in
measures are registered with the B<UCS> core library (see L<UCS> for
details on how to access registered association measures).

The following section gives a full listing of the built-in association
measures from the B<UCS::AM> module with short explanations.  Please refer to
F<http://www.collocations.de/AM/> for the full equations and references.
Further association measures can be imported from B<add-on packages> 
(see the section on L<Add-On Packages|"ADD-ON PACKAGES"> below).


Note that some association measures produce infinite values (I<+inf> or
I<-inf>).  The logarithm of infinity is represented by the return value of the
built-in B<inf> function (see L<the UCS::Expression::Func
manpage|UCS::Expression::Func>).  The association scores of measures with the
suffix B<.pv> can be interpreted as probabilities (the likelihood of the
observed data or the p-value of a statistical hypothesis test).  Such
probabilities are given as B<negative base 10 logarithms>, ranging from 0 to
I<+inf>.  Measures with the suffix B<.tt> (for I<two-tailed>) are derived from
two-sided statistical hypothesis tests.  One-sided versions of these tests are
provided under the same name, but without the suffix.


=head1 BUILT-IN ASSOCIATION MEASURES

=over 4

=item B<random>

Random numbers between 0 and 1 as association scores simulate random
selection of pair types and are used to break ties when sorting a data
set.

=item B<frequency>

Cooccurrence frequency of the pair type.  This association measure is
used to sort data sets by frequency, but requires some systematic
method for breaking ties.


=item B<z.score>

A z-score for the observed cooccurrence frequency O11 compared to the
expected frequency E11.  The value represents a standardised normal
approximation of the binomial sampling distribution of O11 under the
point null hypothesis of independence.

=item B<z.score.corr>

A z-score for O11 compared to E11 with Yates' continuity correction
applied.

=item B<t.score>

Church et al (1991) use Student's t-test to compare the observed
cooccurrence frequency O11 to the null expectation E11 estimated from
the sample (which is a random variate as well), applying several
approximations to simplify the B<t.score> equation.  The computed
value is a t-score with degrees of freedom roughly equal to the sample
size I<N>.  This application of the t-test is highly questionable,
though, and produces extremely conservative results.

=item B<chi.squared>

One-sided version of Pearson's chi-squared test for the independence of rows
and columns in a 2x2 contingency table.  Positive scores indicate positive
association (I<< O11 > E11 >>), and negative scores indicate negative
association (I<< O11 < E11 >>).  The distinction between positive and negative
association is unreliable for small absolute values of the test statistic.
Under the null hypothesis, the one-sided B<chi.squared> statistic approximates
a normal distribution (as the signed root of a chi-squared distribution with
one degree of freedom).

=item B<chi.squared.corr>

One-sided version of Pearson's chi-squared test for the independence of rows
and columns in a 2x2 contingency table, with Yates' continuity correction
applied.

=item B<log.likelihood>

One-sided version of the log-likelihood statistic suggested by Dunning (1993),
a likelihood ratio test for independence of rows and columns in a 2x2
contingency table (Dunning introduced the measure as a test for homogeneity of
the table columns, i.e. equal success probabilites of two independent binomial
distributions).  Positive scores indicate positive association (I<< O11 > E11
>>), and negative scores indicate negative association (I<< O11 < E11 >>).
The distinction between positive and negative association is unreliable for
small absolute values of the test statistic.  Under the null hypothesis, the
one-sided B<log.likelihood> statistic approximates a normal distribution (as
the signed root of a chi-squared distribution with one degree of freedom).

=item B<Poisson.Stirling>

Approximation of the likelihood of the observed cooccurrence frequency I<O11>
under the point null hypothesis of independence (so that the expected
frequency is I<E11>).  The measure is derived from B<Poisson.likelihood> (in
the L<UCS::AM::HTest|UCS::AM::HTest> module) using Stirling's formula,
resulting in a simple expression that can easily be evaluated.  This measure
was proposed by Quasthoff and Wolff (2002) and has been re-scaled to base 10
logarithms to allow a direct comparison with B<Poisson.likelihood>.

=item B<Poisson.pv>

Significance (one-sided p-value) of an exact Poisson test for the observed
cooccurrence frequency O11 compared to the expected frequency E11 under the
point null hypothesis of independence.  This test is based on a Poisson
approximation of the correct binomial sampling distribution of O11.  It is
numerically and analytically much easier to handle than the binomial test.

=item B<Fisher.pv>

Significance (one-sided p-value) of Fisher's exact test for
independence of rows and columns in a 2x2 contingency table with fixed
marginals.  This test is widely accepted as the most appropriate
independence test for contingency tables (cf. Yates 1984).  Its use as
an association measure was suggested by Pedersen (1996).

=item B<MI>

Maximum-likelihood estimate of the base 10 logarithm of the I<mu>-value, which
is identical to pointwise mutual information between the events describing
occurrences of a pair's components.  Note that mutual information is measured
in I<decimal units> rather than the customary I<bits>.  The theoretical range
is from I<-inf> to I<+inf>, but the actural range for a given data set is
restricted depending on the sample size I<N>.

=item B<MI2>

A heuristic variant of B<MI> where the numerator is squared in order
to discount low-frequency pairs.  This measure also has some
theoretical justification, being the square of the B<gmean> measure.

=item B<MI3>

Another heuristic variant of B<MI> where the numerator is cubed, which
boosts the discounting effect considerably.  

=item B<relative.risk>

Maximum-likelihood estimate of the logarithmic relative risk
coefficient of association strength (base 10 logarithm).  Ranges from
I<-inf> to I<+inf>.

=item B<odds.ratio>

Maximum-likelihood estimate of the logarithmic odds ratio as a
coefficient of association strength (base 10 logarithm).  Ranges from
I<-inf> to I<+inf>.

=item B<odds.ratio.disc>

A "discounted" version of B<odds.ratio>, adding 0.5 to each factor in the
equation.  This modification of the odds ratio is commonly used to avoid
infinite values, but does not seem to have a theoretical foundation.

=item B<Dice>

Maximum-likelihood estimate of the Dice coefficient of association
strength.  Ranges from 0 to 1.

=item B<Jaccard>

Maximum-likelihood estimate of the Jaccard coefficient of association
strength, which is equivalent to B<Dice> (i.e., there is a strictly monotonic
mapping between the two association scores).  Ranges from 0 to 1.

=item B<MS>

Maximum-likelihood estimate of the I<minimum sensitivity> coefficient
suggested by Pedersen and Bruce (1996).  Ranges from 0 to 1.

=item B<gmean>

Maximum-likelihood estimate of the I<geometric mean> coefficient of
association strength.  Ranges from 0 to 1.


=item B<average.MI>

Maximum-likelihood estimate of the average mutual information between
the indicator variables X and Y marking instances of a pair type's
components.  This implementation uses base 10 logarithms and
multiplies the mutual information value with the sample size I<N> in
order to obtain readable values.  Interestingly, B<average.MI> is
identical to Dunning's log-likelihood measure (B<log.likelihood> and
its variants) except for a scaling factor.

=item B<local.MI>

Contribution of a given pair type to the (maximum-likelihood estimate
of the) average mutual information of I<all> cooccurrences.  Formally,
this is the mutual information between the random variables U and V,
which represent the component types of a pair token in the random
sample.

=back


=head1 ADD-ON PACKAGES

The B<UCS::AM> module provides a basic set of useful and well-known
association measures.  Except for the B<Poisson.pv> and B<Fisher.pv>, all
measures have simple equations that can be computed efficiently.  Further and
more specialised association measures can be imported from add-on packages.
Currently, the following packages are available:

  UCS::AM::HTest         variants of hypothesis tests, likelihood measures
  UCS::AM::Parametric    parametric association measures

These packages are implemented as Perl modules and can simply be loaded with
the B<use> operator.  Alternatively, the B<UCS::Load_AM_Package> function
provides a convenient interface, where only the last part of the package name
has to be specified, is case-insensitive, and may be abbreviated to a unique
prefix.  For instance, the B<UCS::AM::HTest> package can be loaded with the
specification C<'ht'>.  The empty string C<''> loads B<UCS::AM>, and C<'ALL'>
imports all available AM packages.  (See L<the UCS manpage|UCS> for details.)


=head1 COPYRIGHT

Copyright 2003 Stefan Evert.

This software is provided AS IS and the author makes no warranty as to
its use and performance. You may use the software, redistribute and
modify it under the same terms as Perl itself.

=cut
