## Filename: UCS/AM/HTest.pm
## Modified: Mon Dec  8 17:46:43 2003 (evert)   
##   Author: Stefan Evert
##  Purpose: built-in association measures: hypothesis tests and likelihood measures

package UCS::AM::HTest;

use UCS;

UCS::Register_AM "z.score.pv", 
  "z-score for O11 against E11 (p-value)",
  'my $z = (%O11% - %E11%) / sqrt(%E11%);  
   lpnorm($z, 0, 1, "upper")',
  "Asymptotically, the z-score has a standard normal distribution.  Since \n".
  "association measures should be one-sided, the upper tail probability is returned."
  or UCS::Die "Error in UCS expression for built-in association measure am.z.score.pv";

UCS::Register_AM "z.score.corr.pv", 
  "z-score with Yates' correction (p-value)",
  'my $z = (%O11% - %E11% - 0.5) / sqrt(%E11%);  
   lpnorm($z, 0, 1, "upper")',
  "Yates' continuity correction for a z-score (one-sided probability, upper tail)\n".
  "simply subtracts 0.5 from the observed cooccurrence frequency O11."
  or UCS::Die "Error in UCS expression for built-in association measure am.z.score.corr.pv";

UCS::Register_AM "t.score.pv", 
  "t-test for comparison of O11 against E11 (p-value)",
  'my $t = (%O11% - %E11%) / sqrt(%O11%);
   lpnorm($t, 0, 1, "upper")',
  "Church et. al. (1991) apply Student's t-test to indicator variables, so that O11\n".
  "and E11 represent a sample of size N.  Therfore, the t-score has an asymptotic\n".
  "t distribution with N-1 degrees of freedom.  N is very large and can be taken to\n".
  "equal Infinity, in which case the t distribution becomes a standard normal distribution."
  or UCS::Die "Error in UCS expression for built-in association measure am.t.score.pv";

# store formula for chi-squared statistic (homogeneity version) in (local) variable
my $chi_squared = '%N% * (%O11% * %O22% - %O12% * %O21%) ** 2 / (%R1% * %R2% * %C1% * %C2%)';
my $chi_squared_yates = '%N% * (abs(%O11% * %O22% - %O12% * %O21%) - %N% / 2) ** 2 / (%R1% * %R2% * %C1% * %C2%)';

UCS::Register_AM "chi.squared.tt", 
  "Pearson's chi-squared statistic (two-sided)",
  $chi_squared,
  "Pearson's chi-squared test for equal ratios (success probabilities of two independent\n".
  "binomial distributions).  This equation is identical to the test for independence of rows\n".
  "and columns in a 2x2 contingency table (multinomial distribution)."
  or UCS::Die "Error in UCS expression for built-in association measure am.chi.squared.tt";

UCS::Register_AM "chi.squared.corr.tt",
  "Pearson's chi-squared statistic (two-sided, Yates' correction)",
  $chi_squared_yates,
  "Pearson's chi-squared test with Yates' continuity correction applied."
  or UCS::Die "Error in UCS expression for built-in association measure am.chi.squared.corr.tt";

UCS::Register_AM "chi.squared.tt.pv", 
  "Pearson's chi-squared statistic (two-sided, p-value)",
  'my $chi2 = '.$chi_squared.'; 
   lpchisq($chi2, 1, "upper")',
  "The asymptotic chi-squared distribution of Pearson's test statistic yields a two-sided p-value."
  or UCS::Die "Error in UCS expression for built-in association measure am.chi.squared.tt.pv";

UCS::Register_AM "chi.squared.corr.tt.pv",
  "Pearson's chi-squared statistic (two-sided, Yates' correction, p-value)",
  'my $chi2 = '.$chi_squared_yates.'; 
   lpchisq($chi2, 1, "upper")',
  "Two-sided p-value for Pearson's chi-squared test with Yates' continuity correction."
  or UCS::Die "Error in UCS expression for built-in association measure am.chi.squared.corr.tt.pv";

UCS::Register_AM "chi.squared.pv",
  "Pearson's chi-squared statistic (one-sided, p-value)",
  'my $chi2 = '.$chi_squared.';
   my $z = (%O11% >= %E11%) ? sqrt($chi2) : -sqrt($chi2);
   lpnorm($z, 0, 1, "upper")',
  "It is a little tricky to obtain one-sided p-values for Pearson's chi-squared test.\n".
  "We have to use the fact that, with df=1, the one-sided square root of Pearson's\n".
  "statistic (which is negated iff O11 < E11) has an asymptotic standard normal distribution."
  or UCS::Die "Error in UCS expression for built-in association measure am.chi.squared.pv";

UCS::Register_AM "chi.squared.corr.pv",
  "Pearson's chi-squared statistic (one-sided, Yates' correction, p-value)",
  'my $chi2 = '.$chi_squared_yates.';
   my $z = (%O11% >= %E11%) ? sqrt($chi2) : -sqrt($chi2);
   lpnorm($z, 0, 1, "upper");',
  "One-sided p-value for Pearson's chi-squared test with Yates' continuity correction."
  or UCS::Die "Error in UCS expression for built-in association measure am.chi.squared.corr.pv";

my $log_likelihood = 
  '2*( ((%O11% > 0) ? %O11% * log(%O11% / %E11%) : 0) +
       ((%O12% > 0) ? %O12% * log(%O12% / %E12%) : 0) +
       ((%O21% > 0) ? %O21% * log(%O21% / %E21%) : 0) +
       ((%O22% > 0) ? %O22% * log(%O22% / %E22%) : 0)
     )';

UCS::Register_AM "log.likelihood.tt", 
  "Dunning's log-likelihood ratio (two-sided likelihood ratio test)",
  $log_likelihood,
  "Likelihood ratio test for equal success probabilities of two independent binomial\n".
  "distributions as suggested by Dunning (1993).  This simpler equation can be found\n".
  "in Dunning (1998).  Like the Pearson's chi-squared, this is a two-sided test"
  or UCS::Die "Error in UCS expression for built-in association measure am.log.likelihood.tt";

UCS::Register_AM "log.likelihood.tt.pv", 
  "Dunning's log-likelihood ratio (two-sided, p-value)",
  'my $g2 = '.$log_likelihood.';
   lpchisq($g2, 1, "upper")',
  "The log-likelihood ratio has an asymptotic chi-squared distribution with df=1."
  or UCS::Die "Error in UCS expression for built-in association measure am.log.likelihood.tt.pv";

UCS::Register_AM "log.likelihood.pv", 
  "Dunning's log-likelihood ratio (one-sided, p-value)",
  'my $g2 = '.$log_likelihood.';
   my $z = (%O11% >= %E11%) ? sqrt($g2) : -sqrt($g2);
   lpnorm($z, 0, 1, "upper")',
  "It is a little tricky to obtain one-sided p-values for Dunning's log-likelihood ratio.\n".
  "We have to use the fact that, with df=1, the one-sided square root of the ratio\n".
  "(which is negated iff O11 < E11) has an asymptotic standard normal distribution."
  or UCS::Die "Error in UCS expression for built-in association measure am.log.likelihood.pv";

UCS::Register_AM "multinomial.likelihood.pv", 
  "Multinomial likelihood of the observed contingency table",
  '-(lgamma(%N%+1) - %N% * log10(%N%)
     + %O11% * log10(%E11%) + %O12% * log10(%E12%)
     + %O21% * log10(%E21%) + %O22% * log10(%E22%)
     - lgamma(%O11%+1) - lgamma(%O12%+1) - lgamma(%O21%+1) - lgamma(%O22%+1))',
  "This is a simple implementation which does not try to avoid cancellation problems.\n".
  "Uses the log Gamma function from UCS::R to compute logarithms of factorials efficiently."
  or UCS::Die "Error in UCS expression for built-in association measure am.multinomial.likelihood.pv";

UCS::Register_AM "binomial.likelihood.pv", 
  "Binomial likelihood for null distribution of O11",
  'ldbinom(%O11%, %N%, %E11%/%N%)',
  "Uses binomial distribution from UCS::R for (hopefully) accurate results."
  or UCS::Die "Error in UCS expression for built-in association measure am.binomial.likelihood.pv";

UCS::Register_AM "Poisson.likelihood.pv", 
  "Poisson likelihood for null distribution of O11 (R implementation)",
  'ldpois(%O11%, %E11%)',
  "Uses Poisson distribution from UCS::R for (hopefully) accurate results."
  or UCS::Die "Error in UCS expression for built-in association measure am.Poisson.likelihood.R.pv";

UCS::Register_AM "Poisson.likelihood.Perl.pv", 
  "Poisson likelihood for null distribution of O11 (Perl implementation)",
  'my $p = exp(-%E11%);
   for my $k (1 .. %O11%) {
     $p *=  %E11% / $k;
   }
   -log10($p)',
  "Naive multiplication algorithm for Poisson likelihood, implemented in Perl."
  or UCS::Die "Error in UCS expression for built-in association measure am.Poisson.likelihood.pv";

UCS::Register_AM "hypergeometric.likelihood.pv", 
  "hypergeometric likelihood of observed table with fixed marginals",
  'ldhyper(%O11%, %R1%, %R2%, %C1%, %C2%)',
  ""
  or UCS::Die "Error in UCS expression for built-in association measure am.hypergeometric.likelihood.pv";

UCS::Register_AM "binomial.pv", 
  "binomial test for null distribution of O11 (one-sided, p-value)",
  '(%O11% > 0) ? lpbinom(%O11%-1, %N%, %E11%/%N%, "upper") : 0',
  ""
  or UCS::Die "Error in UCS expression for built-in association measure am.Poisson.pv";



return 1;


UCS::Register_AM "", 
  "",
  '',
  ""
  or UCS::Die "Error in UCS expression for built-in association measure am.";



1;


__END__


=head1 NAME

UCS::AM::HTest - More association measures based on hypothesis tests 


=head1 SYNOPSIS

  use UCS;
  use UCS::AM::HTest;

  @htest_AMs = UCS::AM_Keys();

  # z.score.pv
  # z.score.corr.pv
  # t.score.pv
  # chi.squared.tt
  # chi.squared.tt.pv
  # chi.squared.corr.tt
  # chi.squared.corr.tt.pv
  # chi.squared.pv
  # chi.squared.corr.pv
  # log.likelihood.tt
  # log.likelihood.tt.pv
  # log.likelihood.pv
  # binomial.pv
  # multinomial.likelihood.pv
  # hypergeometric.likelihood.pv
  # binomial.likelihood.pv
  # Poisson.likelihood.pv
  # Poisson.likelihood.Perl.pv



=head1 DESCRIPTION

This module contains some further B<association measures> based on
statistical hypothesis tests, most of which are variants of measures defined
in the B<UCS::AM> module.  There are also several likelihood measures, which
compute the probability of the observed contingency table rather than applying
a full hypothesis test.  The association measures defined in this module are
intended mainly for a detailed comparative study of the properties of the
significance-of-association class of AMs.  Casual users should stick with the
variants found in the B<UCS::AM> module.

The following section gives a full listing of the association measures defined
in the B<UCS::AM::HTest> module with short explanations.  Please refer to
F<http://www.collocations.de/AM/> for the full equations and references.  When
the module is imported, the additional measures are registered with the B<UCS>
core library (see L<the UCS manpage|UCS> for details on how to access
registered association measures).

The association scores of measures with the suffix B<.pv> can be interpreted
as probabilities (i.e. the likelihood of the observed data or the p-value of a
statistical hypothesis test).  Such probabilities are given as B<negative base
10 logarithms>, ranging from 0 to I<+inf> (I<+inf> is represented by the
return value of the built-in B<inf> function (see L<the UCS::Expression::Func
manpage|UCS::Expression::Func>).  Measures with the suffix B<.tt> (for
I<two-tailed>) are derived from two-sided statistical hypothesis tests.
One-sided versions of these tests are provided under the same name without the
suffix.


=head1 ASSOCIATION MEASURES

=over 4

=item B<z.score.pv>

The significance (one-sided p-value) corresponding to B<z.score>,
obtained from the distribution function of the standard normal
distribution.  (The B<z.score> measure computes a z-score for the
observed cooccurrence frequency O11 compared to the expected frequency
E11; see L<the UCS::AM manpage|UCS::AM> for details.)

=item B<z.score.corr.pv>

The significance (one-sided p-value) corresponding to B<z.score.corr>, 
a z-score for O11 against E11 with Yates' continuity correction applied.

=item B<t.score.pv>

The significance (one-sided p-value) corresponding to B<t.score>, obtained
from the distribution function of the standard normal distribution. Since the
number of degrees of freedom is very large, the t-distribution of the test
statistic is practically identical to the standard normal distribution
(t-distribution with I<df=inf>).  (The B<t.score> measure is an application of
Student's t-test to the comparison of O11 against E11; see L<the UCS::AM
manpage|UCS::AM> for details.)

=item B<chi.squared.tt>

Pearson's chi-squared test for independence of rows and columns in a
2x2 contingency table. The equation used in this implementation is
derived from the homogeneity version of the chi-squared test (for
equality of the success probabilities of two independent binomial
distributions), and is fully equivalent to that of the independence
test.  Note that Pearson's chi-squared test is two-sided.

=item B<chi.squared.tt.pv>

The significance (two-sided p-value) corresponding to
B<chi.squared.tt>, obtained from the chi-squared distribution with one
degree of freedom.

=item B<chi.squared.corr.tt>

Pearson's chi-squared test for independence of rows and columns in a
2x2 contingency table, with Yates' continuity correction applied
(two-sided test).

=item B<chi.squared.corr.tt.pv>

The significance (two-sided p-value) corresponding to B<chi.squared.corr.tt>.

=item B<chi.squared.pv>

The significance (one-sided p-value) corresponding to B<chi.squared>, the
one-sided version of Pearson's test for the independence of rows and columns
(see L<the UCS::AM manpage|UCS::AM> for details).  The p-value is obtained
from the standard normal distribution (since the signed square root of the
chi-squared test statistic has a standard normal distribution).

=item B<chi.squared.corr.pv>

The significance (one-sided p-value) corresponding to B<chi.squared.corr>, the
one-sided version of Pearson's chi-squared test with Yates' continuity
correction applied.  Again, the p-value is obtained from the standard normal
distribution.

=item B<log.likelihood.tt>

The log-likelihood statistic suggested by Dunning (1993), a likelihood
ratio test for independence of rows and columns in a 2x2 contingency table.
(Dunning introduced the statistic as a test for homogeneity of the table columns,
i.e. equal success probabilites of two independent binomial distributions).  
Note that all likelihood ratio tests are two-sided tests.

=item B<log.likelihood.tt.pv>

The significance (two-sided p-value) corresponding to
B<log.likelihood.tt>, obtained from the chi-squared distribution with
one degree of freedom.

=item B<log.likelihood.pv>

The significance (one-sided p-value) corresponding to B<log.likelihood>, the
one-sided version of Dunning's likelihood ratio test (see L<the UCS::AM
manpage|UCS::AM> for details).  The p-value is obtained from the standard
normal distribution (since the signed square root of the log-likelihood
statistic has a standard normal distribution.)

=item B<binomial.pv>

Significance (one-sided p-value) of an exact binomial test for the observed
cooccurrence frequency O11 compared to the expected frequency E11 under the
point null hypothesis of independence.  This test is computationally expensive
and may be numerically unstable, so use with caution. (This is also the reason
why it is not included in the B<UCS::AM> module.)

=item B<multinomial.likelihood.pv>

Likelihood of the observed contingency table under the point null
hypothesis of independence (i.e. with expected frequencies E11, E12,
E21, and E22 estimated from the observed table).

=item B<hypergeometric.likelihood.pv>

Likelihood of the observed contingency table under the null hypothesis
of independence of rows and columns, with all marginal frequencies
fixed to the observed values.

=item B<binomial.likelihood.pv>

Binomial likelihood of the observed cooccurrence frequency O11 under the point
null hypothesis (with expected frequency E11 estimated from the observed
table).  This function is relatively slow and may be numerically unstable, so
use with caution.

=item B<Poisson.likelihood.pv>

Poisson approximation of the binomial likelihood B<binomial.likelihood.pv>,
which is numerically and analytically more manageable.

=item B<Poisson.likelihood.Perl.pv>

Alternative version of B<binomial.likelihood.pv>, based on a direct
Perl implementation of the naive multiplicative algorithm.  


=back


=head1 COPYRIGHT

Copyright 2003 Stefan Evert.

This software is provided AS IS and the author makes no warranty as to
its use and performance. You may use the software, redistribute and
modify it under the same terms as Perl itself.

=cut
