## Filename: UCS/AM/Parametric.pm
## Modified: Sun Dec 14 22:51:55 2003 (evert)   
##   Author: Stefan Evert
##  Purpose: built-in association measures: parametric measures

package UCS::AM::Parametric;

use UCS;
use UCS::Expression


# compile parametric UCS expression for MI confidence intervals
# - desired value is the lower boundary mu* of the Poisson confidence interval for the mu coefficient
# - the Poisson probability P(X11 >= O11 | mu) is given by rgamma(O11, mu * E11)
# - for a two-sided confidence level alpha, choose mu* such that P(X11 >= O11 | mu*) = alpha/2
# - alpha is given as a negative base 10 logarithm (note that log10(a/2) = log10(a) - log10(2))
$MI_conf = new UCS::Expression 'my $lp = -%ALPHA% - log10(2); log10(lrgamma_inv(%O11%, $lp) / %E11%)', "ALPHA" => 2;

UCS::Register_AM "MI.conf", 
  "Conservative estimate for MI at two-sided confidence level ALPHA (parameter, negative base 10 log)",
  $MI_conf,
  ""
  or UCS::Die "Error in UCS expression for built-in association measure am.MI.conf";

foreach $alpha (2, 3, 5, 10, 50, 100, 1000) {
  $MI_conf->set_param("ALPHA", $alpha);
UCS::Register_AM "MI.conf.$alpha", 
    "Conservative estimate for MI at two-sided confidence level 1E-$alpha",
  $MI_conf,
    ""
      or UCS::Die "Error in UCS expression for built-in association measure am.MI.conf.$alpha";
}

# compile parametric UCS expression for Poisson test with H0: mu = x
$poisson_mu = new UCS::Expression '(%O11% > 0) ? lppois(%O11%-1, %E11% * %MU%, "upper") : 0', "MU" => 1;

UCS::Register_AM "Poisson.mu.pv", 
  "Poisson test with null hypothesis mu = MU > 1 (parameter)",
  $poisson_mu,
  ""
  or UCS::Die "Error in UCS expression for built-in association measure am.Poisson.mu.pv";

foreach $mu (2, 3, 5, 10, 50, 100, 1000, 10000) {
  $poisson_mu->set_param("MU", $mu);
  UCS::Register_AM "Poisson.mu.$mu.pv", 
      "Poisson test with null hypothesis mu = $mu",
	$poisson_mu,
	  ""
	    or UCS::Die "Error in UCS expression for built-in association measure am.Poisson.mu.$mu.pv";
}


return 1;

UCS::Register_AM "", 
  "",
  '',
  ""
  or UCS::Die "Error in UCS expression for built-in association measure am.";



1;


__END__


=head1 NAME

UCS::AM::Parametric - Parametric association measures


=head1 SYNOPSIS

  use UCS;
  use UCS::AM::Parametric;

  @parametric_AMs = UCS::AM_Keys();

  # MI.conf 
  # MI.conf.<n>        [<n> = 2, 3, 5, 10, 50, 100, 1000]
  # Poisson.mu.pv
  # Poisson.mu.<n>.pv  [<n> = 2, 3, 5, 10, 50, 100, 1000, 10000]


=head1 DESCRIPTION

This module contains some parametric B<association measures>, which are
parametrised extensions of measures defined in the basic B<UCS::AM> module.
Parametric measures are a recent development in cooccurrence statistics, and
the choice of appropriate parameter values is still very much a research
question.  Parametric measures will often be computationally expensive and
may be numerically unstable, so novice users are advised to use the basic 
measures from the B<UCS::AM> module instead.

The following section gives a full listing of the parametric association
measures defined in the B<UCS::AM::Parametric> module with short explanations.
Please refer to F<http://www.collocations.de/AM/> for the full equations and
references.  When the module is imported, the additional measures are
registered with the B<UCS> core library (see L<the UCS manpage|UCS> for
details on how to access registered association measures).

The association scores of measures with the suffix B<.pv> can be interpreted
as probabilities (i.e. the likelihood of the observed data or the p-value of a
statistical hypothesis test).  Such probabilities are given as B<negative base
10 logarithms>, ranging from 0 to I<+inf> (I<+inf> is represented by the
return value of the built-in B<inf> function (see L<the UCS::Expression::Func
manpage|UCS::Expression::Func>).


=head1 ASSOCIATION MEASURES

=over 4

=item B<MI.conf>

Conservative estimate for the base 10 logarithm of the I<mu>-value (whose 
maximum-likelihood estimate is given by the B<MI> measure).  The association
score computed by B<MI.conf> is the lower endpoint of a two-sided confidence
interval for I<mu> at significance level B<alpha>, which is specified by the
C<ALPHA> parameter (as a negative base 10 logarithm).  The "usual" significance 
levels I<.01> and I<.001> correspond to C<ALPHA=2> and C<ALPHA=3>, respectively.

Please duplicate the B<UCS::Expression> object returned by
C<UCS::AM_Expression("MI.conf")> before modifying the C<ALPHA> parameter.

=item B<MI.conf.I<ALPHA>>

Versions of B<MI.conf> with the C<ALPHA> parameter pre-set to the value
specified as part of the name.  Available B<I<ALPHA>> values are B<2>, B<3>,
B<5>, B<10>, B<50>, B<100>, and B<1000>.  For instance, B<MI.conf.10> computes
a two-sided confidence interval at significance level 1E-10. 

Do not modify the C<ALPHA> parameter of these association measures (in the
B<UCS::Expression> object returned by the B<UCS::AM_Expression> function).

=item B<Poisson.mu.pv>

Poisson test for I<O11> under the modified point null hypothesis I<< pi = p1 *
p2 * mu >> (rather than the independence hypothesis I<< pi = p1 * p2 >> used
by the B<Poisson.pv> measure).  The (non-logarithmic) value of I<mu> is given
by the C<MU> parameter.  For C<MU=1>, the association scores computed by
B<Poisson.mu.pv> are identical to those of B<Poisson.pv>.

Please duplicate the B<UCS::Expression> object returned by
C<UCS::AM_Expression("Poisson.mu.pv")> before modifying the C<MU> parameter.

=item B<Poisson.mu.I<MU>.pv>

Versions of B<Poisson.mu.pv> with the C<MU> parameter pre-set to the value
specified as part of the name.  Available B<I<MU>> values are B<2>, B<3>,
B<5>, B<10>, B<50>, B<100>, B<1000>, and B<10000>.  

Do not modify the C<MU> parameter of these association measures (in the
B<UCS::Expression> object returned by the B<UCS::AM_Expression> function).




=back




=head1 COPYRIGHT

Copyright 2003 Stefan Evert.

This software is provided AS IS and the author makes no warranty as to
its use and performance. You may use the software, redistribute and
modify it under the same terms as Perl itself.

=cut
