## Filename: UCS/DS.pm
## Modified: Thu Jul 31 19:14:20 2003 (evert)   
##   Author: Stefan Evert
##  Purpose: base class for data set objects (file streams or in-memory representations)

## UCS::DS objects store information about which fields are defined in a data set,
## and the order in which those fields do (or should occur) in a data stream.
## Both file streams and in-memory representations inherit from UCS::DS.

package UCS::DS;

use UCS;

sub new {
  my $class = shift;
  my $self = {};
  $self->{COLS} = {};                  # $self->{COLS}->{$name} = $column_index (starting from 0, no gaps)
  $self->{TYPE} = {};		       # $self->{TYPE}->{$name} = UCS::VarType($name);
  $self->{TEMP} = {};		       # $self->{TEMP}->{$name} = $is_temporary
  $self->{NEXT_FREE_COL} = 0;          # $self->{NEXT_FREE_COLS_INDEX} = <lowest unused column index>
  $self->{COMMENTS} = [];	       # list of chomped and de'#'ed comment lines
  $self->{GLOBALS} = {};	       # $self->{GLOBALS}->{$var} = $value
  bless($self, $class);
  $self->add_vars(@_);		       # remaining arguments are field names for init
  return $self;
}

sub DESTROY {
  # nothing do to in destroy
}

# this method is used internally by add_vars()
sub add_single_var ( $$ ) {
  my ($self, $name) = @_;
  UCS::Die "UCS::DS:  Cannot add invalid variable '$name' in add_vars() method."
      unless UCS::ValidName($name);
  if (not exists $self->{COLS}->{$name}) {
    my $index = $self->{NEXT_FREE_COL};
    $self->{COLS}->{$name} = $index;
    $self->{NEXT_FREE_COL} = $index + 1;
    $self->{TYPE}->{$name} = UCS::VarType($name);
    $self->{TEMP}->{$name} = 0;
  }
  else {
    # the attempt to add an existing variable will be silently ignored
  }
}

sub add_vars {
  my $self = shift;
  foreach my $name (@_) {
    $self->add_single_var($name);
  }
}

# this method is used internally by delete_vars()
sub delete_single_var ( $$ ) {
  my ($self, $name) = @_;
  UCS::Die "UCS::DS:  Cannot delete invalid variable '$name' in delete_vars() method."
      unless UCS::ValidName($name);
  my $cols = $self->{COLS};
  if (exists $cols->{$name}) {
    my $index = $cols->{$name};
    foreach my $c (values %$cols) {
      $c-- if $c > $index;			# adjust higher indices
    }
    delete $cols->{$name};
    $self->{NEXT_FREE_COL} -= 1;		# highest previous index should be free now
  }
  else {
    # the attempt to delete a non-existing variables will be silently ignored
  }
}

sub delete_vars {
  my $self = shift;
  foreach my $name (@_) {
    $self->delete_single_var($name);
  }
}

sub var {
  my ($self, $name) = @_;
  return $self->{TYPE}->{$name};		# returns undef if variable doesn't exist
}

sub var_index {
  my ($self, $name) = @_;
  return $self->{COLS}->{$name};		# returns undef if variable doesn't exist
}

sub vars {
  my $self = shift;
  my $cols = $self->{COLS};
  my @names = sort { $cols->{$a} <=> $cols->{$b} } keys %$cols;
  return @names;
}

sub temporary {
  my $self = shift;
  my $name = shift;
  UCS::Die "UCS::DS:  temporary() method called on non-existing variable '$name'."
      unless $self->var($name);
  if (@_) {
    my $val = ($_[0]) ? 1 : 0;
    $self->{TEMP}->{$name} = $val;
  }
  else {
    return $self->{TEMP}->{$name};
  }
}

sub comments {
  my $self = shift;
  return @{$self->{COMMENTS}};
}

sub add_comments {
  my $self = shift;
  push @{$self->{COMMENTS}}, @_;
}

sub delete_comments {
  my $self = shift;
  $self->{COMMENTS} = [];
}

sub copy_comments {
  my $self = shift;
  my $ds = shift;
  UCS::Die "UCS::DS:  argument of copy_comments() method must be UCS::DS object (but is ".(ref $ds).")"
      unless $ds->isa(UCS::DS);
  $self->{COMMENTS} = [ $ds->comments ];	# discards old comments and makes copy of $ds's comments
}

sub globals {
  my $self = shift;
  UCS::Die "UCS::DS:  the globals() methods takes no arguments; did you mean \$ds->global(\$var)?"
      if @_ > 0;
  return sort keys %{$self->{GLOBALS}};
}

sub global {
  my $self = shift;
  my $var = shift;
  UCS::Die "UCS::DS:  missing argument in global() method; did you mean \$ds->globals?"
      unless defined $var;
  UCS::Die "UCS::DS:  illegal name '$var' for global variable in global() method."
      unless UCS::ValidKey($var);
  return $self->{GLOBALS}->{$var};		# happens to return undef if global variable doesn't exist
}

sub set_global {
  my ($self, $var, $value) = @_;
  UCS::Die "UCS::DS:  illegal name '$var' for global variable in set_global() method."
      unless UCS::ValidKey($var);
  $self->{GLOBALS}->{$var} = $value;
}

sub delete_global {
  my ($self, $var) = @_;
  UCS::Die "UCS::DS:  illegal name '$var' for global variable in delete_global() method."
      unless UCS::ValidKey($var);
  return delete $self->{GLOBALS}->{$var};
}

sub copy_globals {
  my $self = shift;
  my $ds = shift;
  UCS::Die "UCS::DS:  argument of copy_globals() must be UCS::DS object (but is ".(ref $ds).")."
      unless $ds->isa(UCS::DS);
  $self->{GLOBALS} = {};			# delete all existing global variables
  foreach my $v ($ds->globals) {
    $self->set_global($v, $ds->global($v));
  }
}


1;


__END__


=head1 NAME

UCS::DS - base class for data set implementations


=head1 SYNOPSIS

  use UCS::DS;

  $ds = new UCS::DS;			# "virtual" data set
  $ds->add_vars($name1, $name2, ...);	# append variables (= columns) in this order
  $ds->delete_vars($name1, ...);	# delete variables (column 'gaps' are closed)

  $type = $ds->var($name);		# check whether variable exists, returns data type
  $index = $ds->var_index($name);	# column index of variable
  @names = $ds->vars;			# list all variables in column order

  $ds->temporary($name, 1);		# mark variable as temporary (will not be saved)

  @lines = $ds->comments;		# ordered list of comment lines
  $ds->add_comments($line1, ...);       # append comment lines
  $ds->delete_comments;                 # delete all comments
  $ds->copy_comments($ds2);		# copy all comments from $ds2

  @global_vars = $ds->globals;          # unordered list of global variable names
  $value = $ds->global($var);		# return value of global variable $var
  $ds->set_global($var, $value);        # set value of global variable (may be new variable)
  $ds->delete_global($var);		# delete global variable
  $ds->copy_globals($ds2);              # copy global variables from $ds2


=head1 DESCRIPTION

B<UCS::DS> acts as a base class for B<data set> managers (either file
streams or in-memory representations).  A B<UCS::DS> object manages a
list of B<variables> (with names according to the UCS naming
conventions detailed in L<ucsfile>), and maps them to the B<column
indices> of a data set file.

It is always ensured that the column indices of a data set span a
contiguous range starting at 0.  New variables will be appended to the
existing columns in the order of declaration.  When a variables is
deleted, all columns to its right are shifted to fill the gap.

When it is available, B<UCS::DS> objects also store information from
the B<header> of a data set file.  This information includes B<comment
lines> and B<global variables> (see L<ucsfile> for details).


=head1 METHODS

=over 4

=item $ds = new UCS::DS;

Create a new B<UCS::DS> object, with an empty list of
variables. Normally, this constructor is only invoked implicitly by
derived classes.

=item $ds = new UCS::DS $name1, $name2, ...;

Creates a B<UCS::DS> object with the specified variables. Same as

  $ds = new UCS::DS;
  $ds->add_vars($name1, $name2, ...);

=item $ds->add_vars($name1, $name2, ...);

Add one or more variables I<$name1>, I<$name2>, ... to the data set.
Variables that are already defined will be silently ignored.  New
variables are appended to the existing columns in the specfied order.
I<$name1>, I<$name2>, ... must be valid UCS variable names.

=item $ds->delete_vars($name1, $name2, ...);

Delete the variables I<$name1>, I<$name2>, ... from the data set.
Variables that are not defined in the data set will be silently
ignored.  When a variable has been deleted, all columns to its right
are shifted to fill the gap.  All arguments must be valid UCS
variable names.

=item $type = $ds->var($name);

Check whether the variable I<$name> is defined in the data set I<$ds>.
Returns the data type of the variable (C<BOOL>, C<INT>, C<DOUBLE>, or
C<STRING>, see L<ucsfile>), or B<undef> if it does not exist.

=item $is_temp = $ds->temporary($name);

=item $ds->temporary($name, $val);

Mark variable I<$name> as temporary (if I<$val> is true) or permanent
(if I<$val> is false).  The single-argument version returns true if
the variable I<$name> is temporary.  Temporary variables are
interpreted by in-memory representations of data sets.  They may be
deleted automatically and will not be written to data set files.

=item $index = $ds->var_index($name);

Get column index of variable I<$name>.  I<$index> ranges from 0 to one
less than the number of variables in the data set.  Returns B<undef>
if the variable I<$name> does not exist in the data set.  It is
recommended to test this condition with the B<var> method first.

=item @names = $ds->vars;

Returns the names of all variables in this data set, sorted by their
column indices.  When saved to a data set file, the columns will
appear in this order.

=item @lines = $ds->comments;

Returns all comment lines as an ordered list (i.e. as they would appear in
a data set file).  Comment lines are B<chomp>ed and the initial C<#> character
(followed by an optional blank) is removed.

=item $ds->add_comments($line1, ...);

Add comment lines (which will be appended to existing comments).  Like
the data returned by the B<comments> method, I<$line1> etc. should not
begin with a C<#> character or end in a newline.

=item $ds->delete_comments;

Deletes all comment lines.

=item $ds->copy_comments($ds2);

Copies all comment lines from I<$ds2>, which must be an object derived
from B<UCS::DS>.  Existing comments of I<$ds> are overwritten.  This command
is equivalent to

  $ds->delete_comments;
  $ds->add_comments($ds2->comments);

=item @global_vars = $ds->globals;

Returns the names of all global variables in alphabetical order.
B<NB:> global variable names must be valid UCS identifiers.

=item $value = $ds->global($var);

Returns the value of a global variable I<$var> as a character string.
If the global variable I<$var> does not exist, returns B<undef>.

=item $ds->set_global($var, $value);

Set global variable I<$var> to the string I<$value>.  If I<$var> does
not exist, it is automatically added to the data set.

=item $ds->delete_global($var);

Delete a global variable.  If I<$var> does not exist, the method call
will be silently ignored.

=item $ds->copy_globals($ds2);

Copies all global variables and their values from I<$ds2>, which must
be an object derived from B<UCS::DS>.  Any existing global variables
off the data set I<$ds> will be erased.

=back


=head1 COPYRIGHT

Copyright 2003 Stefan Evert.

This software is provided AS IS and the author makes no warranty as to
its use and performance. You may use the software, redistribute and
modify it under the same terms as Perl itself.

=cut

