## Filename: UCS/DS/Memory.pm
## Modified: Tue Jan 13 20:28:29 2004 (evert)   
##   Author: Stefan Evert
##  Purpose: ASCII-format data set

package UCS::DS::Format;

use UCS;
use UCS::File;
use UCS::DS;
use UCS::DS::Memory;
use UCS::Expression::Func;

sub new {
  my $class = shift;
  my $ds = shift;
  UCS::Die "UCS::DS::Format:  object must be initalised with UCS::DS::Memory object"
      unless defined($ds) and ref($ds) eq "UCS::DS::Memory";
  my $self = {
	      DS => $ds,			# in-memory data set
	      MINDIGITS => 2,			# at least two digits after decimal point in fixed-point format
	      MAXDIGITS => 12,			# at most 12 digits in fixed-point format (otherwise use floating-point)
	      DIGITS => 8,			# eight significant digits for fixed-point and floating-point numbers
	      MODE => "table",			# default mode is simple table
	      PAGESIZE => 0,			# table is not split into pages by default
	      VARS => [$ds->vars],		# by default, all variables will be printed
	      };
  return bless($self, $class);
}

sub digits {
  my $self = shift;
  my $digits = shift;
  UCS::Die "UCS::DS::Format:  number of significant digits ($digits) must be >= 2"
      unless $digits >= 2;
  $self->{DIGITS} = $digits;
}

sub mode {
  my $self = shift;
  my $mode = lc(shift);
  if ($mode =~ /^t(ab(le)?)?$/) {
    $self->{MODE} = "table";
  }
  else {
    UCS::Die "UCS::DS::Format:  unknown mode '$mode' [supported modes: table]"
  }
}

sub pagelength {
  my $self = shift;
  my $pagesize = shift;
  $pagesize = 0 unless $pagesize;
  $self->{PAGESIZE} = $pagesize;
}

sub vars {
  my $self = shift;
  my @vars = ();
  foreach my $pattern (@_) {
    push @vars, UCS::Match($pattern, $self->{DS}->vars);
  }
  $self->{VARS} = [@vars];
}

sub print {
  my $self = shift;
  my $filename = (@_) ? shift : "-";
  my $temp = ($filename =~ /^\s*\|/) ? $filename : "> $filename";
  my $fh = UCS::File::Open($temp);
  my $ds = $self->{DS};
  my $size = $ds->size;				# effective size of data set
  my @vars = @{$self->{VARS}};
  my $digits = $self->{DIGITS};
  my $mindigits = $self->{MINDIGITS};
  my $maxdigits = $self->{MAXDIGITS};
  my $pagesize = $self->{PAGESIZE};
  $pagesize = $size if $pagesize <= 0;		# format entire table as a single page

  my $abort = 0;				# abort when SIGPIPE is received (probably means that output was piped into "less")
  local $SIG{PIPE} = sub { $abort = 1; };

  # format table one page at a time
  for (my $start = 1; $start <= $size and not $abort; $start += $pagesize) {
    $end = $start + $pagesize - 1;
    $end = $size if $end > $size;

    # in paged mode, every page is analysed and formatted independently, so we need an appropriate index
    my $save_index = $ds->{ACTIVE_INDEX};	# will temporarily change active index in page mode
    unless ($start == 1 and $end == $size) { 
      if ($save_index) {			# in paged mode, construct and activate index corresponding to current page
	$ds->make_index(".format", @{$ds->index($save_index)}[$start-1 .. $end-1]);
      }
      else {
	$ds->make_index(".format", $start .. $end);
      }
      $ds->activate_index(".format");
      my @X = @{$ds->index(".format")};
    }
    my $cur_size = $end - $start + 1;		# size of current page = virtual size of data set (with .format activated)

    # examine variables in order to choose formatting options
    %format = ();				# $format{$var}{OPTION} = $value
    # WIDTH = total width of field, FMT = sprintf format string or "BOOL", ALIGN = left/right
    foreach my $v (@vars) {
      my $type = UCS::VarType($v);
      $format{$v}{ALIGN} = ($type eq "STRING" and not $v eq "l1") ? "left" : "right";
      my $width = length($v);
      if ($type eq "BOOL") {			# BOOL = print yes / no / NA
	$format{$v}{FMT} = "BOOL";		# prints "yes" / "no"
	$width = max($width, 3);		# required width is at least 3
      }
      elsif ($type eq "STRING") {		# STRING = print string / NA
	$format{$v}{FMT} = '%s';
	foreach my $i (1 .. $cur_size) {
	  my $s = $ds->cell($v, $i);
	  my $w = (defined $s) ? length($s) : 2;
	  $width = $w if $w > $width;
	}
      }
      else { # "INT" or "DOUBLE" 
	my $r = $ds->summary($v);		# use summary to figure out best formatting
	my $max = $r->{MAX};
	my $min = $r->{MIN};
	if (not defined $max) {
	  $format{$v}{FMT} = '';		# only NA's -> format doesn't really matter, width = 2
	  $width = max($width, 2);
	}
	else {
	  my $absmax = $r->{ABSMAX};
	  my $negative = ($r->{MIN} < 0) ? 1 : 0;
	  if ($type eq "INT") {			# INT = print integer / NA 
	    $format{$v}{FMT} = '%d';
	    # max. number of digits (plus "-" sign if there are negative numbers)
	    my $w = int(log10($absmax+1)) + 1 + $negative; # ($absmax+1) to avoid rounding error when $absmax = 10^n 
	    $width = max($width, $w, 2);
	  }
	  else {				# DOUBLE = need to decide between fixed-point and floating-point format
	    my $unit = ($max - $min) / $cur_size; # approx. unit of measurement, assuming uniform distribution of values
	    $unit = $max if $unit == 0;
	    my $digits_before_point = int(log10($absmax)) + 1; # number of digits required before decimal point
	    $digits_before_point = max($digits_before_point, 1);
	    my $digits_after_point = $digits - int(log10($unit)) - 1; # number of digits after decimal point for desired precision
	    $digits_after_point = max($mindigits, $digits_after_point);
	    if ($digits_before_point + $digits_after_point <= $maxdigits) {
	      $format{$v}{FMT} = '%.'."$digits_after_point".'f'; # fixed-point format
	      my $w = $digits_before_point + $digits_after_point + 1 + $negative;
	      $width = max($width, $w, 2);
	    }
	    else {
	      $format{$v}{FMT} = '%.'.($digits-1).'e'; # floating-point format (engineering notation)
	      my $w = $digits + 1 + $negative + 5; # reserve five characters for exponent
	      $width = max($width, $w, 2);
	    }
	  }
	}
      }
      $format{$v}{WIDTH} = $width;
    }

    # format and print header
    my $error = 0;
    foreach my $v (@vars) {
      $fh->print(fmtcell($v, '%s', $format{$v}{ALIGN}, $format{$v}{WIDTH}), "  ");
    }
    print $fh "\n";
    foreach my $v (@vars) {
      print $fh "-" x $format{$v}{WIDTH}, "  ";
    }
    print $fh "\n";

    # format and print each row from current page 
    foreach my $n (1 .. $cur_size) {
      last if $abort;				# abort immediately after SIGPIPE
      foreach my $v (@vars) {
	$fh->print(fmtcell($ds->cell($v, $n), $format{$v}{FMT}, $format{$v}{ALIGN}, $format{$v}{WIDTH}), "  ");
      }
      print $fh "\n";
    }

    # print separator line (empty line) between pages
    print $fh "\n" 
      unless $end == $size;

    $ds->activate_index($save_index);		# restore previous index (may be undef)
    $ds->delete_index(".format");		# delete temporary index
  }

  $fh->close;
}

# internal helper function for formatting and padding cell values
#   $padded_string = fmtcell($value, $fmt, $align, $width);
sub fmtcell {
  my ($value, $fmt, $align, $width) = @_;
  my $string = "NA";
  if (defined $value) {
    if ($fmt eq "BOOL") { $string = ($value) ? "yes" : "no"; }
    else                { $string = sprintf $fmt, $value; }
  }
  my $l = length($string);
  UCS::Die "UCS::DS::Format:  formatted value '$string' is too wide for column (fmt='$fmt', width=$width)"
      unless $l <= $width;
  my $pad = $width - $l;
  if ($pad > 0) {
    if ($align eq "left") { $string = $string.(" " x $pad); }
    else                  { $string = (" " x $pad).$string; }
  }
  return $string;
}


1;

__END__


=head1 NAME

UCS::DS::Format - ASCII-format data set or subset

=head1 SYNOPSIS

  use UCS::DS::Memory;
  use UCS::DS::Format;

  $ds = new UCS::DS::Memory $filename;  # needs in-memory representation

  $formatter = new UCS::DS::Format $ds;	# formatter object for data set $ds

  $formatter->digits(6);		# number of significant digits 

  $formatter->mode("table");		# only mode so far

  $formatter->pagelength(50);		# print in pages of 50 rows each
  $formatter->pagelength(undef);	# print as single table

  $formatter->vars($pattern, ...);	# select variables that will be shown

  $formatter->print;			# print formatted table on STDOUT
  $formatter->print($filename);         # write to file or pipe


=head1 DESCRIPTION

This module provides a convenient method to format data sets as ASCII
tables, which can then be used for viewing and printing.  The formatter
has to be applied to the in-memory representation implemented by the
B<UCS::DS::Memory> module.  Its output is printed on STDOUT by default,
but it can also be redirected to a file or pipe.


=head1 METHODS

=over 4

=item $formatter = new UCS::DS::Format $ds;

Creates new formatter object for the data set I<$ds>, which must be 
a B<UCS::DS::Memory> object.  The formatter object should be used
immediately after its creation and destroyed afterwards.  When any
changes are made in the data set I<$ds>, a new formatter has to be
created.

=item $formatter->digits($n);

Configure I<$formatter> to display approximately I<$n> significant
digits for floating-point variables (data type C<DOUBLE>).  I<$n>
must be at least 2.

=item $formatter->mode("table");

The default mode C<table> prints the data set in the form of a simple ASCII
table with column headers.  It is the only supported mode so far.

=item $formatter->pagelength($rows);

Configure I<$formatter> to format data set in separate pages of I<$n> rows
each.  The individual pages are separated by a single blank line.  Use of this
option may improve the formatting quality, helps to avoid excessive columns
widths, and reduces the delay before partial results can be displayed.

When I<$rows> is set to 0 or omitted, the entire data set is printed as 
a single table.  This is also the default behaviour.

=item $formatter->vars($pattern, ...);

Display only variables matching the specified wildcard patterns, in the
specified order.  This configuration option can also be used to change
the ordering of the columns or display a variable in more than one column. 
Repeated calls to the B<vars> method will overwrite, rather than add to, 
the previous selection.

=item $formatter->print;

=item $formatter->print($filename);

Format the data set with the specified options, and print the result on
STDOUT.  When the optional argument I<$filename> is specified, the output is
redirected to this file or pipe.

=back


=head1 SEE ALSO

See also the manpage of the L<PRINT|PRINT> utility, which is based on 
the B<UCS::DS::Format> module.


=head1 COPYRIGHT

Copyright 2004 Stefan Evert.

This software is provided AS IS and the author makes no warranty as to
its use and performance.  You may use the software, redistribute and
modify it under the same terms as Perl itself.

=cut



