## Filename: UCS/DS/Memory.pm
## Modified: Wed May 12 14:30:10 2004 (evert)   
##   Author: Stefan Evert
##  Purpose: in-memory representation of UCS data sets

package UCS::DS::Memory;

use locale;

use UCS;
use UCS::DS;
use UCS::DS::Stream;
use UCS::Expression;
use UCS::Expression::Func;

use base qw<UCS::DS>;

sub new {
  my $class = shift;
  my $filename = shift;				# may also be a UCS::DS::Stream::Read object
  my $self = new UCS::DS;
  $self->{SIZE} = 0;				# number of rows (actual length of arrays in DATA is ignored)
  $self->{DATA} = {};				# $self->{DATA}->{$var}->[$n] = $cell;
  $self->{INDEX} = {};				# $self->{INDEX}->{$name} = [$row1, $row2, ...]
  $self->{ACTIVE_INDEX} = undef;		# name of activated index
  bless($self, $class);				# enable overridden data set methods 
  if ($filename) {
    my $ds = (ref $filename eq "UCS::DS::Stream::Read") ? $filename : new UCS::DS::Stream::Read $filename;
    if (@_) {					# parse init options
      my %opts = map {lc($_) => 1} @_;
      $ds->na(0) if delete $opts{"-na"};	# -na  ...  disable missing values (-> 0 / "")
      $ds->na(1) if delete $opts{"+na"};	# +na  ...  enable missing values (-> undef)
      my @unknown = sort keys %opts;
      UCS::Die "UCS::DS::Memory:  unknown option(s) @unknown passed to constructor"
	  if @unknown > 0;
    }
    $self->copy_comments($ds);
    $self->copy_globals($ds);
    my @vars = $ds->vars;			# column order so we can store data by column later on
    $self->add_vars(@vars);
    my $ncol = @vars;
    my $data = $self->{DATA};
    while ($ds->read) {
      UCS::Die "UCS::DS::Memory:  format error in data set file '$filename'"
	  unless $ds->valid;
      my $columns = $ds->columns;
      foreach my $i (0 .. $ncol-1) {
	push @{$data->{$vars[$i]}}, $columns->[$i];
      }
    }
    $self->{SIZE} = $ds->row;
    $ds->close;
    # could do some error checks here re. number of rows, declared size and length of vectors in DATA
  }
  return $self;
}

# ======================================================================

# extend methods for adding and deleting variables so that they update the DATA hash
sub add_vars {
  my $self = shift;
  foreach my $var (@_) {
    my $aref = $self->{DATA}->{$var} = [];      # will behave as if filled with undef's
    $#{$aref} = $self->{SIZE} - 1;		# adjust to correct length (so eval method won't complain)
  }
  $self->SUPER::add_vars(@_);
}

sub delete_vars {
  my $self = shift;
  delete @{$self->{DATA}}{@_};
  $self->SUPER::delete_vars(@_);
}

sub size {					# returns _effective_ size of data set (= size of index when activated)
  my $self = shift;
  my $active = $self->{ACTIVE_INDEX};
  if ($active) {
    return scalar @{$self->{INDEX}->{$active}};
  }
  else {
    return $self->{SIZE};
 }
}

sub set_size {
  my ($self, $newsize) = shift;
  my $size = $self->{SIZE};
  $newsize = 0 if $newsize < 0;
  # adjust length of data vectors (so eval method doesn't complain, 
  # and to avoid stray entries when shortening and re-extending a data set)
  foreach my $vec (values %{$self->{DATA}}) {
    $#{$vec} = $newsize - 1;
  }
  if ($newsize < $size) {
    # adjust indices when the data set is truncated (which is an expensive operation)
    foreach my $index (values %{$self->{INDEX}}) { 
      $index = [ grep {$_ <= $newsize} @$index ];
    }
  }
  $self->{SIZE} = $newsize;
}

# ======================================================================

sub cell {
  my ($self, $var, $n) = @_;
  my $nmax = $self->size;
  UCS::Die "UCS::DS::Memory: row number $n out of range (1 .. $nmax) in cell() method."
      unless 1 <= $n and $n <= $nmax;
  my $active = $self->{ACTIVE_INDEX};
  $n = $self->{INDEX}->{$active}->[$n-1]	# apply activated index
    if $active;
  my $data = $self->{DATA}->{$var};
  UCS::Die "UCS::DS::Memory: variable '$var' does not exist in cell() method call."
      unless defined $data;
  return $data->[$n-1];
}

sub set_cell {
  my ($self, $var, $n, $val) = @_;
  my $nmax = $self->size;
  UCS::Die "UCS::DS::Memory: row number $n out of range (1 .. $nmax) in set_cell() method."
      unless 1 <= $n and $n <= $nmax;
  my $active = $self->{ACTIVE_INDEX};
  $n = $self->{INDEX}->{$active}->[$n-1]	# apply activated index
    if $active;
  my $data = $self->{DATA}->{$var};
  UCS::Die "UCS::DS::Memory: variable '$var' does not exist in set_cell() method call."
      unless defined $data;
  $data->[$n-1] = $val;
}

# ======================================================================

sub row {
  my ($self, $n) = @_;
  my $nmax = $self->size;
  UCS::Die "UCS::DS::Memory: row number $n out of range (1 .. $nmax) in row() method."
      unless 1 <= $n and $n <= $nmax;
  my $active = $self->{ACTIVE_INDEX};
  $n = $self->{INDEX}->{$active}->[$n-1]	# apply activated index
    if $active;
  my $rowdata = {};
  my $data = $self->{DATA};
  foreach my $v ($self->vars) {
    $rowdata->{$v} = $data->{$v}->[$n-1];
  }
  return $rowdata;
}

sub set_row {
  my $self = shift;
  my $n = shift;
  my $nmax = $self->size;
  UCS::Die "UCS::DS::Memory: row number $n out of range (1 .. $nmax) in row() method."
      unless 1 <= $n and $n <= $nmax;
  my $active = $self->{ACTIVE_INDEX};
  $n = $self->{INDEX}->{$active}->[$n-1]	# apply activated index
    if $active;
  my $arghash;
  if (@_ == 1) {
    $arghash = shift;
    UCS::Die 'UCS::DS::Memory Usage:  $ds->set_row($n, $arghash);'
	unless ref($arghash) eq "HASH";
  } 
  else {
    UCS::Die 'UCS::DS::Memory:  uneven number of arguments in $ds->set_row($n, $v1 => $val1, ...);'
	unless (@_ & 0x01) == 0;
    $arghash = {@_} ;
  }
  foreach my $v (keys %$arghash) {
    $coldata = $self->{DATA}->{$v};
    $coldata->[$n-1] = $arghash->{$v}
      if defined $coldata;			# non-existing variables are silently ignored
  }
}

sub append_row {
  my $self = shift;
  my $realsize = $self->{SIZE};	      # real size and effective size may be different when there is an active index
  my $eff_size = $self->size;
  $self->set_size($realsize + 1);     # indices are still valid, but a sorting index will be incomplete now
  my $active = $self->{ACTIVE_INDEX}; # if there is an active index, it will be extended to include the new row
  push @{$self->{INDEX}->{$active}}, $realsize + 1
    if $active;
  $self->set_row($eff_size + 1, @_);
}

sub delete_rows {
  my $self = shift;
  my $from = shift;
  my $to = (@_) ? shift : $from;
  my $nmax = $self->{SIZE};
  UCS::Die "UCS::DS::Memory: invalid range $from .. $to of row numbers in delete_rows() method (max: 1 .. $nmax)."
      unless 1 <= $from and $from <= $to and $to <= $nmax;
  my $length = $to - $from + 1;
  my $data = $self->{DATA};
  foreach my $vec (values %$data) {		# cut rows $from .. $to from each column vector
    splice @$vec, $from - 1, $length;
  }
  $self->set_size($nmax - $length);
  foreach my $index (values %{$self->{INDEX}}) { # adjust indices (which is an expensive operation)
    $index = [ grep {$_ < $from or $_ > $to} @$index ];
  }
}

# ======================================================================

sub column {
  my ($self, $var) = @_;
  my $data = $self->{DATA}->{$var};
  UCS::Die "UCS::DS::Memory: variable '$var' does not exist in column() method call."
      unless defined $data;
  return $data;
}

# ======================================================================

sub missing {
  my ($self, $exp) = @_;
  UCS::Die 'UCS::DS::Memory Usage:  @missing_vars = $ds->missing($exp);'
      unless defined $exp and ref($exp) eq "UCS::Expression";
  my $data = $self->{DATA};
  my @missing = grep { not defined $data->{$_} } $exp->needed;
  return @missing;
}

sub eval {
  my ($self, $var, $exp) = @_;
  UCS::Die 'UCS::DS::Memory Usage:  $ds->eval($var, $exp);'
      unless defined $exp and (not ref($exp) or ref($exp) eq "UCS::Expression");
  $exp = new UCS::Expression $exp		# automatically compiles UCS expression specified as source string
    unless ref($exp);
  my @missing = $self->missing($exp);
  UCS::Die "UCS::DS::Memory:  can't eval() UCS expression because of missing variable(s) ".join(", ", @missing)
      if @missing;
  $self->add_vars($var)				# automatically create variable if necessary
    unless $self->var($var);
  $exp->evalloop($self->{DATA}->{$var}, $self->{SIZE}, $self->{DATA});
}

sub add {
  my $self = shift;
  my $var = shift;
  my ($key) = grep { $var eq "am.$_" } UCS::AM_Keys();
  if (defined $key) {		# requested variable is registered AM
    my $exp = UCS::AM_Expression($key);
    $self->eval($var, $exp);
  }
  elsif (grep {$var eq $_} @UCS::DerivedVars) {	# requested variable is derived variable
    my $exp = new UCS::Expression "\%$var\%";
    $self->eval($var, $exp);
  }
  else {
    UCS::Die "UCS::DS::Memory:  can't add() variable '$var' (not a registered AM or derived variable)";
  }
}

# ======================================================================

sub summary {
  my $self = shift;
  my $var = shift;
  my $type = UCS::VarType($var);
  UCS::Die "UCS::DS::Memory:  summary() available for INT and DOUBLE variables only (not for $var)"
      unless $type eq "INT" or $type eq "DOUBLE";
  UCS::Die "UCS::DS::Memory:  variable '$var' does not exist in data set"
      unless $self->var($var);
  $vector = $self->{DATA}->{$var};		# column vector for this variable
  # count undefined values and remove them for the summary (by constructing appropriate index)
  my $active = $self->{ACTIVE_INDEX};		
  my $na = 0;
  $self->make_index(".summary.ok", grep {	# .summary.ok = index of defined entries
    if (defined $vector->[$_-1]) { 1 } else { $na++; 0 }
  } (($active) ? @{$self->{INDEX}->{$active}} : 1 .. @$vector) );
  $self->activate_index(".summary.ok");		# activate .summary.ok for sort operation
  my $index = $self->sort(".summary", "+$var");	
  $self->activate_index($active);		# return to previously activated index
  # now compute statistics for the remaining entries (by looping over $index)
  my $N = @$index;
  return { NA => $na }				# can't compute statistics -> MEAN, ... are undefined
    if $N == 0;
  my $min = $vector->[$index->[0] - 1];
  my $max = $vector->[$index->[$N-1] - 1];
  my $median = 0;
  my $N2 = int($N / 2);				# midpoint of vector -> median
  if (2 * $N2 == $N) {
    $median =					# even length -> median is average of element nos. $N2 and $N2+1
      ($vector->[$index->[$N2] - 1] + $vector->[$index->[$N2-1] - 1]) / 2;
  }
  else {
    $median = $vector->[$index->[$N2] - 1];	# odd length -> median is element no. $N2+1
  }
  my $absmax = max(abs($min), abs($max));
  my $absmin = abs($min);			# $absmin, $sum, $ss, $step are computed in loop below
  my $sum = 0;					# sum of entries (-> mean)
  my $ss = 0;					# sum of squares (-> variance)
  my $step = $max - $min;			# granularity = minimum step size between consecutive values
  for my $i (0 .. $N-1) {
    my $x = $vector->[$index->[$i] - 1];
    $sum += $x;
    $ss  += $x * $x;
    if ($i > 0) {
      $abs_x = abs($x);
      $absmin = $abs_x 
	if $abs_x < $absmin and $abs_x > 0;
      $x_step = $x - $vector->[$index->[$i-1] - 1]; # non-negative step because vector has been sorted
      $step = $x_step
	if $x_step < $step and $x_step > 0;
    }
  }
  my $mean = $sum / $N;
  my $variance = undef;
  my $sd = undef;
  if ($N >= 2) {				# need at least two values to compute variance and s.d.
    $variance = ($ss - $N * $mean * $mean) / ($N - 1);
    $variance = 0 if $variance < 0;		# possible cancellation error for large mean and small variance
    $sd = sqrt($variance);
  }
  $self->delete_index(".summary.ok");		# delete temporary indices
  $self->delete_index(".summary");
  return {					# return result structure
	  MIN => $min,				# minimum value
	  MAX => $max,				# maximum value
	  ABSMIN => $absmin,			# minimum non-zero absolute value
	  ABSMAX => $absmax,			# maximum absolute value
	  SUM => $sum,				# sum of all values
	  MEAN => $mean,			# mean
	  MEDIAN => $median,			# median
	  VAR => $variance,			# variance
	  SD => $sd,				# standard deviation
	  STEP => $step,			# minimum distance between two values
	  NA => $na,				# number of undefined entries (if > 0, results may be incorrect)
	 };
}

# ======================================================================

sub index {
  my $self = shift;
  my $name = shift;
  UCS::Die 'UCS::DS::Memory Usage:  $vector = $ds->index($idx);'
      if @_ > 0;
  return $self->{INDEX}->{$name};		# undef if not defined
}

sub make_index {
  my $self = shift;
  my $name = shift;
  my $index = $self->{INDEX};
  if (@_) {
    my $data = shift;
    if (@_ == 0 and ref($data) eq "ARRAY") {
      $index->{$name} = [ @$data ];		# make copy of arrayref
    }
    else {
      $index->{$name} = [ $data, @_ ];		# assume that arguments are row numbers
    }
  }
  else {
    $index->{$name} = [];			# empty indices are also allowed
  }
}

sub where {
  my ($self, $name, $exp) = @_;
  UCS::Die 'UCS::DS::Memory Usage:  $ds->where($idx, $exp);'
      unless defined $exp and (not ref($exp) or ref($exp) eq "UCS::Expression");
  $exp = new UCS::Expression $exp		# automatically compiles UCS expression specified as source string
    unless ref($exp);
  my @missing = $self->missing($exp);
  UCS::Die "UCS::DS::Memory:  can't eval() UCS expression because of missing variable(s) ".join(", ", @missing)
      if @missing;
  $self->{INDEX}->{$name} = [];			# create new index or reset existing one
  my $index = $self->{INDEX}->{$name};
  my $selected = [];				# output vector for evalloop (will be interpreted as Boolean vector)
  my $size = $self->{SIZE};
  $exp->evalloop($selected, $size, $self->{DATA});
  my $active = $self->{ACTIVE_INDEX};		# if index is activated, where() computes a subset of this index
  if ($active) {
    foreach my $row (@{$self->{INDEX}->{$active}}) {
      push @$index, $row
	if $selected->[$row-1];
    }
  }
  else {					# otherwise, the new index contains all rows where $exp evaluates to true
    foreach my $row (1 .. $size) {
      push @$index, $row
	if $selected->[$row-1];
    }
  }
  undef($selected);
  return $index;				# also returns newly created index for direct use
}

sub count {
  my ($self, $exp) = @_;
  UCS::Die 'UCS::DS::Memory Usage:  $n = $ds->count($exp);'
      unless defined $exp and (not ref($exp) or ref($exp) eq "UCS::Expression");
  $self->where(".count", $exp);			# create temporary index ...
  my $n = scalar @{$self->index(".count")};	# ... and return its length
  $self->delete_index(".count");
  return $n;
}

sub activate_index {
  my $self = shift;
  my $name = shift;
  if (defined $name) {
    UCS::Die "UCS::DS::Memory:  index '$name' does not exist, activation failed."
	unless exists $self->{INDEX}->{$name};
    $self->{ACTIVE_INDEX} = $name;
  }
  else {
    $self->{ACTIVE_INDEX} = undef;
  }
}

sub delete_index {
  my $self = shift;
  my $active = $self->{ACTIVE_INDEX};
  $active = "" unless defined $active;
  foreach my $name (@_) {
    delete $self->{INDEX}->{$name};
    $self->activate_index()			# if one of the deleted indices was active, de-activate it
      if $name eq $active;
  }
}

# ======================================================================

sub copy {
  my $self = shift;
  my $active = $self->{ACTIVE_INDEX};
  my $copy = new UCS::DS::Memory;		# initialise a copy of $self (without the data)
  my @vars = $self->vars;
  if (@_) {					# copy selected variables only
    foreach my $spec (@_) {
      $copy->add_vars(UCS::Match($spec, @vars)); # the add method avoids duplicates
    }
  } 
  else {					# copy all variables
    $copy->add_vars(@vars);
  }
  $copy->copy_comments($self);
  $copy->copy_globals($self);
  foreach my $name ($self->vars) {		# remove globals explaining variables that will not be copied
    $copy->delete_global($name)
      if defined $copy->global($name) and not $copy->var($name);
  }
  # note that indices aren't preserved in the copy
  if ($active) {
    my $index = $self->index($active);
    UCS::Die "UCS::DS::Memory:  index '$active' does not exist, can't duplicate data set."
	unless defined $index;
    $copy->{SIZE} = scalar @$index;		# size of new data set == length of index
    my @index = map { $_ - 1 } @$index;		# shucks! (UCS indices start from 1)
    foreach my $var ($copy->vars) {
      $copy->{DATA}->{$var} = [ @{$self->{DATA}->{$var}}[@index] ]; # make copies using array slices
    }
  }
  else {
    $copy->{SIZE} = $self->{SIZE};
    foreach my $var ($copy->vars) {
      $copy->{DATA}->{$var} = [ @{$self->{DATA}->{$var}} ]; # make actual copies
    }
  }
  return $copy;
}

# ======================================================================

sub renumber {
  $self->add_vars("id")
    unless defined $self->var("id");
  my $vector = $ds->column("id");
  my $size = $self->{SIZE};			# ignore activated index
  foreach my $n (1 .. $size) {
    $vector->[$n-1] = $n;
  }
}

# ======================================================================

# internal helper function: parse sort key, returning variable, comparison operators, and flag for descending order
# ($var, $op_cmp, $op_eq, $descending) = parse_sort_key($key);
sub parse_sort_key {
  my $key = shift;
  my $descending = undef;			# this will trigger guessing unless order is specified explicitly 
  if ($key =~ s/^[+-]|[+-]$//) {
    $descending = ($& eq "-") ? 1 : 0;
  }
  my $type = UCS::VarType($key);		# determine data type
  UCS::Die "UCS::DS::Memory:  invalid sort key '$key'."
      unless $type;				# $type == "" indicates invalid variable name
  my $op_cmp = ($type eq "STRING") ? "cmp" : "<=>";
  my $op_eq = ($type eq "STRING") ? "eq" : "==";
  unless (defined $descending) {
    my ($spec) = UCS::SplitName($key);		# determine default sort order from variable type
    $descending = (defined($spec) and ($spec eq "b" or $spec eq "am")) ? 1 : 0;
    # Boolean variables and association scores are sorted in descending order by default, everything else ascending
  }
  return ($key, $op_cmp, $op_eq, $descending);
}

# internal function for sorting and ranking (because the two operations are quite similar)
# $ds->sort_or_rank($sort_index, $ranking_var, @sort_keys);
sub sort_or_rank {
  my $self = shift;
  my $idx = shift;				# name of index for sort results (undef for ranking op.)
  my $ranking = shift;				# name of variable for ranking (undef for sort op.)
  my @sort = ();				# comparison statements for each sort key
  my @equal = ();				# equality tests for each sort key (for ranking)
  my @vars = ();				# definitions of lexical variables
  # the Perl code generated below depends on several lexical variables in its environment:
  #   $self    = UCS::DS::Memory object, for access to column vectors (and direct output of ranking)
  #   $N       = size of data set (number of rows), needed for ranking loop
  #   $rows    = array of row numbers to be sorted (either active index or 1 .. $N)
  #   $sorted  = result of sort operation (i.e. an array of row numbers), which is used by the ranking loop
  #   $ranking = name of UCS variable to hold computed ranking
  # the code also defines several local variables for the relevant column vectors
  while (@_) {					# parse sort keys and generate Perl code for each key
    my ($var, $op_cmp, $op_eq, $descending) = parse_sort_key(shift);
    my $vector = $var;				# $vector = suitable name for local Perl variable 
    $vector =~ tr[.][_];
    $vector = '$_'.$vector;
    my $stmt = ($descending) ? "$vector".'->[$b-1] '."$op_cmp $vector".'->[$a-1]' : "$vector".'->[$a-1] '."$op_cmp $vector".'->[$b-1]';
    push @sort, $stmt;
    push @equal, "$vector".'->[$sorted->[$_line-1] - 1] '."$op_eq $vector".'->[$sorted->[$_line-2] - 1]';
    push @vars, 'my '."$vector".' = $self->{DATA}->{'."'$var'".'};';
    UCS::Die "UCS::DS::Memory:  missing variable '$var' for sort or rank operation"
	unless $self->var($var);
  }
  my $sort_code = join("\n", '{',		# this code sorts row numbers in $rows according to the specified sort keys
		       @vars, 
		       'sort {'.join(" or ", @sort).'} @$rows',
		       '}');  
  my $N = $self->{SIZE};			# $N is the actual size of the data set (not used for sort op.)
  my $active = (defined $ranking) ? undef : $self->{ACTIVE_INDEX}; # ignore active index for ranking operation
  my $rows = ($active) ? $self->index($active) : [1 .. $N];
  my $sorted = [ eval $sort_code ];		# execute code for sort op.
  UCS::Die "UCS::DS::Memory:  internal error in sort() method", "Evaluation of generated code", $sort_code, "failed because of", $@
      if $@;					# check that eval was successful
  $self->{INDEX}->{$idx} = $sorted		# store sort result (for sort op.)
    if defined $idx;
  if (defined $ranking) {
    my $rank_code = join("\n", '{',		# this code computes a ranking corresponding to $sorted, assigning equal ranks to ties
			 @vars, 
			 'my $_rank = 0;', 
			 'my $_result = $self->{DATA}->{'."'$ranking'".'};',
			 'foreach my $_line (1 .. $N) {',
			 '  $_rank = $_line if $_line == 1 or not ('.join(" and ", @equal).');',
			 '  $_result->[$sorted->[$_line-1] - 1] = $_rank;',
			 '}',
			 '}');
    eval $rank_code;                            # execute code for ranking op. (writes directly to data set variable)
    UCS::Die "UCS::DS::Memory:  internal error in rank() method", "Evaluation of generated code", $rank_code, "failed because of", $@
	if $@;					# check that eval was succesconveniencesful
  }
  return $sorted;				# return newly created sort index (used by sort() method)
}

sub sort {
  my $self = shift;
  my $idx = shift;
  UCS::Die 'UCS::DS::Memory Usage:  $ds->sort($idx, $key1, $key2, ...);'
      unless @_;
  return $self->sort_or_rank($idx, undef, @_);	# also returns sort index as arrayref for direct use
}

sub rank {
  my $self = shift;
  my $ranking = shift;
  UCS::Die 'UCS::DS::Memory Usage:  $ds->rank($ranking, $key1, $key2, ...);'
      unless @_;
  UCS::Die "UCS::DS::Memory:  illegal variable name '$ranking' in rank() method."
      unless UCS::ValidName($ranking);
  my ($type) = UCS::SplitName($ranking);
  UCS::Die "UCS::DS::Memory:  output variable ($ranking) must have the form r.* or n.* in rank() method"
      unless $type eq "r" or $type eq "n";
  $self->add_vars($ranking);			# make sure that result variable exists
  $self->sort_or_rank(undef, $ranking, @_);
}

# ======================================================================

sub save {
  my $self = shift;
  my $filename = shift;
  UCS::Die 'UCS::DS::Memory Usage:  $ds->save($filename [, @variables]);'
      unless defined($filename);
  my $active = $self->{ACTIVE_INDEX};
  my $size = $self->size;			# effective size, taking active index into account
  my $ds = new UCS::DS::Stream::Write $filename; # open output stream and configure it
  my @vars = grep {not $self->temporary($_)} $self->vars; # don't save temporary variables
  if (@_) {
    foreach my $spec (@_) {			# save selected variables only
      $ds->add_vars(UCS::Match($spec, @vars));  # the add method avoids duplicates
    }
  } 
  else {					# save all variables
    $ds->add_vars(@vars);
  }
  $ds->copy_globals($self);
  foreach my $name ($self->vars) {		# remove globals explaining variables that will not be saved
    $ds->delete_global($name)
      if defined $ds->global($name) and not $ds->var($name);
  }
  $ds->set_global("size", $size);
  $ds->copy_comments($self);
  $ds->open;					# write data set header
  foreach my $n (1 .. $size) {
    my $rowdata = $self->row($n);		# uses index if activated
    $ds->data($rowdata); # not quite the fastest solution - may need to think of an alternative
    $ds->write;
  }
  $ds->close;
}

# ======================================================================

package UCS::DS::Memory::Dict;

sub new {
  my $class = shift;
  my $ds = shift;
  my @keys = @_;				# remaining arguments are variables
  UCS::Die 'UCS::DS::Memory::Dict Usage:  new UCS::DS::Memory::Dict $ds, $var1, $var2, ...;'
      unless @keys > 0 and ref($ds) eq "UCS::DS::Memory";
  my $n_keys = @keys;
  my @missing = grep { not $ds->var($_) } @keys;
  UCS::Die "UCS::DS::Memory::Dict:  variable(s) @missing not defined in data set"
      if @missing > 0;
  UCS::Die "UCS::DS::Memory::Dict:  variable(s) repeated in key specification (@keys)"
      unless UCS::Unique(@keys) == @keys;

  my $self = {
	      N_KEYS => $n_keys,                # number of variables in key (= values for lookup)
	      KEYS => [ @keys ],                # variable names of key items (in correct order)
	      ROWS => 0,			# number of rows indexed in dictionary
	      MAKE_KEY => undef,		# sub to construct hash key from variable values
	      HASH => {},			# lookup hash (stores single row as scalar, multiple rows as arrayref)
	     };

  my @convert = map {				# code to convert each key item to an appropriate string value
    my $type = $ds->var($keys[$_]);
    if ($type eq "BOOL")      { '(($_['.$_.']) ? "1" : "0")' }
    elsif ($type eq "STRING") { '($_['.$_.'])' }
    elsif ($type eq "INT")    { '(sprintf "%d", $_['.$_.']+0)' }
    else                      { '($_['.$_.']+0.0)' }
  } (0 .. $n_keys - 1);
  my $code = 'sub { return join("\t", '.join(", ", @convert).'); }';
  my $sub = eval $code;				# compile make_key subroutine
  UCS::Die "UCS::DS::Memory::Dict:  compilation of make_key sub failed because $@", "[$code]"
      unless defined $sub and ref($sub) eq "CODE";
  $self->{MAKE_KEY} = $sub;

  return bless($self, $class);			# actual initialisation is done by dict() method below
}

sub multiplicity {
  my $self = shift;
  if (@_) {					# called with arguments -> compute multiplicity of specific key
    my $data = $self->find_key("multiplicity", @_);
    return 0 
      unless defined $data;
    return 1
      unless ref($data);
    return scalar @$data;
  }
  else {					# without arguments -> compute max. and avg. multiplicity
    my $rows = $self->{ROWS};
    my $max = 1;
    my $avg = 1;
    if ($rows > 0) {
      my $hash = $self->{HASH};
      $avg = 0;
      my $entries = 0;
      foreach my $val (values %$hash) {
	my $mult = (ref($val) eq "ARRAY") ? @$val : 1; # number of rows for this key
	$max = $mult
	  if $mult > $max;
	$avg += $mult;
	$entries++;
      }
      UCS::Die "UCS::DS::Memory::Dict:  lookup hash corrupt ($avg != $rows)"
	  unless $avg == $rows;
      $avg /= $entries;
    }
    if (wantarray) {
      return ($max, $avg);
    } else {
      return $max;
    }
  }
}

sub unique {
  my $self = shift;
  return ($self->multiplicity > 1) ? 0 : 1;
}

# internal method: insert key (specified as list of variable values) into lookup hash
sub insert {
  my $self = shift;
  my $n_keys = $self->{N_KEYS};
  my $row = shift;
  my $args = @_;
  UCS::Die "UCS::DS::Memory::Dict:  insert() called with wrong number of arguments (expected $n_keys, got $args)"
      unless $n_keys == $args;
  my $key = $self->{MAKE_KEY}->(@_);
  my $hash = $self->{HASH};
  if (exists $hash->{$key}) {
    $hash->{$key} = [ $hash->{$key} ]
      unless ref($hash->{$key}) eq "ARRAY";
    push @{$hash->{$key}}, $row;
  }
  else {
    $hash->{$key} = $row;
  }
  $self->{ROWS}++;
}

# internal method: find dictionary key matching list of variable values or row of other data set
# (used by lookup() and multiplicity() methods so they can take arguments in the same way) 
sub find_key {
  my $self = shift;
  my $caller = shift;				# name of calling method (for error messages)
  my @keys = ();
  my $wanted = $self->{N_KEYS};
  if (@_ == 2 and ref($_[0]) eq "UCS::DS::Memory") {
    my ($ds, $row) = @_;
    foreach my $var (@{$self->{KEYS}}) {
      UCS::Die "UCS::DS::Memory::Dict:  variable $var not defined in data set passed to ${caller}() method"
	  unless $ds->var($var);
      push @keys, $ds->cell($var, $row);
    }
  }
  elsif (@_ > 0 and not grep {ref($_)} @_) {
    @keys = @_;
    my $args = @keys;
    if ($wanted > 1 and $args == 1) {		# so one can also use the internal key representations for lookup
      @keys = split /\t/, $keys[0];
      $args = @keys;
    }
    UCS::Die "UCS::DS::Memory::Dict:  ${caller}() called with wrong number of arguments (expected $wanted, got $args)"
	unless $args == $wanted;
  }
  else {
    UCS::Die 'UCS::DS::Memory::Dict Usage:  $dict->'.$caller.'($x1, $x2, ...); OR $dict->'.$caller.'($other_ds, $row_num);';
  }
  my $key = $self->{MAKE_KEY}->(@keys);
  return $self->{HASH}->{$key};	# may be undef, single number, or array reference (caller has to work this out
}

sub lookup {
  my $self = shift;
  my $data = $self->find_key("lookup", @_);
  return					# key not found -> return undef or empty list
    unless defined $data;
  return $data					# single row -> return this value
    unless ref($data);
  return (wantarray) ? @$data : $data->[0];	# multiple rows -> return list or first row
}

sub keys {
  my $self = shift;
  return keys %{$self->{HASH}};
}

# ======================================================================

package UCS::DS::Memory;

sub dict {
  my $self = shift;
  my @keys = @_;
  UCS::Die 'UCS::DS::Memory Usage:  $dict = $ds->dict($var1, $var2, ...);'
      unless @keys > 0;
  my $dict = new UCS::DS::Memory::Dict $self, @keys;
  foreach my $row (1 .. $self->size) {		# insert keys into $dict (uses activated index)
    $dict->insert($row, map { $self->cell($_, $row) } @keys);
  }
  return $dict;
}

# ======================================================================

1;

__END__


=head1 NAME

UCS::DS::Memory - In-memory representation of data sets


=head1 SYNOPSIS

  use UCS::DS::Memory;

  $ds = new UCS::DS::Memory;		# empty data set
  $ds = new UCS::DS::Memory $filename;	# read from file (using UCS::DS::Stream)

  # access & edit variables, comments, and globals with UCS::DS methods

  $pairs = $ds->size;			# number of pair types
  $ds->set_size($pairs);                # truncate or extend data set

  $value = $ds->cell($var, $n);		# read entry from data set table
  $ds->set_cell($var, $n, $value);	# set entry in data set table

  $rowdata = $ds->row($n);		# returns hashref (varname => value)
  $ds->set_row($n, $rowdata);		# set row data (ignores missing vars)
  $ds->set_row($n, "f1"=>$f1, "f2"=>$f2, ...);
  $ds->append_row($n, $rowdata);        # append row to data set
  $ds->delete_rows($from, $to);		# delete a range of rows from the data set

  $vector = $ds->column($var);		# reference to data vector of $var
  $vector->[$n] = $value;		# fast direct access to cells

  $ds->eval($var, $exp)  		# evaluate expression on data set & store in $var
    unless $ds->missing($exp);		# check first whether all reqd. variables are available
  $ds->add($var);			# auto-compute variable (derived variable or registered AM)

  $stats = $ds->summary($var);          # statistical summary of numerical variable

  $ds->where($idx, $exp);		# define index: rows matching UCS expression
  $n = $ds->count($exp);		# number of rows matching expression
  $vector = $ds->index($idx);		# returns reference to array of row numbers
  $ds->make_index($idx, $row1, $row2, ...);  # define index: explicit list of row numbers
  $ds->make_index($idx, $vector);	     #            or array reference (will be duplicated)
  $ds->activate_index($idx);		# activate index (will be used by most access methods)
  $ds->activate_index();		# de-activate index
  $ds->delete_index($idx);              # delete index

  $ds2 = $ds->copy;		        # make physical copy of data set (using index if activated)
  $ds2 = $ds->copy("*", "am.%");        # copy selected variables only (in specified order)

  $ds->renumber;                        # renumber/add ID values as increasing sequence 1 .. size

  $ds->sort($idx, $var1, $var2, ...);   # sort data set on $var1, breaking ties by $var2 etc.
  $ds->sort($idx, "-$var1", "+$var2");	# - = descending, + = ascending (default depends on variable type)
  $ds->rank($ranking, $key1, ...);      # compute ranking (with ties) and store in data set variable $ranking

  $ds->save($filename);			# save data set to file (using index if activated)

  $dict = $ds->dict($var1, $var2, ...); # lookup hash for variable(s) (UCS::DS::Memory::Dict object)
  ($max, $average) = $dict->multiplicity; # maximum / average number of rows for each key
  if ($dict->unique) { ... }            # whether every key identifies a unique row
  @rows = $dict->lookup($x1, $x2, ...); # look up key in dictionary, returns all matching rows
  $row  = $dict->lookup($x1, $x2, ...); # in scalar context, returns first matching row
  @rows = $dict->lookup($other_ds, $n); # look up row $n from other data set
  $n_rows = $dict->multiplicity($x1, $x2, ...);  # takes same arguments as lookup()
  @keys = $dict->keys;			# return unsorted list of keys entered in dictionary

=head1 DESCRIPTION

This module implements an in-memory representation of UCS data sets.  When a
data set file has been loaded into a B<UCS::DS::Memory> object (or a new empty
data set has been created), then B<variable names>, B<comments>, and
B<globals> can be accessed and modified with the respective B<UCS::DS> methods
(see L<the UCS::DS manpage|UCS::DS>).

Additional methods in the B<UCS::DS::Memory> class allow the user to:

=over 4

=item *

read and write individual B<cells> as well as entire B<rows> or B<columns>

=item *

change the B<size> of a data set

=item *

annotate B<derived variables>, B<association scores>, or arbitrary B<UCS
expressions> in the data set

=item *

compute statistical B<summaries> of numerical variables

=item *

B<select> rows matching given UCS expression from a data set

=item *

B<sort> data sets by one or more variables and compute B<rankings>

=item *

save the data set into a data set file

=back

The individual methods are detailed in the following sections.  In all methods,
columns are identified by the respective variable names, whereas rows (corresponding
to pair types) are identified by row numbers.  B<NB:> Row numbers start with 1
(like R vectors, but unlike Perl arrays)!


=head1 GENERAL METHODS

=over 4

=item $ds = new UCS::DS::Memory;

Create empty data set.  The new data set has zero rows and no variables.
Returns object of class B<UCS::DS::Memory>;

=item $ds = new UCS::DS::Memory $file [, '-na'];

Reads data set file into memory and returns B<UCS::DS::Memory> object.  The
argument I<$file> is either a string giving the name of the data set file or a
B<UCS::DS::Stream::Read> object (see L<the UCS::DS::Stream
manpage|UCS::DS::Stream>), which has been opened but not read from.  When
the specified file does not exist and in the case of a read error, the 
constructor B<die>s with an appropriate error message. 

The option C<'-na'> disables missing value support (which is enabled by
default), so that C<NA> values in the data set file will be replaced by 0 or
the empty string, depending on the data type.  Use C<'+na'> to enable missing
value support explicitly.

=item $V = $ds->size;

Returns the size of the data set, i.e. the number of rows (or pair types).

=item $ds->set_size($V);

Change the size of the data set to I<$V> rows.  This method can both truncate
and extend a data set.  B<NB:> Unlike the B<size> method, B<set_size> always
applies to the real size of the data set and ignores the active row index.
However, all row indices are preserved and adjusted in case of a truncation.
If there is an active row index, it remains active. (See the section L</"ROW
INDEX METHODS"> below for more information on row indices.)

=item $value = $ds->cell($var, $n);

Retrieve the value of variable I<$var> for row I<$n> (i.e. the I<$n>-th pair type).
This method is convenient and performs various error checks, but it involves 
a considerable amount of overhead.  Consider the B<column> method when performance
is an issue. 

=item $ds->set_cell($var, $n, $value);

Set the value of variable I<$var> for row I<$n> to I<$value>.  Like B<cell>, this
method is convenient, but comparatively slow.  Consider the B<column> method when
is an issue. 

=item $rowdata = $ds->row($n);

Returns hash reference containing the entire data from row I<$n> indexed by
variable names.  This method is inefficient and mainly for convenience, e.g.
when applying a UCS expression to individual rows (cf. the description of the
B<eval> method in L<the UCS::Expression manpage|UCS::Expression>).  

=item $ds->set_row($n, $rowdata);

=item $ds->set_row($n, $var1 => $val1, $var2 => $val2, ...);

Set the values of some or all variables for row I<$n>.  The values can either
be passed in a single hash reference indexed by variable names, or as I<$var>
C<< => >> I<$value> pairs.  Any variables that do not exist in the data set
I<$ds> are silently ignored.  This method is faster than calling B<set_cell>
repeatedly, especially when a new row is added to the data set.  

=item $ds->append_row($rowdata);

=item $ds->append_row($var1 => $val1, $var2 => $val2, ...);

Append new row to the data set and fill it with the specified values.  This
method is a combination of B<set_size> and B<set_row>.  Variable values that
are not specified in the argument list are set to B<undef>.  When there is an
active row index, the new row is appended to this index, while all other
indices remain unchanged (see the section L</"ROW INDEX METHODS"> below for
more information on row indices).

=item $ds->delete_rows($from, $to);

Delete rows I<$from> through I<$to> from the data set.  B<NB:> This method
always applies to the real row numbers and ignores the active row index.  All
existing indices are adjusted (which is an expensive operation) and an active
row index remains activated.  (See the section L</"ROW INDEX METHODS"> below
for more information on row indices.)

=item $vector = $ds->column($var);

Returns an array reference to the data vector of variable I<$var>.  I<$vector>
can be used both for read and write access, so care has to be taken that the
data set isn't accidentally modified (e.g. through side effects of a B<map> or
B<grep> operation on I<@$vector>).  Of course, activating a row index has no
effect, since the B<column> method gives direct access to the internal data
structures.  (See the section L</"ROW INDEX METHODS"> below for more
information on row indices.)

=item @missing_vars = $ds->missing($exp);

Determines whether all variables required to evaluate the UCS expression
I<$exp> (an object of class B<UCS::Expression>) are defined in the data 
set I<$ds>.  Returns an empty list if I<$exp> can be evaluated, and
the names of missing variables otherwise.

=item $ds->eval($var, $exp);

Evaluate the UCS expression I<$exp> (an object of class B<UCS::Expression>) on
the data set I<$ds>, and store its values in the variable I<$var>.  When
I<$var> is a new variable, it is automatically added to the data set;
Otherwise, the previous values are overwritten.  This operation is I<much>
faster than repeatedly evaluating I<$exp> for each row.  For convenience,
I<$exp> can also be specified as a source string, which will be compiled on
the fly.  B<NB:> The B<eval> method always operates on the entire data set,
even when a row index is activated.  (See the section L</"ROW INDEX METHODS">
below for more information on row indices.)

=item $ds->add($var);

Add a new variable to the data set and auto-compute its values, or overwrite
an existing variable.  I<$var> must be the name of a B<derived variable> such
as C<E11> or an B<association score> such as C<am.t.score> (see L<the ucsfile
manpage|ucsfile> for details).

=item $stats = $ds->summary($var);

Computes a statistical summary of the numerical variable I<$var> (a numerical
variable is a variable of data type C<INT> or C<DOUBLE>).  I<$stats> is a 
hash reference representing a data structure with the following fields:

  MIN     ...  minimum value
  MAX     ...  maximum value
  ABSMIN  ...  smallest non-zero absolute value
  ABSMAX  ...  largest absolute value
  SUM     ...  sum of all value
  MEAN    ...  mean (= average) 
  MEDIAN  ...  median (= 50% quantile)
  VAR     ...  empirical variance
  SD      ...  empirical standard deviation (sq. root of variance)
  STEP    ...  smallest non-zero difference between any two values
  NA      ...  number of missing values (undef's)

Note that some of these fields may be B<undef> if they have no meaningful
value for the given data set.

=item $ds2 = $ds->copy;

=item $ds2 = $ds->copy(@variables);

Duplicates a data set, so that I<$ds2> is completely independent from I<$ds>
(whereas C<$ds2 = $ds;> would just give another handle on the same data set).
Comments and globals are copied to I<$ds2> as well.  Optionally, a list of
variable names and/or wildcard patterns (see L<the ucsexp manpage|ucsexp>) can
be specified.  In this case, only the selected columns will be copied.  B<NB:>
If there is an active row index, the copy will only include the rows selected
by the index, and they will be arranged in the corresponding order.  However,
no row indices are copied to I<$ds2>.  (See the section L</"ROW INDEX
METHODS"> below for more information on row indices.)

=item $ds->renumber;

When rows have been deleted from a data set, or a copy has been made with an
active row index, the values of the C<id> variable are preserved (and can
be used to match rows against the correspond entries in the original data set).
When an independent numbering is desired, the B<renumber> method can be used
to re-compute the C<id> values so that they form an uninterrupted sequence
starting from 1.  B<NB:> The renumbering ignores an activated row index.

=item $ds->save($filename);

=item $ds->save($filename, @variables);

This method saves the contents of I<$ds> to a UCS data set file I<$filename>.
When an optional list of variable names and/or wildcard patterns (see L<the
ucsexp manpage|ucsexp>) is specified, only the selected columns will be saved.
B<NB:> If there is an active row index, only the rows selected by the index
will be written to I<$filename>, and they will be arranged in the
corresponding order.  The row indices themselves cannot be stored in a data
set file.  (See the section L</"ROW INDEX METHODS"> below for more information
on row indices.)  Also note that B<temporary variables> will not be saved (see
L<the UCS::DS manpage|UCS::DS>).

=back


=head1 ROW INDEX METHODS

A B<row index> is an array reference containing a list of row numbers
(starting from 1, unlike Perl arrays).  Row indices are used to select rows
from an in-memory data set, or to represent a re-ordering of the rows (or
both).  They are usually created by the B<where> and B<sort> methods, but can
also be constructed explicitly.  An arbitrary number of named row indices can
be stored in a B<UCS::DS::Memory> object.

A row index can be B<activated>, creating a "virtual" data set containing only
the rows selected by the index, arranged in the corresponding order.  Most
B<UCS::DS::Memory> methods will then operate on this virtual data set.  All
exceptions are marked clearly in this manpage.  In particular, the B<where>
method selects a subset of the activated index, and B<sort> can be used to
reorder it.  There can only be one active row index at a time.  There is no
way of localising the activation (so that a previously active index is restored 
at the end of a block), so it is highly recommended to use active indices only
locally and de-activate them afterwards.  

Index names must be valid UCS identifiers, i.e. they may only contain
alphanumeric characters (C<A-Z a-z 0-9>) and periods (C<.>)
(cf. L<ucsfile/"VARIABLES">).  Note that index names beginning with a period
are reserved for internal use.

=over 4

=item $ds->make_index($idx, $row1, $row2, ...); 

=item $ds->make_index($idx, $vector);

Construct row index from a list of row numbers or an array reference I<$vector>,
and store it under the name I<$idx> in the data set I<$ds>.  In the second form,
the anonymous array is duplicated, so the contents of I<$vector> can be modified
or destroyed without affecting the stored row index.

=item $vector = $ds->index($idx);

Retrieve row index by name.  Returns an array reference to the internal data, so
be careful not to modify the contents of I<$vector> accidentally.  In most cases,
it is easier to activate I<$idx> and use the normal access methods. 

=item $ds->delete_index($idx);

Delete the row index named I<$idx>.  If it happens to be activated, it will 
automatically de-activated.

=item $ds->activate_index($idx);

Activate row index I<$idx>.  This will clear any previous activations.  Note that
this operation may change the effective size of the data set as returned by the 
B<size> method (unless I<$idx> is just a sort index).

=item $ds->activate_index();

Deactivate the currently active index, re-enabling direct access to the full
data set in its original order.

=item $ds->where($idx, $exp);

Construct I<$idx> selecting all rows for which the UCS expression I<$exp>
(given as a B<UCS::Expression> object) evaluates to true (see L<the ucsexp
manpage|ucsexp> for an introduction to UCS expression, and L<the
UCS::Expression manpage|UCS::Expression> for compilation instructions).  It is
often convenient to compile I<$exp> on the fly, especially when it is a
simple condition, e.g.

  $ds->where("high.freq", new UCS::Expression '%f% >= 10');

which can be shortened to

  $ds->where("high.freq", '%f% >= 10');

The B<where> method will automatically compile the source string passed as
I<$exp> into a B<UCS::Expression> object.  On-the-fly compilation involves
only moderate overhead.  When there is an active row index, B<where> will 
select a subset of this index, preserving its ordering.

=item $n = $ds->count($exp);

Similar to B<where>, this method only counts the number of rows matching the
UCS expression I<$exp>, without creating a named index.  The condition I<$exp>
may be given either as a B<UCS::Expression> object or as a source string,
which is compiled on the fly.  (Internally, the rows are collected in a
temporary index, which is automatically deleted when the method call returns.)

=item $ds->sort($idx, $key1, $key2, ...);

Sort data set I<$ds> by the specified sort keys.  The data set is first
sorted, by I<$key1>.  Ties are then broken by I<$key2>, any remaining ties by
I<$key3>, etc.  If there are any ties left when all sort keys have been used,
their ordering is undefined (and depends on the implementation of the B<sort>
function in Perl).  The resulting ordering is stored in a row index with
the name I<$idx>.  When there is an active row index, B<sort> will re-order
the rows selected by this index.

Each B<sort key> consists of a variable name, optionally preceded or followed
by a C<+> or C<-> character to select ascending or descending sort order,
respectively.  The default order is I<descending> for Boolean variables and
association scores, and I<ascending> for all other variables.  The sort keys
C<'l1'> and C<'l2'> sort in alphabetical order, while C<'f-'> puts the most
frequent pair types first.

In order to break remaining ties randomly, an appropriate additional sort
key has to be specified.  The usual choice are the association scores of
the B<random> measure (see L<the UCS::AM manpage|UCS::AM>).  It may be 
necessary to compute this measure first, which can be conveniently done
with the B<add> method, as shown in the example below.

  # order pair types by frequency (descending), breaking ties randomly
  if (not $ds->var("am.random")) {
    $ds->add("am.random");
    $ds->temporary("am.random", 1);  # temporary, don't save to disk
  }
  $ds->sort("by.freq", "f-", "am.random");

=item $ds->rank($ranking, $key1, $key2, ...);

The B<rank> method is similar to B<sort>, but creates a ranking instead of a
sort index.  The ranking is stored in the integer variable I<$ranking>.  Note
that tied rows are assigned the same rank, which is the lowest available rank
(as in the Olympic Games) rather than the average of all ranks in the group
(as is often done in statistics).  All other remarks about the B<sort> method
apply equally well to the B<rank> method, especially those concerning
randomisation.

=back


=head1 DICTIONARIES (LOOKUP HASHES)

A data set B<dictionary> is a B<hash> structure listing all the different
values that a given variable assumes in the data set (or all the different
value combinations of several variables).  For each value (or value
combination), which is called a B<key> of the dictionary, the corresponding
row numbers in the data set can be retrieved (called a B<lookup> of the key). 
In the terminology of relational databases, such a dictionary is referred to
as an B<index>.  Be careful not to confuse this notion with the B<row index>
described above, which is used for subsetting and/or reordering the rows of
a data set.

A dictionary can be created for any variable (or combination of variables)
with the B<dict> method, and is returned in the form of a
B<UCS::DS::Memory::Dict> object.  B<NB:> This dictionary is only valid as long
as the data set itself is not modified (which includes activation or
deactivation of a row index).  Unlike a database index, the dictionary is not
updated automatically.  It is therefore important to keep operations on the
data set under strict control while a dictionary is in use.  It is always
possible to add, modify, and delete variables that are not included in the
dictionary, though.  For the same reason (as well as to save working memory),
dictionaries should be deleted when they are no longer needed.

The main purpose of a dictionary is to B<look up> keys and find the matching
rows in the data set efficiently (the B<ucs-join> program is an example of a
typical application).  It is often desirable to choose variables in such a way
that every key identifies a unique row in the data set (for instance, the
values of C<l1> and C<l2> identify a pair type, which should have only one
entry in a data set).  A dictionary with this property is called B<unique>.
Both unique and non-unique dictionaries are supported (unique dictionaries are
represented in a memory-efficient fashion).  Lookup and similar operations are
implemented as methods of the B<UCS::DS::Memory::Dict> object.

Although mainly intended for string values, dictionaries support all data
types.  Boolean variables will usually be of interest only in combination with
other variables (possibly also Boolean ones), and dictionaries are rarely
useful for floating-point values.

=over 4

=item $dict = $ds->dict($var1, ..., $varN);

Create a B<dictionary> for the variables I<$var1>, ..., I<$varN> in the data
set I<$ds>.  Each B<key> of this dictionary is a combination of I<N> values,
which must be specified in the same order as the variable names.  When a row
index is in effect, keys and row numbers in the dictionary are taken from the
virtual data set defined by the activated index.  The returned object of class
B<UCS::DS::Memory::Dict> is a read-only dictionary: in order to take changes
in the data set I<$ds> into account (including the activation or deactivation
of a row index), a new object has to be created with the B<dict> method.

=item if ($dict->unique) { ... }

This method returns a true value iff I<$dict> is a B<unique> dictionary.

=item ($max, $avg) = $dict->multiplicity;

=item $max = $dict->multiplicity;

Returns the maximum (I<$max>) and average (I<$avg>) number of rows matching a
key in I<$dict>.  The dictionary is unique iff I<$max> equals 1.

=item @rows = $dict->lookup($x1, ..., $xN);

=item $row  = $dict->lookup($x1, ..., $xN);

Look up a B<key>, specified as an I<N>-tuple of variable values I<($x1, ...,
$xN)>, in the dictionary I<$dict> and return the matching row numbers.  The
values I<$x1>, ..., I<$xN> must be given in the same order as the variables
I<$var1>, ..., I<$varN> in the B<dict> method call when the dictionary was
created.  When the key is not found in I<$dict>, an empty list is returned.

In scalar context, the (number of the) first matching row is returned, or
B<undef> if the key is not found in the dictionary.

=item @rows = $dict->lookup($ds2, $n);

=item $row  = $dict->lookup($ds2, $n);

The B<lookup> method can also be used to look up rows from a second data set
I<$ds2>, i.e. to find rows in the dictionary's data set I<$ds> where the
values of I<$var1>, ..., I<$varN> match the I<$n>-th row of I<$ds2>.  For this
form of invocation, the dictionary variables must be defined in I<$ds2>
(otherwise, a fatal error is raised).

=item $n_rows = $dict->multiplicity($x1, ..., $xN);

=item $n_rows = $dict->multiplicity($ds2, $n);

When called with arguments, the B<multiplicity> method returns the number of
rows matching a specific key in I<$dict>.  The key can be given in the same
two ways as for the B<lookup> method.  (Note that calling B<lookup> in scalar
context returns the first matching row, I<not> the total number of rows.)


=item @keys = $dict->keys;

=item $n_keys = $dict->keys;

Returns an unsorted list of all dictionary keys in the internal representation
(where each key is a single string value).  Such internal representations can
be passed to the B<lookup> and B<multiplicity> methods instead of an
I<N>-tuple I<($x1, ..., $xN)>.  In scalar context, the B<keys> method
efficiently computes the number of keys in I<$dict>.

=back


=head2 Examples

The B<keys> method and the ability to use the returned internal
representations in the B<lookup> method provide an easy way to compute the
(empirical) B<distribution> of a data set variable, i.e. a list of different
values and their multiplicities.  (Note that calling B<lookup> in scalar
context cannot be used to determine the multiplicity of a key because it 
returns the first matching row in this case.)

  # frequency table for variable $v on data set $ds
  $dict = $ds->dict($v);
  @distribution = 
    # sort values by multiplicity
    sort { $b->[1] <=> $a->[1] or $a->[0] cmp $b->[0] }
    # compute multiplicity for each value
    map { [$_, $dict->multiplicity($_)] }
    # for a single variable $v, internal keys are simply the values
    $dict->keys;
  undef $dict;  		# always erase dictionary after use

The following example is a bare-bones version of the B<ucs-join> command,
annotating the pair types of a data set I<$ds1> with a variable I<$var> from
another data set I<$ds2> (matching rows according to the pair types they
represent, i.e. using the variables C<l1> and C<l2>).  Typically, I<$ds2> will
be an annotation database.

  $ds1->add_variables($var);	# assuming $var hasn't previously exist in $ds1
  $dict = $ds2->dict($var);
  $dict->unique 
    or die "Not unique -- can't look up pair types.";
  foreach $n (1 .. $ds1->size) {
    $row = $dict->lookup($ds1, $n);
    $ds1->set_cell($var, $n, $ds2->cell($var, $row))
      if defined $row;
  }
  undef $dict;


=head1 SEE ALSO

The L<ucsfile|ucsfile> manpage for general information about UCS data sets and
the data set file format, the L<ucsexp|ucsexp> manpage for an introduction to
UCS expressions (which are used extensively in the B<UCS::DS::Memory> module)
and wildcard patterns, the L<UCS::Expression|UCS::Expression> manpage for
information on how to compile UCS expressions, and the L<UCS::DS|UCS::DS>
manpage for methods that manipulate the layout of a data set and its header
information.


=head1 COPYRIGHT

Copyright 2004 Stefan Evert.

This software is provided AS IS and the author makes no warranty as to
its use and performance.  You may use the software, redistribute and
modify it under the same terms as Perl itself.

=cut

