## Filename: UCS/DS/Stream.pm
## Modified: Tue Apr 27 10:19:55 2004 (evert)   
##   Author: Stefan Evert
##  Purpose: stream objects for reading/writing UCS data set files

package UCS::DS::Stream;

use UCS;
use UCS::DS;
use UCS::File;

# Read and write streams share so few features (other than those
# derived from UCS::DS) that they are implemented as two completely
# separate classes.

package UCS::DS::Stream::Read;

use base qw<UCS::DS>;
use File::Find;

sub new {
  my $class = shift;
  my $filename = shift;
  my $self = new UCS::DS;
  # try to locate data set file in the global DataSet/ directory tree, 
  # unless it's a fully specified relative or absolute path (^\.*/) or a command pipe (\|)
  if (not -f $filename and $filename !~ m{ ^\.*/ | \| }x) {
    my @matches = ();				# collect matching data set files in global tree
    my $regex = qr{/\Q$filename\E$};		# make sure that partial file names don't match!
    find(sub {
	   if ($File::Find::name =~ $regex) {
	     push @matches, $File::Find::name;
	   }
	 }, "$UCS::BaseDir/DataSet/.");
    UCS::Warn("Found multiple matches for data set file '$filename' in global directory tree:",
	      (map { "  $_" } @matches), "Using first matching file.")
	if @matches > 1;
    $filename = shift @matches
      if @matches;
  }
  my $fh = UCS::File::Open($filename);
  $self->{FILENAME} = $filename;		# name of the data set file
  $self->{FH} = $fh;				# file handle for reading
  $self->{WIDTH} = 0;				# number of columns
  $self->{LINE} = "";				# current line read from stream
  $self->{COLUMNS} = [];			# data from current row, split into columns
  $self->{ROW} = 0;				# row number (1 .. size)
  $self->{EOF} = 0;				# stop reading at EOF
  $self->{VALID} = 0;				# whether current data line is valid (format check)
  $self->{NA} = 1;				# support missing (NA) values as undef's, or replace with 0 / ""
  $self->{IS_NUMERIC} = [];			# flags indicating whether i-th field is numeric (for NA substitution)
  # read data set header: comment lines and global variables
  my $line = $fh->getline;
  while ($line and $line =~ /^\s*\#/) {
    chomp $line;
    if ($line =~ s/^\s*\#\#\:\:\s*//) {	     	# variable definition
      if (not $line =~ /^([A-Za-z0-9\.]+)\s*=\s*(.+)$/) {
	UCS::Warn "$filename: syntax error in definition of global variable:",
	    ">>##:: $line<<";
	return undef;
      }
      my ($var, $value) = ($1, $2);
      if (not UCS::ValidKey($var)) {
	UCS::Warn "$filename: illegal name '$var' for global variable:",
	    " >>##:: $line<<";
	return undef;
      }
      $self->set_global($var, $value);
    }
    else {					# comment line
      $line =~ s/\s*\#\s?//;			# strip leading '#' and one blank (further blanks allowed for indentation)
      $self->add_comments($line);
    }
    $line = $fh->getline;
  }
  # now $line should contain the table header
  UCS::Die "UCS::DS::Stream::Read:  error reading data set file '$filename'"
      unless $line;
  chomp $line;
  my @vars = split /\t/, $line;
  my $N = @vars;
  unless ($N > 0) {
    UCS::Warn "$filename: missing table header",
	"Perhaps a malformed comment line or missing data table??",
	  ">>$line<<";
    return undef;
  }
  $self->{WIDTH} = $N;
  my @invalid =	grep {not UCS::ValidName($_)} @vars;	# check that all variable names are valid
  if (@invalid) {
    UCS::Warn("$filename: invalid variable name(s): @invalid",
	      "Perhaps a malformed comment line or missing table header??",
	      ">>$line<<");
    return undef;
  }
  $self->add_vars(@vars);
  $self->{IS_NUMERIC} = [map {(UCS::VarType($_) eq "STRING") ? 0 : 1} @VARS];
  # next read from $fh should yield first data line
  return bless($self, $class);
}

sub close {					# in analogy to FileHandles, you can either close or destroy
  my $self = shift;
  my $fh = delete $self->{FH};
  $fh->close
    if defined $fh;
  $self->{EOF} = 1;				# do not attempt further reads
}

sub DESTROY {
  my $self = shift;
  $self->close;
}

sub na {
  my $self = shift;
  my $flag = shift;
  if (defined $flag) {
    $self->{NA} = ($flag) ? 1 : 0;
  }
  else {
    return $self->{NA};
  }
}

sub eof {
  my $self = shift;
  return $self->{EOF};
}

sub valid {
  my $self = shift;
  return $self->{VALID};
}

sub row {
  my $self = shift;
  return $self->{ROW};
}

# read one LINE of data from file, and split into COLUMNS array
sub read {
  my $self = shift;
  if ($self->{EOF}) {				# don't attempt to read beyond end of file
    return 0;
  }
  my $line = $self->{FH}->getline;
  if (not $line) {				# reached end of file
    $self->{EOF} = 1;
    $self->{LINE} = "";
    $self->{COLUMNS} = [];
    $self->{VALID} = 0;
    return 0;
  }
  else {
    chomp $line;
    $self->{LINE} = $line;
    $self->{ROW}++;
    $self->{COLUMNS} = [split /\t/, $line, -1];
    my $N = @{$self->{COLUMNS}};
    $self->{VALID} = ($N == $self->{WIDTH});	# validity check
    my $cols = $self->{COLUMNS};
    my $numeric = $self->{IS_NUMERIC};
    my $na = $self->{NA};
    foreach my $i (0 .. $N-1) {
      if ($cols->[$i] eq "NA") {
	if ($na) {
	  $cols->[$i] = undef;
	}
	else {
	  $cols->[$i] = ($numeric->[$i]) ? 0 : "";
	}
      }
    }
    return 1;
  }
}

sub columns {
  my $self = shift;
  return $self->{COLUMNS};
}

sub value {
  my $self = shift;
  my $var = shift;
  UCS::Die("UCS::DS::Stream::Read: illegal variable name '$var' in value() method call.")
      unless UCS::ValidName($var);
  UCS::Die("UCS::DS::Stream::Read: variable '$var' does not exist in value() method call.")
      unless $self->var($var);
  my $index = $self->var_index($var);
  UCS::Die("UCS::DS::Stream::Read: no valid row data available in value() method call.")
      unless $self->valid;
  return $self->columns->[$index];
}

sub data {
  my $self = shift;
  UCS::Die("UCS::DS::Stream::Read: no valid row data available in data() method call.")
      unless $self->valid;
  my @vars = $self->vars;			# all variables in column order
  my $N = $self->{WIDTH};
  my $columns = $self->columns;
  my $hash = {};
  for my $i (0 .. $N-1) {
    $hash->{$vars[$i]} = $columns->[$i];
  }
  return $hash;
}


package UCS::DS::Stream::Write;

use base qw<UCS::DS>;

sub new {
  my $class = shift;
  my $filename = shift;
  my $self = new UCS::DS;
  $self->{FILENAME} = $filename;		# name of the data set file
  $self->{FH} = undef;				# file handle for writing, will be set by open() method
  $self->{WIDTH} = 0;				# number of columns
  $self->{COLUMNS} = [];			# data for current row, split into columns
  return bless($self, $class);
}

sub open {
  my $self = shift;
  my $filename = $self->{FILENAME};
  UCS::Die "UCS::DS::Stream::Write:  stream for data set file $filename already open."
      if defined $self->{FH};
  my $into = (not $filename =~ /^\s*[\|>]/) ? "> " : ""; # so output pipes are supported as well
  my $fh = UCS::File::Open "$into$filename";
  $self->{FH} = $fh;
  $self->{WIDTH} = $self->vars;
  $self->{COLUMNS} = [];
  # write data set header: first comments, then global variables
  foreach my $line ($self->comments) {
    print $fh "# $line\n";
  }
  foreach my $var ($self->globals) {
    print $fh "##:: $var = ", $self->global($var), "\n";
  }
  # write table header, i.e. names of variables in column order
  print $fh join("\t", $self->vars), "\n";
}

sub columns {
  my $self = shift;
  my $W = $self->{WIDTH};
  my $N = @_;
  UCS::Die "UCS::DS::Stream::Write:  incorrect number of values passed to columns() method (expected $W, got $N)"
      unless $N == $W;
  $self->{COLUMNS} = [@_];			# might change this to \@_ to improve performance (?)
}

sub data {
  my $self = shift;
  my $arghash;
  if (@_ == 1) {
    $arghash = shift;
    UCS::Die 'UCS::DS::Stream::Write Usage:  $ds->data($arghash);'
	unless ref($arghash) eq "HASH";
  } 
  else {
    UCS::Die 'UCS::DS::Stream::Write:  uneven number of arguments in $ds->data($v1 => $val1, ...);'
	unless (@_ & 0x01) == 0;
    $arghash = {@_} ;
  }
  foreach my $var (keys %$arghash) {
    my $idx = $self->var_index($var);
    $self->{COLUMNS}->[$idx] = $arghash->{$var}
      if defined $idx;
  }
}

sub write {
  my $self = shift;
  my $fh = $self->{FH};
  my $filename = $self->{FILENAME};
  UCS::Die "UCS::DS::Stream::Write:  \$ds->write called for non-open data set file $filename"
      unless defined $fh;
  print $fh join("\t", map { (defined $_) ? $_ : "NA" } @{$self->{COLUMNS}}[0 .. $self->{WIDTH}-1]), "\n"
    or UCS::Die "UCS::DS::Stream::Write:  error writing data set file $filename ($!)";
  $self->{COLUMNS} = [];
}

sub close {
  my $self = shift;
  my $fh = $self->{FH};
  my $filename = $self->{FILENAME};
  if (defined $fh) {
    UCS::Warn "UCS::DS::Stream::Write:  closing data set file $filename discards unflushed buffer data"
	if @{$self->{COLUMNS}} > 0;
    $fh->close
      or UCS::Die "UCS::DS::Stream::Write:  error writing data set file $filename ($!)";
    $self->{FH} = undef;
  }
}

sub DESTROY {
  my $self = shift;
  $self->close;
}
 

package UCS::DS::Stream;

1;

__END__


=head1 NAME

UCS::DS::Stream - I/O streams for data set files

=head1 SYNOPSIS

  use UCS::DS::Stream;

  $ds = new UCS::DS::Stream::Read $filename;
  die "format error" unless defined $ds;
  # access variables, comments, and globals with UCS::DS methods
  while ($ds->read) {
    die "read/format error"
      unless $ds->valid;		# valid row data available?
    $n = $ds->row;			# row number
    $idx = $ds->var_index("am.logl");	# see 'ucsdoc UCS::DS'
    $logl = $ds->columns->[$idx];	# $ds->columns returns arrayref
    $logl = $ds->value("am.logl");	# short and safe, but slower
    $rowdata = $ds->data;		# returns hashref (varname => value)
    $logl = $rowdata->{"am.logl"};	# == $ds->value("am.logl") 
  }
  ds->close;

  $ds = new UCS::DS::Stream::Write $filename;
  # set up variables, comments, and globals with UCS::DS methods
  $ds->open;				# write data set header
  foreach $i (1 .. $N) {
    $ds->data("id"=>$i, "l1"=>$l1, ...);# takes hashref or list of pairs
    $ds->data("am.logl"=>$logl, ...);   # may be called repeatedly to add data
    $ds->columns($i, $l1, $l2, ...);	# complete list of column data
    $ds->write;				# write row and clear data cache
  }
  $ds->close;

=head1 DESCRIPTION

B<UCS data set streams> are used to read and write data set files one
row at a time.  When an B<input stream> is created, the corresponding
data set file is opened immediately and its header is read in.  The
header information can then be accessed through B<UCS::DS> methods.
Each B<read> method call loads a single row from the data set file
into an internal representation, from which it is available to the
main program.

An B<output stream> creates / overwrites its associated data set file
only when the B<open> method is called.  This allows the main program
to set up variables and header data with B<UCS::DS> method calls.
After opening the file, the data for each row is first stored in an
internal representation, and then written to disk with the B<write>
method.

Note that there are no objects of class B<UCS::DS::Stream>.  Both
input and output streams inherit directly from the B<UCS::DS> class.


=head1 INPUT STREAMS

B<Input streams> are implemented as B<UCS::DS::Stream::Read> objects.
When an input stream is created, the header of the associated data set
file is read in.  Header data and information about the variables in
the data set can then be accessed using B<UCS::DS> methods.

The actual data set table is then loaded one row (= pair type) at a
time by calling the B<read> method.  The row data are extracted into
an internal representation where they can be accessed with various
methods (some of them being safe, others more efficient). 

The B<na> method controls whether missing values (represented by the
string C<NA> in the data set file) are recognised and stored
internally as B<undef>s, or whether they are silently translated into
0 (C<BOOL>, C<INT>, and C<DOUBLE> variables) and the empty string
(C<STRING> variables), respectively.

=over 4

=item $ds = new UCS::DS::Stream::Read $filename;

Open data set file I<$filename> and read header information.  Header variables
and comments, as well as information about the variables in the data set can
then be accessed with B<UCS::DS> methods.  If I<$filename> is a plain filename
or a partial path (i.e., neither a full relative or absolute path starting
with C</> or C<./> nor a command pipe) and the file is not found in the
current working directory, the standard UCS libary is automatically searched
for a data set with this name.

If there is a syntax error in the data set header, B<undef> is returned.  Note
that the object constructor will B<die> if the file I<$filename> does not
exist or cannot be opened for reading.

=item $ds->na(1);

Enables recognition of missing values represented by the string C<NA>
(as used by B<R>).  When enabled, missing values are represented by
B<undef>s.  Otherwise, they will be silently translated into 0
(C<BOOL>, C<INT>, and C<DOUBLE> variables) and the empty string
(C<STRING> variables), respectively.  Use C<< $ds->na(0); >> to
disable missing value support, which is by default activated.

=item $ok = $ds->read;

Read one line of data from the data set file and extract the field
values into an internal representation.  Returns false when the entire
data set has already been processed.  Typically used in a B<while>
loop similar to the diamond operator: C<< while ($ds->read) {...} >>.

=item $at_end = $ds->eof;

Returns true when the entire data set has been read, i.e. the logical
complement of the value returned by the last B<read> call.

=item $ok = $ds->valid;

Returns true if the internal representation contains valid row data.
Currently, this only compares the number of columns in the file
against the number of variables in the data set.  Later on, values may
also be syntax-checked and coerced into the correct data type.

=item $n = $ds->row;

Returns the current row number (of the row read in by the last B<read>
call, which is now stored in the internal representation).

=item $value = $ds->value($name);

Get value by variables name.  Returns the value of variable I<$name>
currently stored in the internal representation.  This method is
convenient and safe (because it checks that the variable I<$name>
exists in the given data set), but incurs considerable overhead.

=item $cols = $ds->columns;

Return entire row data as an array reference.  Individual variables
have to be identified by their index, which can be obtained with the
B<var_index> method (C<< $cols->[$idx] >>. Since index lookup can be
moved out of the row processing loop, this access method is much more
efficient than its alternatives.  B<NB:> the array I<@$rowdata> is not
reused for the next line of input and can safely be integrated into
user-defined data structures.

=item $rowdata = $ds->data;

Returns hash reference containing entire row data indexed by variable
names.  Thus, the values of individual variables can be accessed with
the expression C<< $rowdata->{$varname} >>, similar to using the
B<value> method.  Access with the B<data> method is convenient for
copying row data to an output stream.  It is relatively slow, though,
and should not be used in tight loops.

=item $ds->close;

Close the data set file.  This method is automatically invoked when
the object I<$ds> is destroyed.

=back


=head1 OUTPUT STREAMS

B<Output streams> are implemented as B<UCS::DS::Stream::Write>
objects.  After creating an output stream object, variables and header
data are set up with the B<UCS::DS> methods.  The data set header is
written to disk when the B<open> method is called.

After that, the actual data set table is generated one row at a time.
Row data is first stored in the internal presentation (using the
B<data> or the B<columns> method), and then written to disk when the
B<write> method is called.

=over 4

=item $ds = new UCS::DS::Stream::Write $filename;

Create output stream for data set file I<$filename>.  Note that this
file will only be created or overwritten when the B<open> method is
called (in contrast to input streams, which open the data set file
immediately).

=item $ds->open;

After setting up variables and header data (comment lines and global
variables) with the respective B<UCS::DS> methods, the B<open> method
opens the data set file and writes the data set header.  If the file
cannot be opened for writing, the B<open> method will B<die> with an
error message.

=item $ds->data($v1 => $val1, $v2 => $val2, ...);

=item $ds->data($hashref);

Store data for the next row to be written in an internal
representation.  When using the B<data> method, variables are
identified by name (I<$v1>, I<$v2>, ...) and can be specified in any
order.  The variable-value pairs can also be passed with a single hash
reference.  Variables that do not exist in the data set will be
silently ignored.  The B<data> method can be called repeatedly for a
single row.

=item $ds->columns($val1, $val2, ...);

The B<columns> method provides a more efficient way to specify row
data.  Here, all column values are passed in a single method call, and
care has to be taken to list them in the correct order (namely, the
order in which the variables were set up with the B<add_vars> method).
B<NB:> the B<data> and B<columns> methods cannot be mixed.  It is also
not possible to set up the row data incrementally with repeated
B<columns> calls.

=item $ds->write;

Writes the row data currently stored in the internal buffer to the
data set file, and resets the buffer (to B<undef> values).  Any
B<undef> values in the buffer (including the case where some variables
were not specified with the B<data> method) are interpreted as missing
values and substituted by the string C<NA>.

=item $ds->close;

Completes and closes the data set file. 


=back



=head1 EXAMPLES

The recommended way of B<copying rows> from one data set file to
another is to use the B<data> methods of both streams, so that
variables are copied by name rather than column position.  It would be
more efficient to pass row data directly (using the B<columns>
methods), but this approach is prone to lead to errors when the order
of the columns is different between the input and output data sets.

The following example makes a copy of a data set file, adding an
(enumerative) C<id> variable if it is not present in the source file.

  $in = new UCS::DS::Stream::Read $input_file;
  die "$input_file: format error"
    unless defined $in;
  @vars = $in->vars;
  $add_id = not $in->var("id");

  $out = new UCS::DS::Stream::Write $output_file;
  $out->copy_comments($in);             # copy comments and
  $out->copy_globals($in);              # global variables from input file
  $out->add_vars("id")                  # conventionally, the "id" variables
    if $add_id;                         # is in the first column
  $out->add_vars(@vars);
  $out->open;                           # writes header to $output_file

  while ($in->read) {
    die "read/format error"
      unless $in->valid;
    $out->data($in->data);		# copy row data by field name
    $out->data("id" => $in->row)        # use row number as ID value
      if $add_id;
    $out->write;
  }

  $in->close;
  $out->close;

=head1 COPYRIGHT

Copyright 2004 Stefan Evert.

This software is provided AS IS and the author makes no warranty as to
its use and performance.  You may use the software, redistribute and
modify it under the same terms as Perl itself.

=cut

