## Filename: UCS/Expression.pm
## Modified: Sun Jun 27 11:33:25 2004 (evert)   
##   Author: Stefan Evert
##  Purpose: Compile and execute UCS expressions

package UCS::Expression;

use locale;

use UCS;
use UCS::SFunc;					# import special functions and probability distributions
use UCS::Expression::Func;			# import utility functions for UCS expressions


# $Debug = 1;					# just while developing it ...

##
## UCS::Expression object format:
##   $exp->{EXP}        = $ucs_expression;
##   $exp->{NEEDED}     = [@additional_required_variables];
##   $exp->{DERIVED}    = [@variables_derived_from_frequency_signature];
##   $exp->{PARAM}      = [@parameter_names];
##   $exp->{VALUE}->{$param} = $value;
##   $exp->{CODE}       = $compiled_perl_code;
##   $exp->{LOOPCODE}   = $compiled_code_for_evaluation_loop;


##
##  helper functions for compiling UCS expressions
##

# rules for computing derived variables from frequency signature
%derivation_rule = (
		    "O11" => '%f%',
		    "O12" => '%f1% - %f%',
		    "O21" => '%f2% - %f%',
		    "O22" => '%N% - %f1% - %f2% + %f%',
		    "E11" => '%f1% * %f2% / %N%',
		    "E12" => '%f1% * (%N% - %f2%) / %N%',
		    "E21" => '(%N% - %f1%) * %f2% / %N%',
		    "E22" => '(%N% - %f1%) * (%N% - %f2%) / %N%',
		    "R1"  => '%f1%',
		    "R2"  => '%N% - %f1%',
		    "C1"  => '%f2%',
		    "C2"  => '%N% - %f2%',
		    "lf"  => 'log(%f%) / log(10)',
		    "lf1" => 'log(%f1%) / log(10)',
		    "lf2" => 'log(%f2%) / log(10)',
		    "e"   => '(%f1% * %f2%) / %N%',
		    "b"   => '%f2% / %f1%',
		    "o"   => '%f%',
		    "le"  => 'log((%f1% * %f2%) / %N%) / log(10)',
		    "lb"  => 'log(%f2% / %f1%) / log(10)',
		    "lo"  => 'log(%f%) / log(10)',
		    "p"   => '%f% / %N%',	# for backward compatibility (cf. UCS.pm)
		    "p1"  => '%f1% / %N%',
		    "p2"  => '%f2% / %N%',
		    "lp"  => '-log(%f% / %N%) / log(10)',
		    "lp1" => '-log(%f1% / %N%) / log(10)',
		    "lp2" => '-log(%f2% / %N%) / log(10)',
		    );
		    
# @vars = extract_vars($exp, $ignore);
# extract list of variables referenced in UCS expression;
# $ignore is a lookup hash of expression paramters (values are set to 1 for parameters actually used in $exp)
sub extract_vars {
  my $exp = shift;
  my $ignore = (@_) ? shift : {};
  my %vars = map {$_ => 1} ( $exp =~ /\%([A-Za-z0-9.]+)\%/g ); # extract variable references
  foreach my $v (keys %vars) {			# remove parameter references from extracted list
    if (exists $ignore->{$v}) {
      $ignore->{$v} = 1;
      delete $vars{$v};
    }
  }
  my @vars = sort keys %vars;			# make unique list
  my @invalids = grep { not UCS::ValidName($_) } @vars;
  if (@invalids) {
    UCS::Warn("UCS::Expression: illegal variable reference(s) " . 
	      join(", ", map {"%$_%"} @invalids)." in UCS expression",  
	      "$exp");
    return undef;				# single undefined value indicates failure
  }
  return @vars;
}

# ($core_vars, $derived_vars, $other_vars) = split_variables(@vars);
# split list of variables into core variables (except frequency signature), derived variables, and others;
# @vars may include repetitions (which are deleted); for derived variables, the core variables from which
# they are derived (f, f1, f2, N) are automatically added to $core_vars
sub split_variables {
  my %core = ();
  my %derived = ();
  my %other = ();
  my %is_core = map {$_ => 1} @UCS::CoreVars;
  my %is_derived = map {$_ => 1} @UCS::DerivedVars;
  foreach my $v (@_) {
    if (exists $is_core{$v}) {
      $core{$v}++;
    }
    elsif (exists $is_derived{$v}) {
      $derived{$v}++;
      foreach my $w (extract_vars($derivation_rule{$v})) {
	$core{$w}++;
      }
    }
    else {
      $other{$v}++;
    }
  }
  return ([sort keys %core], [sort keys %derived], [sort keys %other]);
}

# $perl_var = ucs2perl($ucs_var);
# translate UCS variable reference to Perl variable name (%am.logl% -> $_am_logl, etc.)
sub ucs2perl {
  my $var = shift;
  $var =~ tr[.][_];
  return '$_'.$var;
}

# $perl_exp = substitute_variables($exp);
# substitute all UCS variable references in $exp by Perl variables
sub substitute_variables {
  my $exp = shift;
  $exp =~ s{\%([A-Za-z0-9.]+)\%}{" ".ucs2perl($1)." "}eg;
  return $exp;
}

sub new {
  my $class = shift;
  my $exp = shift;				# the UCS expression to be compiled
  UCS::Die('UCS::Expression Usage:  $exp = new UCS::Expression $code [, $param => $value, ...];')
      unless defined $exp and (@_ % 2) == 0;
  chomp($exp);
  my $self = {};
  $self->{EXP} = $exp;				# uncompiled UCS expression
  print STDERR "EXPRESSION: $exp\n"
    if $Debug;

  $self->{VALUE} = {};		                # parameters with default values (if any)
  while (@_) {
    my $par = shift;
    my $val = shift;
    UCS::Die "UCS::Expression:  illegal parameter name '$par'."
	unless UCS::ValidKey($par) and not UCS::ValidName($par);
    UCS::Die "UCS::Expression:  parameter '$par' specified more than once in constructor."
	if exists $self->{VALUE}->{$par};
    $self->{VALUE}->{$par} = $val;
  }
  my @param = sort keys %{$self->{VALUE}};
  $self->{PARAM} = [@param];

  my %pars = map {$_ => 0} @param;	        # check $exp for parameter references
  my @vars = extract_vars($exp, \%pars);        # extract variable references from $exp
  return undef if @vars == 1 and not defined $vars[0];
  print STDERR "Variables:  @vars\n"
    if $Debug;
  my @unused = map { "\%$_\%" } grep { not $pars{$_} } keys %pars; # warn about unused parameters
  UCS::Warn "UCS::Expression:  parameter(s) @unused defined but not used."
      if @unused > 0;

  my ($c, $d, $o) = split_variables(@vars);
  my @args = (@$c, @$o);
  my @derived = @$d;
  $self->{NEEDED} = [@args];			# data set variables required for evaluation
  $self->{DERIVED} = [@derived];		# derived variables computed from frequency signature
  if ($Debug) {
    print STDERR "Derived:    @derived\n";
    print STDERR "Needed:     @args\n";
  }

  # now build full Perl sourcecode to be compiled into anonymous subroutine
  my $body = "";				# body of subroutine: computes derived variables & evaluates UCS expression
  foreach my $v (@derived) {			# - compute derived variables
    $body .= ucs2perl($v) . " = " . substitute_variables($derivation_rule{$v}) . ";\n";
  }
  $body .= substitute_variables($exp) . "\n";	# - translate UCS expression into Perl

  # single evaluation Perl code: parameters are scalar values
  my $perl = "sub {\n";				# (i) assign arguments to local variables
  $perl .= "my (" . join(", ", map {ucs2perl($_)} (@param, @args)) . ") = \@_;\n"
    if (@param + @args) > 0;
  $perl .= $body;
  $perl .= "}\n";
  print STDERR "PERL CODE:\n$perl"
    if $Debug;

  # try to compile Perl sourcecode for single evaluation version
  my $code = eval($perl);
  unless (defined $code) {
    UCS::Warn "UCS::Expression:  Compilation of UCS expression", $exp, "failed because of", $@;
    return undef;
  }
  $self->{CODE} = $code;
  
  # loop evaluation Perl code: parameters are array references
  $perl = "sub {\n";
  $perl .= "my \$_OUT=shift; my \$_SIZE = shift;\n";
  foreach my $par (@param) {
    $perl .= "my ".ucs2perl($par)." = shift;\n";
  }
  $perl .= "foreach my \$_ROW (0 .. \$_SIZE-1) {\n";
  for my $i (0 .. $#args) {
    $perl .= "my " . ucs2perl($args[$i]) . " = \$_[$i][\$_ROW];\n";
  }
  $perl .= "\$_OUT->[\$_ROW] = eval {\n$body} # eval\n";
  $perl .= "} # foreach\n} # sub\n";
  print STDERR "LOOP CODE:\n$perl"
    if $Debug;

  # compile Perl sourcecode for loop evaluation version (shouldn't fail if first version was ok)
  $code = eval($perl);
  UCS::Die "UCS::Expression:  Internal error. Could compile CODE for expression", $exp, "but LOOPCODE failed because", $@
      unless defined $code;
  $self->{LOOPCODE} = $code;
  
  return bless($self, $class);
}

sub copy {
  my $self = shift;
  my $new = {};
  $new->{EXP} = $self->string;			# make actual copies of all data fields
  $new->{NEEDED} = [ $self->needed ];
  $new->{DERIVED} = [ @{$self->{DERIVED}} ];
  $new->{PARAM} = [ $self->params ]; 
  $new->{VALUE} = { %{$self->{VALUE}} };
  $new->{CODE} = $self->{CODE};			# these are just references to the same code objects
  $new->{LOOPCODE} = $self->{LOOPCODE};		# in order to avoid recompilation
  return bless($new, ref $self);
}

sub string {
  my $self = shift;
  return $self->{EXP};
}

sub needed {
  my $self = shift;
  return @{$self->{NEEDED}};
}

sub code {
  my $self = shift;
  return $self->{CODE};
}

sub param {
  my $self = shift;
  UCS::Die 'UCS::Expression Usage:  $value = $exp->param($name);'
      unless @_ == 1;
  my $par = shift;
  UCS::Die "UCS::Expression:  parameter '$par' doesn't exist in param() call"
      unless exists $self->{VALUE}->{$par};
  return $self->{VALUE}->{$par};
}

sub params {
  my $self = shift;
  UCS::Die 'UCS::Expression Usage:  $value = $exp->params();'
      unless @_ == 0;
  return @{$self->{PARAM}};
}

sub set_param {
  my $self = shift;
  UCS::Die 'UCS::Expression Usage:  $exp->set_param($name, $value);'
      unless @_ == 2;
  my $par = shift;
  my $value = shift;
  UCS::Die "UCS::Expression:  parameter '$par' doesn't exist in set_param() call"
      unless exists $self->{VALUE}->{$par};
  $self->{VALUE}->{$par} = $value;
}

sub eval {
  my $self = shift;
  my @needed = @{$self->{NEEDED}};		# list of required variables
  my @param_values = @{ $self->{VALUE} }{ @{$self->{PARAM}} }; # extract parameter values as hash slice
  if (@_ == 1) {				# $exp->eval($arghash);
    my $arghash = shift;
    UCS::Die 'UCS::Expression Usage:  $exp->eval($arghash);'
	unless ref($arghash) eq "HASH";
    my @missing = grep { not exists $arghash->{$_} } @needed; # check for missing variables
    UCS::Die "UCS::Expression:  missing argument(s) in \$exp->eval() call. (".join(", ", @missing).")"
	if @missing;
    my @values = @{$arghash}{@needed};		# retrieve variables values from hash
    return $self->{CODE}->(@param_values, @values); # evaluate Perl code
  }
  else {					# $exp->eval(@args);
    UCS::Die 'UCS::Expression: wrong number of arguments in $exp->eval(@args).',
	"(expected ".(@needed+0).", got ".(@_+0).")"
	  unless @needed == @_;			# check that number of arguments is correct at least
    return $self->{CODE}->(@param_values, @_);
  }
}

sub evalloop {
  my $self = shift;
  my @needed = @{$self->{NEEDED}};		# list of required variables
  my @param_values = @{ $self->{VALUE} }{ @{$self->{PARAM}} }; # extract parameter values as hash slice
  my ($size, $result, $arghash);
  if (@_ == 3) {
    ($result, $size, $arghash) = @_;
    UCS::Die 'UCS::Expression:  first argument in $exp->evalloop(\@result, $size, $arghash) must be array reference.'
	unless ref($result) eq "ARRAY";
  }
  elsif (@_ == 2) {
    ($size, $arghash) = @_;
    $result = [];
    $result->[$size-1] = 0;			# pre-allocate output array
  }
  else {
    UCS::Die 'UCS::Expression Usage:  $arrayref = $exp->evallop($size, $arghash); OR $exp->evalloop(\@result, $size, $arghash);';
  }
  UCS::Die 'UCS::Expession:  last argument in $exp->evalloop(...) call must be hash reference.'
      unless ref($arghash) eq "HASH";
  my @arefs = ();				# retrieve variables (= array references) from hash and run error checks 
  foreach my $v (@needed) {
    my $aref = $arghash->{$v};
    UCS::Die "UCS::Expression:  missing argument $v in \$exp->evalloop() call."
	unless defined $aref;
    UCS::Die "UCS::Expression:  \$arghash->{$v} is not an array reference in \$exp->evalloop(..., \$arghash)"
	unless ref($aref) eq "ARRAY";
    my $asize = @$aref;
    UCS::Die "UCS::Expression:  array \$arghash->{$v} has incorrect size ($asize instead of $size) in \$exp->evalloop(..., \$arghash)"
	unless $asize == $size;
    push @arefs, $aref;
  }
  $self->{LOOPCODE}->($result, $size, @param_values, @arefs);	# evaluate loop version of Perl code
  return $result;				# always return reference to result array
}


1;



__END__


=head1 NAME

UCS::Expression - Compile and execute UCS expressions


=head1 SYNOPSIS

  use UCS::Expression;

  $exp = new UCS::Expression $code;  # compile UCS expression
  @vars = $exp->needed;		     # variables needed to evaluate expression
  $code = $exp->string;		     # retrieve sourcecode of UCS expression
  $result = $exp->eval(@args);	     # evaluate UCS expression (argument list)
  $result = $exp->eval($arghash);    # named arguments (UCS variable names)

  $exp = new UCS::Expression $code, "MU" => 10, ...; # expression with parameters
  @params = $exp->params;	     # sorted list of parameter names
  $value = $exp->param("MU");        # current value of parameter
  $exp->set_param("MU", 1);          # set parameter value
  $exp2 = $exp->copy;                # clone expression (e.g. when changing parameters) 
  
  $sub = $exp->code;		     # reference to compiled Perl expression
  $result = $sub->(@args);           # argument list is same as for eval()

  $listref = $exp->evalloop($size, $arghash);  # evaluate expression on full data set
  $exp->evalloop(\@result, $size, $arghash);   # directly writes to array @result


=head1 DESCRIPTION

B<UCS expressions> provide a convenient way to evaluate functions and
conditions on the pair types in a data set.  They consist of arbitrary
Perl code with a syntax extension for direct access to data set
variables: the character sequence C<%I<varname>%> (where C<I<varname>>
is a legal UCS variable name) is replaced by the value of this
variable (for the current pair type).  See L<ucsexp> for a more
detailed description of UCS expressions and some cautionary remarks.

A B<UCS::Expression> object represents a compiled UCS expression.  The
B<needed> method returns a list of UCS variables that are required for
evaluation of the expression.  When B<derived variables> are used in a UCS
expression, they are automatically computed from the frequency signature.

The B<eval> method is normally invoked with a (reference to a) hash of
arguments, using UCS variable names as keys.  It selects the variables needed
to evaluate the UCS expression automatically from the hash, and ensures that
all of them are present.  Better performance is achieved by passing the
required variables as an argument list in the correct order (as returned by
B<needed>).

The B<evalloop> method greatly reduces overhead when a UCS expression
is applied to a list of pair types (i.e. a full data set).  It expects
array references instead of simple variable values, and returns a
reference to an array of the specified length.  Optionally, B<evalloop>
can write directly to an existing array.


=head1 METHODS

=over 4

=item $exp = new UCS::Expression $code;

Compiles the UCS expression I<$code> into a B<UCS::Expression> object.
If compilation fails for some reason, an B<undef>ined value is
returned.  Compiling a UCS expression involves the follwing steps:

=over 2

=item * 

All UCS variable references in I<$code> are identified and validated.

=item * 

A list of required variables is constructed.  Derived variables are implicitly
computed from the frequency signature, and the necessary core variables are
automatically added to the list of required variables.

=item * 

The UCS variable references are substituted with lexical Perl variables,
which are initialised from the parameter list C<@_>.

=item *

The resulting Perl code is compiled into an anonymous subroutine,
which is stored in the B<UCS::Expression> object and can be executed
through the B<eval> method.  

=back

Since B<UCS::Expressions> are comparatively small structures, it is
usually not necessary to destroy them explicitly.

=item $exp = new UCS::Expression $code, $param => $value, ...;

This form of the constructor defines a UCS expression with parameters, given
as pairs of parameter name I<$param> and default value I<$value>.  Parameters
can be used like variables in the UCS expression.  Their names are simple UCS
identifiers, but B<must not> be valid UCS variable names.  The recommended
convention is to write parameter names all in upper case.

=item @names = $exp->params;

Returns the names of all parameters in alphabetical order.

=item $value = $exp->param($name);

Returns the current value of parameter I<$name>;

=item $exp->set_param($name, $value);

Set the parameter I<$name> to the value I<$value>.  The new value will be used
by all subsequent calls to the B<eval> and B<evalloop> methods.

=item $new_exp = $exp->copy;

Makes a clone of the B<UCS::Expression> object I<$exp>.  Cloning is a fast
operation and should always be used when changing the parameters of an
expression shared between different modules (e.g. a registered association
measure).

=item @vars = $exp->needed;

The B<needed> methods returns a list of UCS variable names, corresponding to
the data set variables needed to evaluate I<$exp>.

=item $code = $exp->string;

Returns the original UCS expression represented by I<$exp> as a
string, and can be used to modify and recompile UCS expressions
(especially those of built-in association measures).  Note that
I<$code> is B<chomp>ed, but may contain internal linebreaks (C<\n>).

=item $result = $exp->eval($arghash);

The B<eval> method evaluates a compiled UCS expression on the data
passed in I<$arghash>, which must be a reference to a hash of variable
names and the corresponding variable values.  The necessary variables
are extracted from I<$arghash> by name, and the method B<die>s with an
error message unless all required variables are present.  Unused
variables are silently ignored.

=item $result = $exp->eval(@args);

The second form of the B<eval> method avoids the overhead of variable name
lookup and error checking.  Here, the argument list I<@arg> consists of the
values of all required variables in the order defined by the B<needed> method.
The list I<@args> is passed directly to the compiled Perl code, so that errors
will usually go undetected.

=item $sub = $exp->code;

The B<code> method returns a code reference to the anonymous subroutine that
resulted from compilation of the UCS expression.  For an expression without
parameters, the subroutine call

  $result = $sub->(@args);

is equivalent to

  $exp->eval(@args);

and further reduces overhead (by a small amount).  It may be useful
when the UCS expression is repeatedly applied, looping over a list of
pair types.  In most such cases, the B<evalloop> method provides a
better solution, though.

=item $listref = $exp->evalloop($size, $arghash);

=item $exp->evalloop(\@result, $size, $arghash);

The B<evalloop> method is used to apply I<$exp> to an entire list of
pair types (i.e. a data set) with a single call.  Its invocation is
similar to the firs form of the B<eval> method.  The additional
parameter I<$size> specifies the number of pair types to be processed.
Each value in I<$arghash> must be a reference to an array of length
I<$size>.  The return value is a reference to an array of the same
length.

The three-parameter form allows B<evalloop> to write the results
directly into an existing array, which may save a considerable amount
of overhead when I<$size> is large.

=back


=head1 SEE ALSO

See L<the ucsexp manpage|ucsexp> for an introduction to UCS expressions, as
well as the L<UCS::SFunc|UCS::SFunc> and
L<UCS::Expression::Func|UCS::Expression::Func> manpages for pre-defined
functions that may be used in UCS expressions.


=head1 COPYRIGHT

Copyright 2004 Stefan Evert.

This software is provided AS IS and the author makes no warranty as to
its use and performance. You may use the software, redistribute and
modify it under the same terms as Perl itself.

=cut

