## Filename: UCS/Expression/Func.pm
## Modified: Tue Apr  6 10:29:20 2004 (evert)   
##   Author: Stefan Evert
##  Purpose: Useful functions that are available in UCS expressions

package UCS::Expression::Func;

use base qw<Exporter>;
@EXPORT = (qw<min max log10 inf>);


sub min ( @ ) {
  my $min = shift;
  foreach my $x (@_) {
    $min = $x if $x < $min;
  }
  return $min;
}

sub max ( @ ) {
  my $max = shift;
  foreach my $x (@_) {
    $max = $x if $x > $max;
  }
  return $max;
}

sub log10 ( $ ) {
  my $x = shift;
  return ($x > 0) ? log($x) / log(10) : -inf();
}

sub inf ( ) {
  return 10e6;
}


1;



__END__


=head1 NAME

UCS::Expression::Func - Utility functions for UCS expressions


=head1 SYNOPSIS

  use UCS::Expression::Func;

  $min_x = min($x1, $x2, ...);	   # minimum of two or more values
  $max_y = max(@y);		   # maximum of two or more values

  $log_prob = -log10($prob);	   # base 10 logarithm

  $log_prob = inf()                # replace log(Infinity) = -log(0)
    if $prob == 0;                 #   by a very large value 


=head1 DESCRIPTION

This module provides a collection of simple but useful functions,
which are automatically imported into the B<UCS::Expression> namespace
so that they can be used in B<UCS expressions> without full qualification.


=head1 FUNCTIONS

=over 4

=item $min_x = min($x1, $x2, ...);

Minimum of two or more numbers.  The argument could also be an array C<@x>.

=item $max_x = max($x1, $x2, ...);

Maximum of two or more numbers.  The argument could also be an array C<@x>.

=item $log_prob = -log10($prob);

Base 10 logarithm, which is used for all logarithmic scales in UCS
(especially logarithmic p-values).  Returns C<-inf()> if I<$prob> is
zero or negative.

=item $log_infinity = inf();

The B<inf> function returns a large positive floating-point value that
represents the logarithm of Infinity in UCS/Perl.  Note that the
logarithm of 0 should consequently be represented by C<-inf()>, as does
the B<log10> function.  In order to find out the exact value on your
system, you can use the command line

  ucs-config -e 'print inf(),"\n"'

=back


=head1 COPYRIGHT

Copyright 2004 Stefan Evert.

This software is provided AS IS and the author makes no warranty as to
its use and performance. You may use the software, redistribute and
modify it under the same terms as Perl itself.

=cut

