## Filename: UCS/File.pm
## Modified: Mon Jan 17 21:08:06 2005 (evert)   
##   Author: Stefan Evert
##  Purpose: some utilities for working with files (does not require PDL)

package UCS::File;

use UCS;

# internal helper function: quote command-line arguments for Bourne shell
sub quote_sh ( @ ) {
  my @args = @_;				# don't modify original data
  my @quoted = map { 
    if (/^[\/A-Za-z0-9._=-]+$/) {
      $_;					# these strings should be safe (and cover most filenames & options)
    } 
    else {
      s/\'/'"'"'/g; "'$_'"; 			# enclose in single quotes, escape ' as '"'"' :o)
    } 
  } @args;
  return (wantarray or @_ > 1) ? @quoted : $quoted[0];
}

# open file for reading or writing, with 'magical' compression/decompression
# returns FileHandle object; if open fails, exits with error message
sub Open ( $ ) {
  my $name = shift;
  my $fh = TryOpen($name);
  UCS::Die("UCS::File::Open(): Can't open file/pipe '$name'.", "[$!]")
    unless defined $fh;
  return $fh;
}

# same as UCS::File::Open(), but just returns undef if open fails
sub TryOpen ( $ ) {
  my $name = shift;
  my $openstring = "";				# string passed to open() function
  UCS::Die("UCS::File::TryOpen():  Sorry, read/write access ('$name') not supported")
      if $name =~ /^\s*\+/;
  if ($name =~ /^\s*\| | \|\s*$/x) {
    $openstring = $name;			# no magic on read or write pipes
  }
  else {
    my $write = "";				# default operation is read ($write is true in write mode)
    if ($name =~ s/^\s*(<|>>?)//) {
      $write = $1 unless $1 eq "<" ;		# value of $write distinguishes between write and append mode
    }
    $name =~ s/^\s+|\s+$//g;			# remove leading and trailing whitespace from filename
    my $exists = ($name eq "-" or -f $name);	# check whether file exists (stdin/stdout always exists)
    return undef if (not $write) and (not $exists); # ... can't read if it doesn't exist
    ### Don't erase file before overwriting, as this might change access permissions and group membership
    # unlink $name if $exists and ($write eq ">");    # ... erase file if it already exists and will be overwritten
    if ($name =~ /\.(tgz|tar\.(gz|Z))$/) {
      return undef if $write;	  # can't _create_ archive from program output
      $openstring = "gzip -cd ".quote_sh($name)." | tar xfO - |";
    }
    elsif ($name =~ /\.tar$/) {
      return undef if $write;	  # can't _create_ archive from program output
      $openstring = "gtar xfO ".quote_sh($name)." |";
    }
    elsif ($name =~ /\.gz$/) {			# Yes, gzip and bzip2 can append to a compressed file. Cool.
      $openstring = ($write) ? "| gzip $write ".quote_sh($name)."" : "gzip -cd ".quote_sh($name)." |"; 
    }
    elsif ($name =~ /\.bz2$/) {
      $openstring = ($write) ? "| bzip2 $write ".quote_sh($name)."" : "bzcat ".quote_sh($name)." |"; 
    }
    elsif ($name =~ /\.Z$/) {
      $openstring = ($write) ? "| compress $write ".quote_sh($name)."" : "zcat ".quote_sh($name)." |";
    }
    else {
      $openstring = ($write) ? "$write $name" : $name;
    }
  }
  open($fh = undef, $openstring);
  return $fh;
}


##
##  temporary file objects
##

package UCS::File::Temp;

sub new {
  my $class = shift;
  my $prefix = shift;
  $prefix = "UCS_TempFile" unless defined $prefix;
  my $suffix = "";
  if ($prefix =~ s/\.(gz|bz2|Z)$//) {
    $suffix = $&;
  }
  my $self = {};
  my $basedir = ($prefix =~ /\//) ? "" : "/tmp/";  # if $prefix isn't absolute or relative path, create temp file in /tmp directory
  my $name = $basedir.$prefix.".$$".$suffix;
  my $num = 1;
  while (-e $name) {		# choose unique name in case file already exists
    $name = $basedir.$prefix.".$$-".$num.$suffix;
    $num++;
  }
  my $fh = UCS::File::Open "> $name";
  $self->{NAME} = $name;
  $self->{FH} = $fh;
  $self->{STATUS} = "W";	# W = writing, F = finished, R = reading, D = deleted
  return bless($self, $class);
}

sub DESTROY {
  my $self = shift;
  if ($self->{STATUS} ne "D") {
    $self->close;
  }
}

sub close {
  my $self = shift;
  my $status = $self->{STATUS};
  if ($status eq "W" or $status eq "R" and defined $self->{FH}) {
    $self->{FH}->close
      or UCS::Die("UCS::File::Temp: Error writing/reading temporary file.", "[".$self->name.": $!]");
  }
  unlink $self->{NAME}
    if -f $self->{NAME};
  $self->{STATUS} = "D";
}

sub name {
  my $self = shift;
  return $self->{NAME};
}

sub status {
  my $self = shift;
  return {
	  "W" => "WRITING",
	  "F" => "FINISHED",
	  "R" => "READING",
	  "D" => "DELETED",
	 }->{$self->{STATUS}};
}

sub write {
  my $self = shift;
  UCS::Die("UCS::File::Temp: Can't write to temporary file ".$self->name." with status ".$self->status)
      unless $self->{STATUS} eq "W";
  $self->{FH}->print(@_)
    or UCS::Die("UCS::File::Temp: Error writing temporary file.", "[".$self->name.": $!]");
}

sub finish {
  my $self = shift;
  UCS::Die("UCS::File::Temp: Can't finish temporary file ".$self->name." with status ".$self->status)
      unless $self->{STATUS} eq "W";
  $self->{FH}->close
    or UCS::Die("UCS::File::Temp: Error writing temporary file.", "[".$self->name.": $!]");
  $self->{FH} = undef;
  $self->{STATUS} = "F";
}

sub read {
  my $self = shift;
  my $status = $self->{STATUS};
  UCS::Die("UCS::File::Temp: Can't read from temporary file ".$self->name.", already ".$self->status)
    if $status eq "D";
  if ($status eq "W") {
    $self->finish;
  }
  if ($status ne "R") {
    $self->{FH} = UCS::File::Open $self->name;
    $self->{STATUS} = "R";
  }
  return $self->{FH}->getline;
}

sub rewind {
  my $self = shift;
  my $status = $self->{STATUS};
  UCS::Die("UCS::File::Temp: Can't rewind temporary file ".$self->name." with status ".$self->status)
      if $status eq "D" or $status eq "W";
  if ($status ne "R") {		
    # if rewind is called before first read, it does nothing
  }
  else {
    $self->{FH}->close
      or UCS::Die("UCS::File::Temp: Error writing temporary file.", "[".$self->name.": $!]");
    $self->{FH} = UCS::File::Open $self->name;
  }
}


package UCS::File;

##
##  execute shell commands with extensive error checks
##

#   $UCS::File::Paranoid = 1;
# enables rather paranoid checks in UCS::File::ShellCmd(), which will
# die if there is _any_ output on stderr or if the output on
# stdout contains the word "error" or something similar
#   $UCS::File::Paranoid = -1;
# makes UCS::File::ShellCmd() less paranoid: it will only issue a warning
# for non-zero exit codes, and ignore "warn" on stderr
$Paranoid = 0;

# use global variables and sub to handle warn/die situations
$return_status = 0;
$current_cmd = "";

#   ShellError $errlevel, $message [, $message ...];
# error levels are:
# LVL  <less p.>   <normal>   <extra-p.>
#  6 :  fatal      fatal       fatal 
#  5 :  warn       fatal       fatal
#  4 :  warn       warn        fatal
#  3 :  nothing    warn        fatal
#  2 :  nothing    nothing     fatal
#  1 :  nothing    nothing     warn
#  0 :  nothing    nothing     nothing
sub ShellError ( $@ ) {
  my $errlevel = shift;
  my @message = @_;
  # $action is: 0=fatal, 1=warn, 2=nothing 
  my $action;
  
  if ($errlevel >= $return_status) {  
    $return_status = $errlevel;

    if ($Paranoid < 0) {
      $action = [qw<2 2 2 2 1 1 0>]->[$errlevel];
    }
    elsif ($Paranoid == 0) {
      $action = [qw<2 2 2 1 1 0 0>]->[$errlevel];
    }
    else {
      $action = [qw<2 1 0 0 0 0 0>]->[$errlevel];
    }
  }
  else {
    $action = 2;		# don't report this error if a more serious one has already occurred
  }

  if ($action == 0) {
    UCS::Die("SHELL CMD '$current_cmd' FAILED:",
	     map {chomp; ">> $_"} @message);
  }
  elsif ($action == 1) {
    UCS::Warn("PROBLEMS WITH SHELL CMD '$current_cmd':",
	      map {chomp; print ">> $_"} @message);
  }
  else {
    # nothing :o)
  }
}


#   $error_level = UCS::File::ShellCmd($cmd [, ($outfile | \@data)]);
# Runs shell command $cmd with extensive error checks. If an error
# is detected it exits the program with an error message on stderr or
# issues a warning message on stdout, depending on the seriousness of
# the error and the setting of $UCS::File::Paranoid. 
# Returns error level detected, which is either 0 (no error) or a 
# number between 1 (minor problems) and 6 (fatal error). It is normally
# not necessary to check the returned error level.
# Optionally, the standard output of $cmd can either be written to
# $outfile or stored as an array of chomped text lines in a hashref.
# If $cmd is an arrayref, the first array element is interpreted as the
# command name, and any further elements are passed directly as 
# command-line arguments, bypassing shell expansion (if IPC::Run3 is available).
sub ShellCmd {
  my $cmd = shift;
  my $outfile = shift;
  defined $outfile or $outfile = "";

  # create arrays into which stdout / stderr will be read
  # if second argument is arrayref, store stdout in that array, else create private anonymous array
  my $stdout = [];
  if ((ref $outfile) eq 'ARRAY') {
    $stdout = $outfile;
    $outfile = "";		# so we'll create a temporary file
  }
  my $stderr = [];

  my $stdout_tmp = undef;	# create temporary files for capturing stdout and stderr
  my $stderr_tmp = new UCS::File::Temp "UCS-File-ShellCmd-STDERR";

  my $stdout_file = $outfile;
  if (not $outfile) {
    $stdout_tmp = new UCS::File::Temp "UCS-File-ShellCmd-STDOUT";
    $stdout_tmp->finish;	# now we're allowed to access the file directly
    $stdout_file = $stdout_tmp->name;
  }
  $stderr_tmp->finish;
  my $stderr_file = $stderr_tmp->name;

  if (ref($cmd) eq "ARRAY") {			# explicit argument list -> avoid shell expansion
    my ($prog, @args) = @$cmd;
    # for now, quote arguments individually (single quote, using '"'"' to escape embedded single quotes);
    # should use IPC::Run3 module if installed, with possible fallback to IPC::Open3 
    $cmd = join(" ", $prog, quote_sh(@args));
  }

  my $status = system "($cmd) 1>$stdout_file 2>$stderr_file";
  my $syscode = $status & 0xff;
  my $exitval = $status >> 8;

  my $fh = UCS::File::Open $stderr_file;
  @$stderr = <$fh>;
  map {chomp;} @$stderr;
  $fh->close;
  if ($outfile) {
    @$stdout = ();		# don't check STDOUT if caller wants it in file
  }
  else {
    $fh = UCS::File::Open $stdout_file;
    @$stdout = <$fh>;
    map {chomp;} @$stdout;
    $fh->close;
  }

  $current_cmd = $cmd;		# Error() may want to report the command that failed
  $return_status = 0;		# error level will be increased (but not decreased) by Error() function

  ShellError 6, "System error: $!", @$stderr
    if $syscode != 0;
  ShellError 5, "Non-zero exit value $exitval.", @$stderr
    if $exitval != 0;
  ShellError 5, "Error message on stderr:", @$stderr
    if grep { /error|fail|abnormal|abort/i  } @$stderr;
  ShellError 3, "Warning on stderr:", @$stderr
    if grep { /warn|problem/i } @$stderr;
  ShellError 2, "Stderr output:", @$stderr
    if @$stderr;
  ShellError 1, "Error message on stdout:", @$stdout
    if grep { /error|fail|abnormal|abort/i } @$stdout;

  # return highest error status set by one of the previous commands
  return $return_status;
}


1;



__END__


=head1 NAME

UCS::File - File access utilities

=head1 SYNOPSIS

  use UCS::File;

  ## open filehandle for reading or writing
  # automagically compresses/decompresses files and dies on error
  $fh = UCS::File::Open("> my_file.gz");
  # the same without error checks (may return undefined value)
  $fh = UCS::File::TryOpen("> my_file.bz2");

  ## temporary file objects (disk files are automatically removed)
  $t1 = new UCS::File::Temp;	         # picks a unique filename
  $t2 = new UCS::File::Temp "mytemp";    # extends prefix to unique name
  $t3 = new UCS::File::Temp "mytemp.gz"; # compressed temporary file
  $filename = $t1->name;	# full pathname of temporary file
  $t1->write(...);		# works like $fh->print() ;
  $t1->finish;			# stop writing file
  print $t1->status, "\n";	# WRITING/FINISHED/READING/DELETED
  # main program can read or overwrite file <$filename> now
  $line = $t1->read;		# read one line (like $fh->getline())
  $t1->rewind;			# re-read from beginning of file
  $line = $t1->read;            # (reads first line again)
  $t1->close;			# stop reading and remove temporary file
  # other files will be removed when objects $t2 and $t3 are destroyed

  ## execute shell command with error detection
  $cmd = "ls -l";
  $errlevel = UCS::File::ShellCmd($cmd); # dies with error message if not ok
  $UCS::File::Paranoid = 1;	# more paranoid checks (-1 for less paranoid)
  # $errlevel == 0 (ok), 1 (minor problems), ..., 6 (fatal error)

  UCS::File::ShellCmd($cmd, \@lines);    # capture standard output in array
  UCS::File::ShellCmd($cmd, "file.txt"); # ... or in file (for large amounts of data)
  UCS::File::ShellCmd(["ls", "-l", @files], \@lines);  # bypass shell expansion

=head1 DESCRIPTION

This module provides some useful routines for handling files and
external programs.  This includes B<opening files> with error checks
and automagical compression/decompression, B<temporary file objects>
that are automatically created and deleted, and the execution of
B<shell commands> with extensive error checks.


=head1 OPENING FILES

=over 4

=item $fh = UCS::File::Open($name);

Open file I<$name> for reading, writing, or appending. Returns B<FileHandle>
object if successful, otherwise it B<die>s with an error message. It is thus
never necessary to check whether I<$fh> is defined.

If I<$name> starts with C<< > >>, the file is opened for writing (an
existing file will be overwritten). If I<$name> starts with
C<<< >> >>>, the file is opened for appending.

Files with the extensions C<.Z>, C<.gz>, and C<.bz2> are automagically 
compressed and decompressed, provided that the necessary tools are installed.
It is also possible to append to C<.gz> and C<.bz2> files.

Note that I<$name> may also be a read or write pipe (C<"... |"> or C<"| ...">,
respectively), which is passed directly to the built-in B<open> command.
It is thus subject to shell expansion and does not support automagic compression
and decompression.


=item $fh = UCS::File::TryOpen($name);

Same as B<UCS::File::Open>, but without the error checks. Returns B<undef>
if the C<open()> call fails.

=back


=head1 TEMPORARY FILES

Temporary files (implemented by B<UCS::>B<File::Temp> objects) are assigned 
a unique name and are automatically deleted when the script exits. The
life cycle of a temporary file consists of four stages: B<create>, 
B<write>, B<read> (possibly B<re-read>), B<delete>. This cycle corresponds
to the following method calls:

  $tf = new UCS::File::Temp; # create new temporary file in /tmp dir
  $tf->write(...);     # write cycle (buffered output, like print function)
  $tf->finish;         # complete write cycle (flushes buffer)
  $line = $tf->read;   # read cycle (like getline method for FileHandle)
 [$tf->rewind;         # optional: start re-reading temporary file ]
 [$line = $tf->read;                                               ]
  $tf->close;          # delete temporary file

Once the temporary file has been read from, it cannot be re-written; a
new B<UCS::File::Temp> object has to be created for the next cycle. When
the write stage is completed (but before reading has started, i.e. after
calling the B<finish> method), the temporary file can be accessed 
and/or overwritten by external programs. Use the B<name> method to
obtain its full pathname. If no direct access to the temporary file is
required, the B<finish> method is optional. The write cycle will
automatically be completed before the first B<read> method call.

=over 4

=item $tf = new UCS::File::Temp [ $prefix ];

Creates temporary file in F</tmp> directory. If the optional argument
I<$prefix> is specified, the filename will begin with I<$prefix> and
be extended to a unique name. If I<$prefix> contains a C</> character,
it is interpreted as an absolute or relative path, and the temporary
file will not be created in the F</tmp> directory. To create a
temporary file in the current working directory, use C<./MyPrefix>.

You can add the extension C<.Z>, C<.gz>, or C<.bz2> to I<$prefix> in
order to create a compressed temporary file. The actual filename (as
returned by the B<name> method) will have the same extension in this
case.

The temporary file is immediately created and opened for writing.

=item $filename = $tf->name;

Returns the real filename of the temporary file. B<NB:> direct access
to this file (e.g. by external programs) is only allowed after
calling B<finish>, and before the first B<read>.

=item $tf->write(...);

Write data to the temporary file. All arguments are passed to Perl's 
built-in B<print> function. Like B<print>, this method does not automatically
add newlines to its arguments.

=item $tf->finish;

Stop writing to the temporary file, flush the output buffer, and close
the associated file handle. Afer B<finish> has been called, the
temporary file can be accessed directly by the script or external
programs, and may also be overwritten.  In order to delete a file
created by an external program automatically, B<finish> the temporary
file immediately after its creation and then allow the external tool
to overwrite it:

  $tf = new UCS::File::Temp;
  $tf->finish;  # temporary file has size of 0 bytes now
  $filename = $tf->name;
  system "$my_shell_command > $filename";

=item $line = $tf->read;

Read one line from temporary file (same as calling B<getline> on a
B<FileHandle> object).  Automatically invokes B<finish> if called
during write cycle.

=item $tf->rewind;

Allows re-reading of the temporary file.  The next B<read> call will
return the first line of the temporary file.  Internally this is
achieved by closing and re-opening the associated file handle.

=item $tf->close;

Closes any open file handles and deletes the temporary file.  This will be done
automatically when the B<UCS::File::Temp> object is destroyed. Use B<close> to
free disk space immediately.

=back


=head1 SHELL COMMANDS

The B<UCS::File::ShellCmd> function provides a convenient replacement
for the built-in B<system> command.  Standard output and error
messages produced by the invoked shell command are captured to avoid
screen clutter.  The collected standard ouput of the command can
optionally be returned to the caller (similar to the backtick operator
C<`$shell_cmd`>).  B<UCS::File::ShellCmd> also checks for a variety of
error conditions and returns an error level ranging from 0
(successful) to 6 (fatal error):

  Error Level  Description
    6          command execution failed (system error)
    5          non-zero exit value or error message on STDERR
    4          -- reserved for future use --
    3          warning message on STDERR
    2          any output on STDERR
    1          error message on STDOUT

Depending on the value of I<$UCS::File::Paranoid> and the error level,
a warning message may be issued or the function may B<die> with an
error message.

=over 4

=item $UCS::File::Paranoid = 0;

With the default setting of 0, B<UCS::File::ShellCmd> will B<die> if
the error level is 5 or greater.  In the B<extra paranoid> setting
(+1), it will almost always B<die> (error level 2 or greater).  In the
B<less paranoid> setting (-1) only an error level of 6 (i.e. failure
to execute the shell command) will cause the script to abort.

=item $errlvl = UCS::File::ShellCmd($cmd);

=item $errlvl = UCS::File::ShellCmd($cmd, $filename);

=item $errlvl = UCS::File::ShellCmd($cmd, \@lines);

The first form executes I<$cmd> as a shell command (through the
built-in B<system> function) and returns an error level.  With the
default setting of I<$UCS::File::Paranoid>, serious errors are usually
detected and cause the script to B<die>, so it is not necessary to
check the value of I<$errlvl>.

The second form stores the standard output of the shell command in a
file named I<$filename>, where it can then be processed with external
programs or read in by the Perl script.  B<NB:> Compressed files are
not supported!  It is recommended to use an uncompressed temporary
file (B<UCS::File::Temp> object).

The third form takes an array reference as its second argument, splits
the standard output of I<$cmd> into B<chomp>ed lines and stores them
in the array I<@lines>.  If there is a large amount of standard ouput,
it is more efficient to use the second form.

Note that I<$cmd> is passed to the shell for metacharacter expansion.
In order to avoid this (e.g. when filename arguments may contain blanks),
specify an array reference of the form C<[$program, @args]> instead:

  $errlvl = UCS::File::ShellCmd(["ls", "-l", @files], \@lines);

=back 


=head1 COPYRIGHT

Copyright 2003 Stefan Evert.

This software is provided AS IS and the author makes no warranty as to
its use and performance. You may use the software, redistribute and
modify it under the same terms as Perl itself.

=cut








