## -*-cperl-*-
## Filename: UCS/R/RSperl.pm
## Modified: Wed Aug 17 23:45:32 2005 (severt)   
##   Author: Stefan Evert
##  Purpose: communicate with R backend interactively using Expect

package UCS::R::Expect;

use UCS;
use Expect;
use FileHandle;
use UCS::File;

use base Exporter;
@EXPORT = qw(Start Stop Exec LoadVector DumpVector);

$R = undef;					# Expect object for R server process
$R_Version = "";

# start R server process
sub Start ( ) {
  return 1 if defined $R;			# already running

  UCS::Die "Sorry, the R interpreter does not seem to be installed on your computer."
      unless defined $UCS::Rbin;

  UCS::Status "Spawning R server ... ";
  $R = Expect->spawn($UCS::Rbin, "--vanilla", "--no-readline"); # R must be in standard search path
  UCS::Die "Couldn't start R process in background.\n"
      unless defined $R;
  $R->log_stdout(0);

  my (undef, $error) =				# wait for version information
    $R->expect(15,
	       ["-re", 'Version\s+([0-9]+(\.[0-9]+)+)',
		sub { my $obj = shift; 
		      $R_Version = ($obj->exp_matchlist)[0];
		      exp_continue;
		    }
	       ],
	       "-re", '^>');
  UCS::Die "Couldn't start R process: $error\n"
    if defined $error;

  $R->clear_accum;
  print $R "options(digits=22, width=10)\n"; # maximal accuracy for floating-point value, print only 1 value per line (usually)
  (undef, $error) = $R->expect(5, "-re", '^>');
  UCS::Die "Couldn't setup R process: $error\n"
      if defined $error;
  print $R "source('$UCS::R_BaseDir/R/lib/ucs.R'); ucs.library(\"sfunc\")\n";
  (undef, $error) = $R->expect(5, "-re", '^>');
  UCS::Die "Couldn't load R libraries: $error\n"
      if defined $error;
  
  UCS::Status "R Version $R_Version running.\n";
  return 1;
}

END {
  # make sure to stop the server when the program exits or dies
  # (otherwise Expect will hang up when the object $R ist destroyed)
  Stop() if defined $R;
  # perhaps it's safer to store the server's PID and kill it here? ("kill 9, $pid;")
}

# stop R server process
sub Stop ( ) {
  return 1 unless defined $R;			# no server running
  
  UCS::Status "Stopping R server ... ";
  print $R "q()\n";
  # should check whether R has really exited and kill it if necessary, otherwise the next line will hang
  undef($R);
  $R_Version = "";

  UCS::Status "ok\n";
  return 1;
}

# @results = UCS::R::Exec("$cmd");
# execute R command and read result (which must be a numeric vector)
sub Exec ( $ ) {
  my $cmd = shift;
  if (defined $R and defined $R->exp_exitstatus) { # re-start R server if it has died or something
    undef $R;
  }
  Start() unless defined $R;			# auto-start R server when necessary
  $cmd =~ tr[\n][ ]s;				# command should be on single line
  print $R "$cmd\n";
  my @Result = ();
  if (defined wantarray) {	# try to parse and pass back return value of the command (must be a vector)
    $R->expect(60,
	       ["-re", qr/^\s*\[([0-9]+)\]\s*(.*?)\r?\n/,
		sub { my $obj = shift; 
		      my ($idx, $line) = $obj->exp_matchlist;
		      my @items = split " ", $line;
		      while (@items) {
			my $exp_idx = @Result + 1;
			UCS::Die "UCS::R: indexing problem with R output (expected $exp_idx, got $idx)"
			    unless $exp_idx == $idx;
			push @Result, shift @items;
			$idx++;
		      }
		      exp_continue;
		    }
	       ],
	       "-re", "^>");
  }
  else {			# void context: just wait for next prompt
    $R->expect(60, "-re", "^>");
  }
  UCS::Warn "UCS::R: communication error with R backend (".$R->error.")"
    if $R->error;
  return (wantarray) ? @Result : $Result[0];
}

# UCS::R::LoadVector($varname, \@data);
# load numeric vector into R through temporary file, using the scan() function
sub LoadVector ( $$ ) {
  my $var = shift;
  my $data = shift;
  my $l = scalar @$data;
  if ($l == 0) {
    Exec("$var <- numeric(0)");
  }
  else {
    my $tf = new UCS::File::Temp;
    for my $i (0 .. $l - 1) {
      $tf->write($data->[$i], "\n");
    }
    $tf->finish;
    my $filename = $tf->name;
    Exec("$var <- scan(file=\"$filename\", quiet=TRUE, sep=\"\\n\", n=$l)");
    $tf->close;
  }
  my $l2 = Exec("length($var)");
  UCS::Die "UCS::R::LoadVector(): load failed (sent $l entries, received $l2)"
      unless $l2 == $l;
}

# $arrayref = UCS::R::DumpVector($varname);
# read numeric vector from R through temporary file, using the write() function;
# the data are returned as an anonymous array reference for efficiency reasons
sub DumpVector ( $ ) {
  my $var = shift;
  my @data = ();
  my $l = Exec("length($var)");
  if ($l == 0) {
    return [];
  }
  else {
    my $tf = new UCS::File::Temp;
    $tf->finish;
    my $filename = $tf->name;
    Exec("write($var, file=\"$filename\", ncolumns=1)");
    $data[$l-1] = 0;				# pre-allocate array
    my $size = 0;
    while (my $line = $tf->read) {
      chomp $line;
      $data[$size] = $line;
      $size++;
    }
    $tf->close;
    UCS::Die "UCS::R::DumpVector(): dump failed (expected $l entries, got $size)"
	unless $size == $l;
    return \@data;
  }
}


1;


__END__


=head1 NAME

UCS::R::Expect - Expect-based implementation of R backend

=head1 SYNOPSIS

  use UCS::R::Expect;
  ## exports Start(), Stop() and Exec() functions into current namespace
  ## as well as LoadVector() and DumpVector()


=head1 DESCRIPTION

This module should I<only> be used implicitly through B<UCS::R>, which
loads the more efficient B<UCS::R::RSPerl> implementation if available, 
and falls back on B<UCS::R::Expect> otherwise.

=head1 LIMITATIONS

This module starts an R process in the background and communicates with it
interactively through the B<Expect> module.  This approach has several
disadvantages:

=over 4

=item *

Invoking R commands, waiting for output from the R backend, and parsing that
output causes substantial overhead for R function invocations, allowing less
than 1000 invocations per second even on a fast machine.

=item *

The return value of a function call has to be printed by R, then the resulting
output has to be parsed by Perl.  This interfacing method is rather frail and
currently supports only numeric vectors as return values.

=item *

The interface is extremely inefficient for exchanging large amounts of data
between Perl and R.  It may hang if the input/output buffers used by B<Expect>
overflow.  Use the B<LoadVector> and B<DumpVector> functions to pass large
numeric vectors to R and back.

=back

Because of these limitations, it is highly recommended that you install and
use the B<RSPerl> interface (available from L<http://www.omegahat.org/>) on
Unix platforms.  When RSPerl has been installed with support for calling R
from Perl, it will automatically be detected and configured for use by the UCS
installation script.  See F<doc/install.txt> for more information and
installation tips.

=head1 COPYRIGHT

Copyright (C) 2004-2005 Stefan Evert.

This software is provided AS IS and the author makes no warranty as to
its use and performance. You may use the software, redistribute and
modify it under the same terms as Perl itself.

=cut

