##
##  UCS / The R ToolBox
##

## Filename: base.R
## Modified: Thu Aug  5 16:16:27 2004 (evert)   
##   Author: Stefan Evert
##  Purpose: Basic functions and constants for handling UCS data sets



## builtin.AM   (built-in association measures with short descriptions)
## (builtin.AM is a vector of description strings with the corresponding AM names as keys

builtin.AM <- local({
  ucs.list.am <- paste(ucs.basedir, "bin", "ucs-list-am", sep="/") # use ucs-list-am tool to initialise list
  lines <- ucs.system(paste(ucs.list.am, "-x ALL"), intern=TRUE)
  l <- length(lines)
  if (l > 0 && ((l %% 2) == 0)) {
    res <- sapply(lines[seq(2,l,2)], function (t) substring(t,2))  # removes \t at start of descriptions
    names(res) <- lines[seq(1,l,2)]
  } else {
    warning("error running ucs-list-am command; can't generate list of built-in AMs")
    res <- NULL
  }
  res
})
  
## if (is.builtin.am("chi.squared")) { ... }
##   check whether the specified AM is found in the list of built-ins
is.builtin.am <- function (key) {
  ok <- key %in% names(builtin.AM)
  return(ok)
}

## keys <- builtin.ams()
##   returns list of the names of registered built-in AMs in UCS/Perl;
##   the functions below can be used to convert these into the corresponding data set variables,
##   but note that the conversion also works for AM names that are not registered in UCS/Perl
builtin.ams <- function () {
  names(builtin.AM)
}

## description <- am.key2desc("chi.squared")
##   returns short description (one-line string) for built-in AM
am.key2desc <- function (key) {
  builtin.AM[key]
}

## if (is.valid.key(key)) { ... }
## is.valid.key(key, warn=TRUE)
##   check whether key is a valid AM name, i.e. a UCS identifier;
##   if key is a vector, all elements must be valid AM names;
##   with warn=TRUE, a warning is issued listing all invalid keys
is.valid.key <- function (key, warn=FALSE) {
  ok <- grep("^[A-Za-z0-9.]+$", key, extended=TRUE)
  if (length(ok) < length(key)) {
    select <- if (length(ok) > 0) -ok else 1:length(key)
    if (warn) warning("invalid AM name(s): ", paste(key[select], collapse=", "))
    FALSE
  } else {
    TRUE
  }    
}

## am.var <- am.key2var("chi.squared" [, rank=TRUE])
## r.var <- am.key2rvar("chi.squared")
## key <- am.var2key("am.chi.squared")
##   convert AM names ("keys") into the corresponding data set variables for association scores
##   (rank=FALSE) or rankings (rank=TRUE), and vice versa (am.var2key works for both types);
##   am.key2rvar() is provided for backward compatibility
am.key2var <- function (key, rank=FALSE) {
  if (! is.valid.key(key, warn=TRUE)) stop("invalid AM name")
  paste( (if (rank) "r" else "am") , key, sep=".")
}
am.key2rvar <- function (key) {
  am.key2var(key, rank=TRUE)
}
am.var2key <- function (var) {
  as.vector(sapply(var, function (v) {
    if (is.valid.key(v)) {
      if (substring(v, 1, 3) == "am.") return(substring(v, 4))
      if (substring(v, 1, 2) == "r.") return(substring(v, 3))
    } 
    stop(v, " is not a valid association score variable")
  }))
}

## if (am.in.ds(glaw, "t.score" [, rank=TRUE])) { ... }
## am.in.ds(glaw, c("MI", "t.score"), fail=TRUE)
##   check whether data set contains scores for given AMs (or ranking if rank=TRUE);
##   aborts with error message if fail=TRUE and one or more AMs aren't found in the data set;
##   always aborts if one of the keys is not a valid AM name
am.in.ds <- function (ds, keys, rank=FALSE, fail=FALSE) {
  if (! is.valid.key(keys, warn=TRUE)) stop("invalid AM name")
  vars <- (if (rank) am.key2rvar(keys) else am.key2var(keys))
  matches  <- vars %in% names(ds)
  ok <- (sum(matches) == length(keys))
  if (fail && !ok) {
    stop("data set doesn't contain ",(if (rank) "ranking" else "scores")," for AM(s) ",paste(keys[!matches], collapse=", "))
  }
  return(ok)
}

## keys <- ds.find.am(ds [, rank=TRUE])
##   find AMs annotated in data set (association scores, or rankings if rank=TRUE);
##   returns bare AM names ("keys") without the prefix "am." or "r."
ds.find.am <- function (ds, rank=FALSE) {
  vars <- colnames(ds)
  if (rank) {
    idx <- substring(vars, 1, 2) == "r."
  } else {
    idx <- substring(vars, 1, 3) == "am."
  }
  return(am.var2key(vars[idx]))
}

## keys <- ds.match.am(ds, abbrevs [, rank=TRUE])
##   match full or partial AM keys in vector abbrevs against AM scores (or ranks) annotated in data set;
##   aborts with error message unless all abbrevs have a unique match; returns vector of same length otherwise
ds.match.am <- function (ds, abbrevs, rank=FALSE) {
  keys <- ds.find.am(ds, rank=rank)
  matches <- pmatch(abbrevs, keys, duplicates.ok=TRUE)
  errors <- is.na(matches)
  if (any(errors)) stop("the following AM keys don't match or have multiple matches: ", paste(abbrevs[errors], collapse=", "))
  keys[matches]
}

## glaw <- read.ds.gz("DataSet/glaw.ds.gz")
##   read (gzipped) data set (.ds or .ds.gz) into data frame;
##   compressed data set files are automatically decompressed, and comment lines are ignored;
##   the variables l1 and l2 are read as character vectors; b.* are converted to logicals and f.* to factors
read.ds.gz <- function (filename) {

  if ((! file.exists(filename)) && (length(grep('^[.]*[/]', filename)) == 0)) {        
    ## search data set in global directory tree if it isn't found in the current directory
    ## and doesn't include a full relative (./...) or absolute (/...) path
    files <- list.files(ucs.dataset.dir, full.names=TRUE, recursive=TRUE) # find all files in directory tree
    matches <- ucs.rmatch(paste("/",filename,sep=""), files)
    l <- length(matches)
    if (l > 0) {
      if (l > 1) {
        warning("Multiple matches for data set '", filename, "' in global directory tree.\n",
                "   ", l, " matching files found:\n",
                paste("     ", matches, collapse="\n", sep=""), "\n",
                "   Using first match.")
      }
      filename <- matches[1]
    } 
  }

  if (! file.exists(filename))
    stop ("Can't load data set '", filename, "', file doesn't exist.")

  tmp <- ucs.tempfile("UCS-R-TEMP.")
  on.exit(unlink(tmp))
  if (length(grep("\\.gz$", filename))) {
    ## uncompress gzipped data set to temporary file and remove comments
    cmd <-  paste("gzip -dc", filename, ">", tmp)
    ucs.system(cmd)
    filename <- tmp                     # now read temporary file
  }

  df <- read.delim(filename, quote="", as.is=c("l1", "l2"), comment.char="#")
  
  ## 'magic' data type conversions (might use colClasses=... in read.table() to enforce correct type directly)
  ## df$l1 <- as.character(df$l1)  # conversion to factor should have been avoided with as.is option above
  ## df$l2 <- as.character(df$l2)
  booleans <- grep("^b\\.", colnames(df)) # Boolean attributes in the data frame
  for (att in booleans) {
    df[att] <- as.logical(df[[att]])
  }
  
  ## return the data frame (invisibly)
  invisible(df)
}



## ordering <- order.by.am(glaw, "chi.squared" [, randomise=FALSE])
##   order data set by given AM (randomly breaking ties if randomise==TRUE (default),
##   using the variable am.random, which must be present);  returns row index vector
order.by.am <- function (ds, am, randomise=TRUE) {
  am.in.ds(ds, am, fail=TRUE)           # check that data set contains requested AM
  am.var <- am.key2var(am)              # get data set variable name for AM
  if (randomise) {
    am.in.ds(ds, "random", fail=TRUE) 
    rand.var <- am.key2var("random")    # use 'am.random' variable for breaking ties
    order(-ds[,am.var], ds[,rand.var])  # sort into descending order 
  }
  else {
    order(-ds[,am.var], ds[,"id"])      # sort rows with identical AM scores in original (i.e. ID) order
  }
}


## glaw <- add.ranks(glaw [, keys=c("MI", "MI2")] [, randomise=TRUE] [, overwrite=TRUE])
##   add rankings for given association measures (identified by their keys) to data set;
##   if the parameters keys is omitted, rankings will be computed for all registered AMs;
##   measures whose scores are not annotated in the data set are silently ignored,
##   and existing rankings will be overwritten; if randomise==TRUE (default), breaks
##   ties in the ranking randomly (using the variable am.random, which must be present);
##   if overwrite==FALSE, existing rankings won't be recomputed, which is useful when
##   adding association scores at a later time (computing them within R)
add.ranks <- function (ds, keys=ds.find.am(ds), randomise=TRUE, overwrite=TRUE) {
  n.rows <- nrow(ds)
  for (am in keys) {
    if (am.in.ds(ds, am)) {             # silently skip AMs not defined in data set
      if (overwrite || !am.in.ds(ds, am, rank=TRUE)) {
        cat(am, " ", sep="")
        ordering <- order.by.am(ds, am, randomise)      # list of row indices sorted by AM scores
        if (randomise) {
          sorted.ranks <- 1:n.rows      # with randomisation: unique ranks
        } else {
          am.var <- am.key2var(am)
          groups <- rle(ds[[am.var]][ordering]) # use run-length encoding to determine groups of candidates with equal scores
          groups$values <- (1 + c(0, cumsum(groups$lengths)))[1:length(groups$lengths)]  # compute lowest rank for each group
          sorted.ranks <- inverse.rle(groups)   # use run-length decoding to assign this rank to all group members
        }
        ranks <- vector(mode="numeric", length=n.rows)  # create dummy vector of appropriate length
        ranks[ordering] <- sorted.ranks  # this assigns the correct rank to each row, I believe
        am.rvar <- am.key2rvar(am)       # UCS variable name for ranking
        ds[[am.rvar]] <- ranks
      }
    }
  }
  cat("\n")
  invisible(ds)                         # don't forget to return modified data set :o)
}



## pr.tbl <- precision.recall(glaw, "chi.squared" [, tp=glaw$b.TP, step=1, first=1, cut=<none>, window=<none>])
## compute precision/recall table for specified AM; the data set must contain a
## ranking for the given AM and a user-defined variable "b.TP" marking TPs (unless tp is specified and non-null);
## optionally, increase size of n-best lists by <step> and skip lists with n < <first>;
## if <cut> is specified, compute n-best lists up to n=<cut> only, but pretend this
## corresponds to 100% of the candidate list (used for combined data from multiple n-best lists);
## in addition, local precision can be computed, averaging over a window of approx. <window>
## candidates around the current value of n (in the ranked candidate list)
##   returns data frame with columns
##     n         =  number of pairs in n-best list
##     perc      =  as percentage of full data set
##     tp        =  number of TPs in n-best list
##     fp        =  number of FPs in n-best list (note that tp + fp != n for random sample evaluation!)
##     precision =  precision value for this n-best list
##     recall    =  recall value for this n-best list (relative recall if <cut> was specified)
##     lp        =  local precision, averaged over symmetric window of specified size [only if requested]
precision.recall <- function (ds, am, tp=NULL, step=1, first=1, cut=0, window=0) {
  am.in.ds(ds, am, fail=TRUE)           # check whether requested AM exists in data set
  if (! am.in.ds(ds, am, rank=TRUE)) {  # ... and try to be helpful about ranks
    stop("You must add ranks to your data set first! (with the add.ranks() function)")
  }
  if (is.null(tp)) {                    # so one can also pass tp=NULL explicitly for default behavioue
    if (!("b.TP" %in% names(ds))) {     # ... be helpful about defining TPs, too :o)
      stop('You must define TPs for your data set first! (variable named "b.TP")')
    }
    tp <- ds$b.TP
  }
  if (length(tp) != nrow(ds)) stop("vector of TPs doesn't match data set (different lengths)")
  am.rvar <- am.key2rvar(am)
  order <- order(ds[[am.rvar]])         # re-order data set by AM's ranking
  if (cut > 0) {
    order <- order[1:cut]               # take only first <cut> elements of sig. list
  }
  tp <- tp[order]                       # (Boolean) ordered list of FPs and TPs
  tp.na <- is.na(tp)
  if (sum(tp.na) > 0) {                 # data set with missing annotations (NA's) -> random sample evaluation
    tp[tp.na] <- FALSE                  # replace NA's with FALSE, so they're not counted as TPs
  }
  size <- length(tp)                    # size of data set
  n <- 1:size                           # number of pair types in n-best list
  perc <- 100 * n / size                # corresponding percentage of full data set
  tp.count <- cumsum(tp)                # number of TPs in n-best list
  fp.count <- cumsum(! tp.na) - tp.count # number of FPs in n-best list
  n.count <- tp.count + fp.count        # nuber of annotated candidates in n-best list ( != n iff sum(tp.na) > 0)
  tp.total <- sum(tp)                   # total number of TPs in sample
  prec <- ifelse(n.count > 0,           # precision values (percentage)
                 100 * tp.count / n.count, 0) # set to zero if there are no annotated candidates in the n-best list
  recall <- 100 * tp.count / tp.total   # recall values (percentage)
  pr.rec <- data.frame(n=n,             # combine results into data frame
                       perc=perc, tp=tp.count, fp=fp.count, precision=prec, recall=recall)
  if (window > 0) {
    ## use kernel density estimate to compute local precision
    max.n <- 2 ^ ceiling(log(size, base=2)) # evaluate density estimates for n = 1:max.n (density() rounds up to power of 2!)
    bw <- window / 4                    # Gaussian kernel -> 95% window size is approx. 4 * bandwidth (= s.d.)
    x <- which(tp)                      # "coordinates" of true positives 
    ## use mirrored data to improve non-zero density estimate near end of support (cf. MASS, p.139)
    res <- density(c(-rev(x)+1, x), bw=bw, from=-max.n+1, to=max.n, n=2*max.n)
    d <- res$y                          # density estimates for n = -max.n+1 .. max.n
    ## check that the specified number of points is returned, so that their x-coordinates correspond to integers -max.n+1 .. max.n
    if (length(d) != 2 * max.n) stop("kernel density estimation returned wrong number of points")
    ## we need to re-norm density because of reflection method (so density for 1 .. max.n should some to 0.5)
    ## we may also have lost a small amount of probability mass in the range n+1 .. max.n
    ##  cat("Renorm:", sum(d[n+max.n]), "\n")     # check: should be close to 0.5
    tp.density <- d[n+max.n] / sum(d[n+max.n])    # retrieve density for n = 1:size and re-norm
    ## the following line scales density to average precision values, taking random-sample evaluation into account
    pr.rec$lp <- 100 * tp.density * size * tp.total / n.count[size] # scale to precision values (integral over tp.density == 1)
  }
  if ((step > 1) || (first > 1)) {
    subset <- rev(seq(size, first, by=-step)) # construct index seq. from <first> exactly up to <size> by <step>
    invisible(pr.rec[subset,])
  } else {
    invisible(pr.rec)                     # don't print automatically, of course
  }
}


## evaluation.table(glaw, keys=c("t.score", "chi.squared"), N=c(500,1000,2000) [, tp=glaw$b.TP, recall=TRUE])
## the "text version" of evaluation.plot(), can be used to create precision and recall tables for various
## association measures and n-best lists (might add confidence intervals etc. later)
evaluation.table <- function (ds, keys, N, tp=NULL, recall=FALSE) {
  ## check whether requested AMs are annotated and expand abbreviations
  keys <- ds.match.am(ds, keys, rank=TRUE) # stop()s unless all keys have unique matches
  n.keys <- length(keys)                # number of AMs in table

  ## check vector of TPs and set default if unspecified
  if (is.null(tp)) {                    # so one can also pass tp=NULL explicitly for default behavioue
    if (!("b.TP" %in% names(ds))) {
      stop('You must define TPs for your data set first! (variable named "b.TP")')
    }
    tp <- ds$b.TP
  }
  if (length(tp) != nrow(ds)) stop("vector of TPs doesn't match data set (different lengths)")

  ## create table of precision or recall values
  tbl <- data.frame(n=N)
  for (i in 1:n.keys) {
    pr <- precision.recall(ds, keys[i], tp=tp)
    tbl[[ keys[i] ]] <- if (recall) pr$recall[N] else pr$prec[N]
  }

  return(tbl)
}
