##
##  UCS / The R ToolBox
##

## Filename: eo.R
## Modified: Fri Oct 21 20:53:22 2005 (severt)   
##   Author: Stefan Evert
##  Purpose: Visualisation of data sets and GAMs in the e-o plane

ucs.library("base")
ucs.library("gam")


## EO plot configuration data
.eo.PAR <-
  list(
       bw=FALSE,                        # whether to generate b/w plots by default
       xlim=NULL,                       # default range on x-axis (in orders of magnitude)
       ylim=NULL,                       # default range on y-axis (in orders of magnitude)
       aspect=TRUE,                     # whether to enfore 1:1 aspect (for square plotting region)
       log.marks=FALSE,                 # whether labels on tick marks are in logarithmic units
       steps=100,                       # number of steps for drawing iso-lines
       jitter=FALSE,                    # whether to use jittering for points and n-best lines
       cex=1.3,                         # overall character expansion (for labels and tick marks)
       col=c("black", "blue", "red", "#AA00AA", "#006699"),
       lty="solid",
       lwd=4,
       solid=c("#AAAAAA", "#8888FF", "#FF8888", "#FF88FF", "#88DDFF"),
       angle=c(90, 45, 135, 0, 110),
       density=10,
       bw.col=c("black", "black", "#888888", "black", "#444444"),
       bw.lty=c("solid", "33", "solid", "11", "dotdash"),
       bw.lwd=c(3, 3, 3, 4, 4),
       bw.solid="#AAAAAA",
       bw.angle=c(45, 135, 90, 0, 110),
       bw.density=10,
       pt.pch=20,
       pt.cex=0.5,
       pt.col=c("black", "#00BB00", "#DD0000", "#DDDD00", "#FF9900"),
       bw.pt.pch=c(20, 1, 3, 2, 4),
       bw.pt.cex=c(0.3, 0.4, 0.5, 0.4, 0.5),
       bw.pt.col="black",
       )

## EO plot normalized style data (style vectors expanded to same length)
.eo.STYLE <- list()

## EO plot internal state (dimensions and settings of current plot)
.eo.STATE <- list()

## eo.par("n.step")
## eo.par(bw=TRUE, col=c(...))
##   functional interface to EO plot graphics parameters
eo.par <- function (...) {
  args <- list(...)
  if (! length(args)) {                 # eo.par()
    return(.eo.PAR)
  }
  else if (is.null(names(args))) {      
    if (length(args) == 1 && is.list(args[[1]])) {  # eo.par(list(col=c(...), lty=c(...))), used to reset saved values
      args <- args[[1]]                 # then fall through to default case below
    } else {                            # eo.par("col", "lty") or eo.par(c("col", "lty"))
      pars <- as.character(unlist(args))
      return( if (length(pars) == 1) .eo.PAR[[ pars[1] ]] else .eo.PAR[pars] )
    }
  }
  known.pars <- names(.eo.PAR)          # eo.par(bw=TRUE, col=c(...), lty=c(...), ...)
  invalid <- !(names(args) %in% known.pars)
  if (any(invalid)) {
    warning("invalid graphics parameter(s) ", paste(names(args)[invalid], collapse=", "), " for EO plots ignored")
    args <- args[! invalid]
  }
  old <- .eo.PAR[names(args)]
  for (i in names(args)) { .eo.PAR[[i]] <<- args[[i]] }
  check.eo.pars()
  return(invisible(old))
}

## check.eo.pars()
##   used internally to ensure that .eo.PAR settings are sensible and to update .eo.STYLE
check.eo.pars <- function ( ) {
  max.style <- max(length(.eo.PAR$col), length(.eo.PAR$lty), length(.eo.PAR$lwd))
  .eo.STYLE$col <<- rep(.eo.PAR$col, length=max.style)
  .eo.STYLE$lty <<- rep(.eo.PAR$lty, length=max.style)
  .eo.STYLE$lwd <<- rep(.eo.PAR$lwd, length=max.style)
  .eo.STYLE$solid <<- rep(.eo.PAR$solid, length=max.style)
  .eo.STYLE$angle <<- rep(.eo.PAR$angle, length=max.style)
  .eo.STYLE$density <<- rep(.eo.PAR$density, length=max.style)
  .eo.STYLE$max.style <<- max.style             # largest valid style number for iso lines (colour plots)
  bw.max.style <- max(length(.eo.PAR$bw.col), length(.eo.PAR$bw.lty), length(.eo.PAR$bw.lwd))
  .eo.STYLE$bw.col <<- rep(.eo.PAR$bw.col, length=bw.max.style)
  .eo.STYLE$bw.lty <<- rep(.eo.PAR$bw.lty, length=bw.max.style)
  .eo.STYLE$bw.lwd <<- rep(.eo.PAR$bw.lwd, length=bw.max.style)
  .eo.STYLE$bw.solid <<- rep(.eo.PAR$bw.solid, length=bw.max.style)
  .eo.STYLE$bw.angle <<- rep(.eo.PAR$bw.angle, length=bw.max.style)
  .eo.STYLE$bw.density <<- rep(.eo.PAR$bw.density, length=bw.max.style)
  .eo.STYLE$bw.max.style <<- bw.max.style       # largest valid style number for iso lines (B/W plots)
  pt.max.style <- max(length(.eo.PAR$pt.pch), length(.eo.PAR$pt.cex), length(.eo.PAR$pt.col))
  .eo.STYLE$pt.pch <<- rep(.eo.PAR$pt.pch, length=pt.max.style)
  .eo.STYLE$pt.cex <<- rep(.eo.PAR$pt.cex, length=pt.max.style)
  .eo.STYLE$pt.col <<- rep(.eo.PAR$pt.col, length=pt.max.style)
  .eo.STYLE$pt.max.style <<- pt.max.style       # largest valid style number for data set points (colour plots)
  bw.pt.max.style <- max(length(.eo.PAR$bw.pt.pch), length(.eo.PAR$bw.pt.cex), length(.eo.PAR$bw.pt.col))
  .eo.STYLE$bw.pt.pch <<- rep(.eo.PAR$bw.pt.pch, length=bw.pt.max.style)
  .eo.STYLE$bw.pt.cex <<- rep(.eo.PAR$bw.pt.cex, length=bw.pt.max.style)
  .eo.STYLE$bw.pt.col <<- rep(.eo.PAR$bw.pt.col, length=bw.pt.max.style)
  .eo.STYLE$bw.pt.max.style <<- bw.pt.max.style # largest valid style number for data set points (B/W plots)
  ## might add some sanity checks here
}
check.eo.pars()                         # call now to initialize .eo.STYLE

## retrieve style options, either b/w or colour (depending on internal state)
eo.style <- function (option, style, bw=.eo.STATE$bw) {
  max.name <- if (substr(option, 1, 3) == "pt.") "pt.max.style" else "max.style"
  if (bw) max.name <- paste("bw", max.name, sep=".")
  max.style <- .eo.STYLE[[max.name]]
  if (any(style < 1 | style > max.style)) stop ("style number(s) must be in range 1 .. ", max.style)
  if (bw) option <- paste("bw", option, sep=".")
  style.vec <- .eo.STYLE[[option]]
  return(style.vec[style])
}

## internal: check whether there is an active eo plot
eo.is.open <- function (fail=FALSE) {
  if (all(c("xlim", "ylim", "bw", "par.save") %in% names(.eo.STATE))) {
    return(TRUE)
  } else {
    if (fail) stop("no active (e,o) plot or setup was unsuccessful")
    return(FALSE)
  }
}

## internal: compute suitable range for axis (rounded to full orders of magnitude)
eo.fit.axis <- function (x) {
  x.min <- min(x)
  if (x.min <= 0) stop ("all data points must be positive for logarithmic plot")
  x.max <- max(x)
  return(c(floor(log10(x.min)), ceiling(log10(x.max))))
}

## set up plot of (e,o)-plane (with ranges as specified or computed from data set)
eo.setup <- function (xlim=eo.par("xlim"), ylim=eo.par("ylim"), ds=NULL, bw=eo.par("bw"), file=NULL,
                      aspect=eo.par("aspect"), log.marks=eo.par("log.marks"), cex=eo.par("cex"), ...) {
  if (eo.is.open()) eo.close()          # automatically closes active plot
  if (! missing(ds)) { # data set overrides eo.par() defaults but not explicitly specified limits
    if (missing(xlim)) xlim <- eo.fit.axis(ds$f1 * ds$f2 / ds$N)
    if (missing(ylim)) ylim <- eo.fit.axis(ds$f)
  }
  if (is.null(xlim)) stop("you must specify either xlim or ds argument") # check that axis ranges are specified in some way
  if (is.null(ylim)) stop("you must specify either ylim or ds argument")
  if (aspect) {
    x.range <- diff(xlim)                 # ensure 1:1 aspect ratio, i.e. same ranges on both axes
    y.range <- diff(ylim)
    range <- max(x.range, y.range)
    if (x.range < range) xlim <- xlim + c(-.5,.5) * (range - x.range)
    if (y.range < range) ylim <- ylim + c(-.5,.5) * (range - y.range)
  }
  .eo.STATE$xlim <<- xlim               # record dimensions of the new plot in internal state variable
  .eo.STATE$ylim <<- ylim
  .eo.STATE$bw <<- bw
  .eo.STATE$file <<- FALSE
  if (! missing(file)) {                # save plot to postscript file
    postscript(file=file, width=6, height=6, paper="special", horizontal=FALSE, onefile=FALSE)
    .eo.STATE$file <<- TRUE
  } 
  .eo.STATE$par.save <<- par(mar=c(4,4,1,1)+.1, cex=cex)    # set plot margins so the actual drawing region is square
  xlab <- if (log.marks) "E (magnitude)" else "E"
  ylab <- if (log.marks) "O (magnitude)" else "O"
  plot(xlim, ylim, xlim=xlim, ylim=ylim, type="n", axes=FALSE, xlab=xlab, ylab=ylab, ...) # set up plot
  box()                                 # draw box with computed tick marks
  x.ticks <- ceiling(xlim[1]):floor(xlim[2]) # may have been expanded to non-integer values
  y.ticks <- ceiling(ylim[1]):floor(ylim[2])
  x.labels <- as.character(if (log.marks) x.ticks else 10^x.ticks)
  y.labels <- as.character(if (log.marks) y.ticks else 10^y.ticks)
  axis(1, at=x.ticks, labels=x.labels)
  axis(2, at=y.ticks, labels=y.labels)
}

## draw iso-line for specified gamma (or n-best line), and optionally fill in acceptance region
eo.iso <- function (gam, gamma=0, b=1, N=1e6, n.best=NULL, ds=NULL, style=1, fill=solid, solid=FALSE,
                    steps=eo.par("steps"), jitter=eo.par("jitter"), bw=.eo.STATE$bw,
                    col=eo.style("col",style,bw), lty=eo.style("lty",style,bw), lwd=eo.style("lwd",style,bw),
                    angle=eo.style("angle",style,bw), density=eo.style("density",style,bw), solid.col=eo.style("solid",style,bw)) {
  eo.is.open(fail=TRUE)
  if (! missing(n.best)) {              # compute n-best threshold from data set (overrides any value for gamma)
    if (missing(ds)) stop ("need data set (ds) to compute n-best threshold")
    gamma <- gamma.nbest(ds, gam, n.best, jitter=jitter)
  }
  xlim <- .eo.STATE$xlim
  ylim <- .eo.STATE$ylim
  e.log <- seq(xlim[1]-.5, xlim[2]+.5, length=steps)
  o.log <- log10(gam.iso(gam, gamma, e=10^e.log, b=b, N=N))
  if (fill) {
    y.max <- ylim[2] + .5
    if (solid) {
      polygon(c(e.log, max(e.log), min(e.log)), c(o.log, y.max, y.max), col=solid.col, border=NA)
    } else {
      polygon(c(e.log, max(e.log), min(e.log)), c(o.log, y.max, y.max), col=col, density=density, angle=angle, border=NA)
    }
  }
  lines(e.log, o.log, col=col, lwd=lwd, lty=lty)
}

## highlight differences between acceptance regions for two measures ("t", "X2", "G2")
eo.iso.diff <- function (gam1, gam2, gamma1=0, gamma2=0, b=1, N=1e6, n.best1=NULL, n.best2=NULL, ds=NULL, 
                         style1=4, style2=5, solid=FALSE, bw=.eo.STATE$bw,
                         steps=eo.par("steps"), jitter=eo.par("jitter"), 
                         col1=eo.style("col",style1,bw), angle1=eo.style("angle",style1,bw),
                         density1=eo.style("density",style1,bw), solid.col1=eo.style("solid",style1,bw),
                         col2=eo.style("col",style2,bw), angle2=eo.style("angle",style2,bw),
                         density2=eo.style("density",style2,bw), solid.col2=eo.style("solid",style2,bw)) {
  eo.is.open(fail=TRUE)
  if (! (missing(n.best1) && missing(n.best2))) { # compute n-best threshold from data set (overrides any value for gamma)
    if (missing(ds)) stop ("need data set (ds) to compute n-best threshold")
    if (! missing(n.best1)) gamma1 <- gamma.nbest(ds, gam1, n.best1, jitter=jitter)
    if (! missing(n.best2)) gamma2 <- gamma.nbest(ds, gam2, n.best2, jitter=jitter)
  }
  xlim <- .eo.STATE$xlim                # compute equidistant steps for e and the corresponding iso-values for o
  ylim <- .eo.STATE$ylim
  e.log <- seq(xlim[1]-.5, xlim[2]+.5, length=steps)
  o1.log <- log10(gam.iso(gam1, gamma1, e=10^e.log, b=b, N=N))
  o2.log <- log10(gam.iso(gam2, gamma2, e=10^e.log, b=b, N=N))
  plus <- o1.log >= o2.log
  minus <- o1.log <= o2.log
  if (any(minus)) {
    x <- c(e.log[minus], rev(e.log[minus]))
    y <- c(o1.log[minus], rev(o2.log[minus]))
    if (solid) {
      polygon(x, y, col=solid.col1, border=NA)
    }
    else {
      polygon(x, y, col=col1, density=density1, angle=angle1, border=NA)
    }
  }
  if (any(plus)) {
    x <- c(e.log[plus], rev(e.log[plus]))
    y <- c(o2.log[plus], rev(o1.log[plus]))
    if (solid) {
      polygon(x, y, col=solid.col2, border=NA)
    }
    else {
      polygon(x, y, col=col2, density=density2, angle=angle2, border=NA)
    }
  }
}

## draw (selected) points from data set, with optional jittering
eo.points <- function (ds, style=1, select=NULL, bw=.eo.STATE$bw, jitter=eo.par("jitter"),
                       pch=eo.style("pt.pch",style,bw), cex=eo.style("pt.cex",style,bw), col=eo.style("pt.col",style,bw),
                       ...) {
  eo.is.open(fail=TRUE)
  if (! missing(select)) {
    idx <- eval(substitute(select), envir=ds)
  } else {
    idx <- 1:nrow(ds)
  }
  f  <- ds$f[idx]
  f1 <- ds$f1[idx]
  f2 <- ds$f2[idx]
  N  <- ds$N[idx]
  ## if style is a vector, it has to be recycled and subsetted in parallel to the other vectors
  ## (note that the defaults for pch, cex and col should be computed lazily and use the new value of "style")
  style <- rep(style, length=nrow(ds))[idx]
  ## when pch, cex or col are set explicitly to a vector, the length of this vector has to be adjusted
  ## manually by the caller -- use of the style argument is highly encouraged!
  if (jitter) {
    has.jitter(ds, fail=TRUE)
    f <-  f  + ds$x.jitter.f[idx]
    f1 <- f1 + ds$x.jitter.f1[idx]
    f2 <- f2 + ds$x.jitter.f2[idx]
  }
  o <- f
  e <- f1 * f2 / N
  points(log10(e), log10(o), pch=pch, cex=cex, col=col, ...)
}

## internal: draw legend box in specified corner (1=top left, 2=top right, 3=bottom right, 4=bottom left)
eo.legend.corner <- function (corner, ...) {
  xlim <- .eo.STATE$xlim
  ylim <- .eo.STATE$ylim
  offset.x <- switch(corner, 0.00, 1.00, 1.00, 0.00) # use these offsets for fine-tuning the position of legend boxes
  offset.y <- switch(corner, 1.00, 1.00, 0.00, 0.00)
  x <- xlim[1] + diff(xlim) * offset.x
  y <- ylim[1] + diff(ylim) * offset.y
  xjust <- switch(corner, 0, 1, 1, 0)
  yjust <- switch(corner, 1, 1, 0, 0)
  legend(x=x, y=y, xjust=xjust, yjust=yjust, bg="white", ...)
}

## draw a legend box listing the names of AMs for which iso-lines are shown (with standard colors)
eo.legend.iso <- function (corner, legend, styles, bw=.eo.STATE$bw, fill=solid, solid=FALSE,
                           lwd.add=0, density.mul=2, ...) {
  eo.is.open(fail=TRUE)
  if (length(legend) != length(styles)) stop("number of styles and labels must be the same")
  col <- eo.style("col", styles, bw)
  lty <- eo.style("lty", styles, bw)
  lwd <- eo.style("lwd", styles, bw) + lwd.add
  if (fill) {
    if (solid) {
      solid.bg <- eo.style("solid", styles, bw)
      eo.legend.corner(corner, legend=legend, fill=solid.bg, ...)
    } else {
      angle <- eo.style("angle", styles, bw)
      density <- eo.style("density", styles, bw)
      eo.legend.corner(corner, legend=legend, fill=col, angle=angle, density=density * density.mul, ...)
    }
  } else {
    eo.legend.corner(corner, legend=legend, col=col, lwd=lwd, lty=lty, ...)
  }
}

## draw a legend box for (different types of) points
eo.legend.points <- function (corner, legend, styles, bw=.eo.STATE$bw, cex.mul=2.5, mark=FALSE, ...) {
  eo.is.open(fail=TRUE)
  if (length(legend) != length(styles)) stop("number of styles and labels must be the same")
  pch <- eo.style("pt.pch", styles, bw)
  cex <- eo.style("pt.cex", styles, bw)
  col <- eo.style("pt.col", styles, bw)
  if (mark) {
    pch <- rep(1, length(pch))
    cex <- 1.3 * cex                    # make circle symbol slightly larger
    eo.legend.corner(corner, legend=legend, pch=pch, col=col, pt.cex=cex * cex.mul, pt.lwd=4, ...)
  } else {
    eo.legend.corner(corner, legend=legend, pch=pch, col=col, pt.cex=cex * cex.mul, ...)
  }
}

## draw a legend box for difference regions
eo.legend.diff <- function (corner, legend, style1=4, style2=5, bw=.eo.STATE$bw, solid=FALSE, density.mul=2, ...) {
  eo.is.open(fail=TRUE)
  if (length(legend) != 2) stop("legend must contain exactly 2 entries for eo.legend.diff()")
  eo.legend.iso(corner, legend, c(style1,style2), bw=bw, fill=TRUE, solid=solid, density.mul=density.mul, ...)
}

## mark specified data set points in eo plot
eo.mark <- function (ds, select, style=1, bw=.eo.STATE$bw, cex=1.5, lwd=3, jitter=eo.par("jitter")) {
  eo.is.open(fail=TRUE)
  idx <- eval(substitute(select), envir=ds)
  eo.points(ds[idx,], style=style, jitter=jitter, bw=bw, pch=1, cex=cex, lwd=lwd)
}

## terminate (e,o) plot, closing postscript device when writing to a file
eo.close <- function () {
  if (eo.is.open()) {
    box()           # box around plot may have been overwritten
  } else {
    warning("no active (e,o) plot or setup was unsuccessful")
  }
  vars <- names(.eo.STATE)          # try to close file / reset graphics parameters even when eo.setup() failed
  if ("par.save" %in% vars) par(.eo.STATE$par.save) # reset graphics parameters ("mar")
  if ("file" %in% vars && .eo.STATE$file) dev.off()
  .eo.STATE <<- list()                  # reset internal state
}
