##
##  UCS / The R ToolBox
##

## Filename: fzm.R
## Modified: Sun May 15 13:55:47 2005 (evert)   
##   Author: Stefan Evert
##  Purpose: The finite ZipF-Mandelbrot LNRE model described in (Evert 2004)

## References:
## Evert, Stefan (2004).  A Simple LNRE Model for Random Character Sequences.
##   In: Proceedings of the JADT 2004 Conference, Louvain-la-Neuve, Belgium.


ucs.library("sfunc")
ucs.library("lexstats")

## use logit transform from the "boot" library
## (could also easily implement logit() and inv.logit() in R)
library("boot")                         


## create object representing finite Zipf-Mandelbrot model;
## the constructor can be invoked in several different ways:
##   model <- fzm(alpha=0.8, A=1e-20, B=0.001)         ...  specify parameters explicitly
##   model <- fzm(alpha=0.8, A=1e-20, N=1e6, V=77800)  ...  estimate B from number V of observed types
##   model <- fzm(alpha=0.8, spc=c(V1,V2,...), N=1e6, V=77800)  ...  estimate A and B from observed spectrum (and V)
##   model <- fzm(spc=c(V1,V2,...), N=1e6, V=77800)    ...  estimate all parameters from observed spectrum (and V)
## (N, V, and spc can be specified in addition to explicit parameters and
## will be stored in the returne object for use by the write.lexstats() function)
fzm <- function (alpha=NULL, A=NULL, B=NULL, N=NULL, V=NULL, spc=NULL, m.max=15, stepmax=10, iterlim=100, debug=FALSE) {
  model <- list()
  obs.ok <- (!missing(N)) && (!missing(V))  # observed data available? (N and V)
  spc.ok <- !missing(spc)               # spectrum available (spc=c(V1,V2,...), for sample size N)
  if (obs.ok) {
    model$N <- N
    model$V <- V
    if (spc.ok) model$spc <- spc
  }
  ## if either alpha or A is not specified, estimate them from observed data
  if (missing(alpha) || missing(A)) {
    if (spc.ok && obs.ok) {
      ## perform parameter optimisation, using up to m.max spectrum elements
      ## (but stop before the first rank m with Vm < 100)
      m.max <- min(length(spc), m.max)
      spc.exclude <- (1:m.max)[spc[1:m.max] < 100]
      if (length(spc.exclude) >= 1) m.max <- min(spc.exclude) - 1
      if (m.max < 2) m.max <- 2         # minimum number of spectrum elements for parameter estimation
      ## Cost function for non-linear minimsation (multinomial chi-squared statistic):
      ## optimisation parameters alpha' and A' are logit-transformed to the scale [-inf, inf]
      ## pre-set parameter values are not transformed (passed as optional arguments to the cost() function)
      cost <- function (params, N, V, Vm, m.max, debug, alpha=NULL, A=NULL) {
        if (missing(alpha)) {           # params are optimisation parameters, alpha and A optional pre-set values
          alpha <- inv.logit(params[1])
          if (missing(A)) A <- inv.logit(params[2])
        } else { # (missing(A))
          A <- inv.logit(params[1])
        }
        if (alpha <= 0 || alpha >= 1 || A <= 0 || A >= 1) { # return +infinity when either parameter is out of range
          if (debug) cat("NLM:  alpha =", alpha, " A =", A, "  (out of range)\n")
          return(Inf)
        }
        temp <- fzm(alpha=alpha, A=A, N=N, V=V) # fZM model with specified alpha and A (B estimated such that E[V]==V)
        E.Vm <- EVm(temp, 1:m.max, N )  # compute chi-squared statistic (Baayen 2001, 118)
        E.V <- EV(temp, N)
        Vplus <- V - sum(Vm)
        E.Vplus <- E.V - sum(E.Vm)
        X2 <- sum((Vm - E.Vm)^2 / E.Vm) + (Vplus - E.Vplus)^2 / E.Vplus
        if (debug) cat("NLM:  alpha =", alpha, " A =", A, "  X2 =", X2, "\n")
        return(X2)
      }
      ## perform minimisation of cost function, and use optimised parameter estimate if successful
      if (missing(alpha) && missing(A)) {
        alpha <- spc[1] / V             # direct estimate from infinite ZM model
        A <- 1e-6 / N                   # start with A well below smallest relative frequency in sample
        params <- c(logit(alpha), logit(A)) # initial estimates for (alpha', A')
        res <- nlm(cost, params, print.level=debug, N=N, V=V, Vm=spc[1:m.max], m.max=m.max, stepmax=stepmax, iterlim=iterlim, debug=debug)
        if (res$code <= 4) {            # don't reject result if iteration limit is exceeded
          if (res$code == 4) warning("iteration limit exceeded (using parameters from last step)") 
          alpha <- inv.logit(res$estimate[1])
          A <-  inv.logit(res$estimate[2])
        } else stop("Parameter optimisation failed.")
      } else if (missing(alpha)) {
        alpha <- spc[1] / V             # direct estimate from infinite ZM model
        params <- logit(alpha)
        res <- nlm(cost, params, print.level=debug, N=N, V=V, Vm=spc[1:m.max], m.max=m.max, stepmax=stepmax, iterlim=iterlim, debug=debug, A=A)
        if (res$code <= 2) {
          alpha <- inv.logit(res$estimate[1])
        } else stop("Parameter optimisation failed.")
      } else { # (missing(A))
        A <- 1e-6 / N
        params <- logit(A)
        res <- nlm(cost, params, print.level=debug, N=N, V=V, Vm=spc[1:m.max], m.max=m.max, stepmax=stepmax, iterlim=iterlim, debug=debug, alpha=alpha)
        if (res$code <= 2) {
          A <-  inv.logit(res$estimate[1])
        } else stop("Parameter optimisation failed.")
      }
      if ((alpha <= 0) || (alpha >= 1)) stop("Sorry, automatic estimation of alpha parameter failed (out of valid range).")
    } else {
      stop("You have to specify N, V, and spc for automatic estimation of the alpha parameter.")
    }
  } else {
    if ((alpha <= 0) || (alpha >= 1)) stop("The parameter alpha (",alpha,") must be in the range 0 < alpha < 1.")
    if ((A <= 0) || (A >= 1)) stop("The parameter A (",A,") must be in the range 0 < A < 1.")
  }
  model$alpha <- alpha
  model$A <- A
  if (missing(B)) {
    if (obs.ok) {
      term1 <- (N ^ alpha) * Igamma(1 - alpha, N * A, lower=FALSE)
      term2 <- (A ^ (-alpha)) * (1 - exp(-N * A))
      C <- alpha * V / (term1 + term2)
    } else {
      stop("You have to specify N and V for automatic estimation of the B parameter.") 
    }
    B <- ( (1 - alpha) / C + A ^ (1 - alpha) ) ^ (1 / (1 - alpha))
  } else {
    C <-  (1 - alpha) / ( B ^ (1 - alpha) - A ^ (1 - alpha) )
  }
  if (B <= A) stop("The parameter B (",B,") must be greater than A (",A,")")
  model$B <- B
  model$C <- C
  model$S <- C / alpha * ((A ^ -alpha) - (B ^ -alpha)) # predicted population size
  model$n.param <- 3                    # number of estimated parameters (for multivariate chi-squared test)
  class(model) <- "fzm"                 # bless into "fzm" class
  return(model)
}

# print summary of model parameters and observed data stored in the model
# with "print(model)" or simply by typing "model" in an R session 
print.fzm <- function (model) {
  cat("\nfinite Zipf-Mandelbrot LNRE model.\n")
  cat("Parameters:  alpha =", model$alpha, " A =", model$A, " B =", model$B, " [ C =", model$C, "]\n")
  alpha <- model$alpha
  cat("Vocabulary:  S =", model$S, "\n")
  if ((!is.null(model$N)) && (!is.null(model$V))) {
    cat("Observed:  N =", model$N, "   V  =", model$V)
    if (!is.null(model$spc)) {
      cat("     Vm  = ")
      if (length(model$spc) > 5) {
        cat(model$spc[1:5], "...")
      } else {
        cat(model$spc)
      }
    }
    cat("\n")
    cat("Expected:  N =", model$N, " E[V] =", EV(model, model$N))
    cat("   E[Vm] =", EVm(model, 1:5, model$N), "...\n")
  } else {
    cat("(No observed data available.)\n")
  }
  cat("\n")
}

## expected spectrum and conditional parameter distribution of finite Zipf-Mandelbrot LNRE model, plus variances
##   E[V(N)]                      = EV(model, N)
##   E[V_m(N)]                    = EVm(model, m, N)
##   E[V_m(N)] / E[V(N)]          = EVm(model, m, N, relative=TRUE)
##   E[V_{m+1}(N)] / E[V_m(N)]    = EVm(model, m, N, ratio=TRUE)
##   E[V_{m, rho}(N)]             = EVm(model, m, N, rho=<rho>) 
##   E[V_{m, rho}(N)] / E[V_m(N)] = EVm(model, m, N, rho=<rho>, relative=TRUE) 
##   E[V_{m, >rho}(N)]            = EVm(model, m, N, rho=<rho>, lower=FALSE) 
##   E[V_{m, >rho}(N)] / E[V_m(N)]= EVm(model, m, N, rho=<rho>, relative=TRUE, lower=FALSE) 
EV.fzm <- function (model, N) {
  alpha <- model$alpha
  A <- model$A
  C <- model$C
  term1 <- (N ^ alpha) * Igamma(1 - alpha, N * A, lower=FALSE)
  term2 <- (A ^ (-alpha)) * (1 - exp(-N * A))
  return( (C / alpha) * (term1 + term2) )
}

EVm.fzm <- function (model, m, N, rho=1, relative=FALSE, ratio=FALSE, lower=TRUE) {
  m <- as.integer(m)
  if (any(m <= 0)) stop("Error: m must be an integer > 0.")
  alpha <- model$alpha
  A <- model$A
  C <- model$C
  if (missing(rho)) {
    if (!missing(lower)) warning("'lower' parameter can only be used together with 'rho' (ignored)")
    if (ratio) {
      return( Igamma(m + 1 - alpha, N * A, lower=FALSE) / (gamma(m+2) * Igamma(m - alpha, N * A, lower=FALSE)) )
    } else if (relative) {
      return( EVm(model, m, N) / EV(model, N) )
    } else {
      return( C * (N ^ alpha) * Igamma(m - alpha, N * A, lower=FALSE) / gamma(m + 1) )
    }
  } else {
    if (any(rho < 0) || any(rho > 1)) stop("Error: rho must be in the range 0 <= rho <= 1.")
    N.rho <- N * pmax(rho, A)         # so that value becomes 0 for rho <= A
    if (relative || ratio) {
      fraction <- Igamma(m - alpha, N.rho, lower=FALSE) / Igamma(m - alpha, N * A, lower=FALSE)
      if (lower) return(1 - fraction) else return(fraction)
    } else {
      numerator <- if (lower) {
        Igamma(m - alpha, N * A, lower=FALSE) - Igamma(m - alpha, N.rho, lower=FALSE)
      } else {
        Igamma(m - alpha, N.rho, lower=FALSE)
      }
      return( C * (N ^ alpha) * numerator / gamma(m + 1) )
    }
  }
}

VV.fzm <- function (model, N) {
  return( EV.fzm(model, 2*N) - EV.fzm(model, N) )
}

VVm.fzm <- function (model, m, N, rho=1, relative=FALSE, lower=TRUE) {
  m <- as.integer(m)
  if (any(m <= 0)) stop("Error: m must be an integer > 0.")
  if (missing(rho)) {
    if (!missing(relative)) stop("Error: can't compute variances for relative spectrum")
    if (!missing(lower)) warning("'lower' parameter can only be used together with 'rho' (ignored)")
    return( EVm.fzm(model, m, N) - choose(2*m, m) * 2^(-2*m) * EVm.fzm(model, 2*m, 2*N) )
  } else {
    if ((rho < 0) || (rho > 1)) stop("Error: rho must be in the range 0 <= rho <= 1.")
    if (relative) {
      mu.1 <- EVm.fzm(model, m, N, rho=rho, lower=TRUE) # (Evert 2004, Appendix A.3)
      mu.2 <- EVm.fzm(model, m, N, rho=rho, lower=FALSE)
      var.1 <- VVm.fzm(model, m, N, rho=rho, lower=TRUE)
      var.2 <- VVm.fzm(model, m, N, rho=rho, lower=FALSE)
      mu <- mu.1 + mu.2
      var <- var.1 + var.2
      r <- if (lower) mu.1 / mu else mu.2 / mu # actually, variance is the same for lower=FALSE and lower=TRUE 
      s <- if (lower) var.1 / var else var.2 / var
      return( (var / mu^2) * ( (r-s)^2 + s*(1-s) ) )
    } else {
      return( EVm.fzm(model, m, N, rho=rho, lower=lower) - choose(2*m, m) * 2^(-2*m) * EVm.fzm(model, 2*m, 2*N, rho=rho, lower=lower) )
    }
  }
}

## generic functions for object-oriented calling syntax
EV <- function(model, N, ...) {
  UseMethod("EV")
}
EVm <- function(model, m, N, ...) {
  UseMethod("EVm")
}
VV <- function(model, N, ...) {
  UseMethod("VV")
}
VVm <- function(model, m, N, ...) {
  UseMethod("VVm")
}

## write necessary files to evaluate goodness-of-fit with lexstats tools
##   write.lexstats(model, file=<prefix>)
## model = fZM model object based on empirical data (N, V, and spc)
## file  = creates files: <prefix>_bfZM.spc, <prefix>_bfZM.sp2, <prefix>_bfZM.ev2
write.lexstats.fzm <- function (model, file) {
  N <- model$N
  V <- model$V
  Vm <- model$spc
  if (is.null(N) || is.null(V) || is.null(Vm) || length(Vm) < 15) {
    stop("fZM model must be based on empirical data (N, V, and spectrum of length >= 15 required).")
  }
  file <- paste(file, "_bfZM", sep="")  # trick lnreChi2 into assuming three estimated parameters
  Vm <- Vm[1:15]
  ## write .spc file
  EVm <- EVm(model, 1:15, N)
  EalphaM <- EVm(model, 1:15, N, relative=TRUE)
  X <- data.frame(m=1:15, Vm=Vm, EVm=EVm, alphaM=Vm/V, EalphaM=EalphaM)
  write.table(X, file=paste(file, ".spc", sep=""), sep="\t", row.names=FALSE, quote=FALSE)
  ## write .sp2 file
  X <- data.frame(m=1:30, EVm2N=EVm(model, 1:30, 2*N))
  write.table(X, file=paste(file, ".sp2", sep=""), sep="\t", row.names=FALSE, quote=FALSE)
  ## write .ev2 file
  X <- data.frame(V=V, EV=EV(model, N), EV2N=EV(model, 2*N))
  write.table(X, file=paste(file, ".ev2", sep=""), sep="\t", row.names=FALSE, quote=FALSE)
  return(file)                          # returns basename of data files
}

## generic function for object-oriented calling syntax
write.lexstats <- function(model, file, ...) {
  UseMethod("write.lexstats")
}


## plot estimate for proportion of low-frequency types in frequency class m at sample size N
##   fzm.rho.plot(list(model1, model2, ...), m, N, xmin=-6, ymin=-6, xmax=2, ymax=1, filename="", ...)
## models   = list of fZM model objects (up to 3 with default line styles)
## m        = frequency class
## N        = sample size
## filename = if specified, save plot to EPS file
## xmin, xmax = range on x-axis (expected frequency) in log10 units (default: 1e-6 .. 100)
## ymin, ymax = range on y-axis (p-value / estimate for proportion) in log10 units (default: 1e-6 .. 10)
## lty, col, lwd = configure line styles (should have same length as parameter vector)
fzm.rho.plot <- function (models, m, N,
                          xmin=-6, ymin=-6, xmax=2, ymax=1,
                          lty=c("solid","33","11"), col=c("grey50","black","black"), lwd=c(3,3,4),
                          legend=NULL,
                          filename="") {
  n.models <- max(length(models), length(N))
  models <- rep(models, length=n.models)
  N <- rep(N, length=n.models)
  if (n.models > length(lty)) stop("Too many models. Please specify additional line styles.")
  if (length(lty) != length(col) || length(lty) != length(lwd)) stop("Error: lty, col, lwd must have the same length.")
  lty <- lty[1:n.models]                # adjust length of line style vectors
  col <- col[1:n.models]
  lwd <- lwd[1:n.models]
  log.ef <- seq(xmin-1, xmax+1, by=0.1) # logarithmic expected frequency (base 10)
  ef <- 10 ^ log.ef
  pois.p <- ppois(m-1, lambda=ef, lower=F) # p-value from Poisson test, i.e. P(X >= m)
  ylab <- eval(substitute(expression(paste("p-value / ", E * group("[",R[list(m,rho)],"]"))), list(m=m)))
  par.save <- par(mar=c(5,5,2,2)+.1)
  plot(ef, pois.p, log="xy", lwd=lwd[1], type="l",
       xlim=c(10^xmin, 10^xmax), ylim=c(10^ymin, 10^ymax),
       xlab="expected frequency", ylab=ylab)
  for (i in ymin:ymax) abline(h=10^i, col="black", lty="22")
  for (i in xmin:xmax) abline(v=10^i, col="black", lty="22")
  abline(h=1, col="black", lty="solid", lwd=2)
  abline(v=m, col="black", lty="solid", lwd=2)
  text(m, 10^-1, pos=4, offset=0.5, labels=paste("m =",m), srt=-90)
  for (i in 1:n.models) {
    model <- models[[i]]
    prop <- EVm(model, m, N[i], rho=ef/N[i], relative=TRUE)
    lines(ef, prop, lwd=lwd[i], col=col[i], lty=lty[i])
  }
  legtext <- expression("Poisson test")
  for (i in 1:length(models)) {
    if (missing(legend)) {
      model <- models[[i]]
      legtext <- append(legtext, parse(text=paste("list(alpha==",model$alpha,", A==",model$A,")")))
    } else {
      legtext <- append(legtext, legend[i])
    }
  }
  legend(x=10^(xmax-1/4), y=10^(ymin+1/4), xjust=1, yjust=0, legend=legtext,
         lty=c("solid",lty), col=c("black",col), lwd=4, bg="white")
  par(par.save)
  if (filename != "") {
    dev.copy2eps(file=filename)
  }
}
