##
##  UCS / The R ToolBox
##

## Filename: gam.R
## Modified: Thu Aug 25 02:23:49 2005 (severt)   
##   Author: Stefan Evert
##  Purpose: Implementations of generalised association measures

ucs.library("sfunc")
ucs.library("base")


## Yates discounting function (smoothed version of Yates' correction so that measures remain sound)
gam.yates <- function (d) {
  d.sign <- sign(d)
  d.abs <- abs(d)
  d.corr <- ifelse(d.abs >= 1, d.abs - 0.5, d.abs / 2)
  d.sign * d.corr
}

## inverse of Yates discounting function (for iso-surfaces)
gam.yates.inv <- function (d.corr) {
  d.sign <- sign(d.corr)
  d.corr.abs <- abs(d.corr)
  d <- ifelse(d.corr.abs >= 1/2, d.corr.abs + 0.5, 2 * d.corr.abs)
  d.sign * d
}

## standard abbreviations for GAMs in ebo-coordinates:
##   b* = b.star(b), ||b|| = b.norm(b), \bar{e} = e.bar(e, b, N), \sqrt{x} = signed.sqrt(x)
b.star <- function (b) { (b + 1) / sqrt(b) }
b.norm <- function (b) { ifelse(b >= 1, sqrt(b), 1/sqrt(b)) }
e.bar <- function (e, b, N) { N - sqrt(N * e) * b.star(b) + e }
signed.sqrt <- function (x) { sign(x) * sqrt(abs(x)) }

## registry of generalised association measures (GAMs)
##   GAM$name = function object that computes GAM scores
##   ISO$name = function object that computes iso-surface for GAM (if available)
.ucs.GAM <- list()
.ucs.ISO <- list()

## register a generalised association measure (equation is required, iso-surface is optional);
## the registered function will be provided with full coordinates in both ebo (o,e,b)
## and standard (f,f1,f2,N) systems and can use whichever are most natural; the required signatures are
##   (o, e, b, f, f1, f2, N)  for the GAM equation
##   (gamma, e, b, f1, f2, N) for the iso-surface
## all arguments are passed as vectors of the same length
register.gam <- function (name, equation, iso.equation=NULL) {
  eq.args <- strsplit("o e b f f1 f2 N", " ", fixed=TRUE)[[1]] # required signature for equation function
  iso.args <- strsplit("gamma e b f1 f2 N", " ", fixed=TRUE)[[1]] # required signature for iso function
  if (! is.function(equation)) stop("<equation> argument of register.gam() must be a function object")
  args <- names(formals(equation))
  if (length(args) != length(eq.args) || any(args != eq.args)) {
    stop("GAM equation function must have signature (o, e, b, f, f1, f2, N)")
  }
  .ucs.GAM[[name]] <<- equation
  if (! missing(iso.equation)) {        # iso function is optional
    if (! is.function(iso.equation)) stop("<iso.equation> argument of register.gam() must be a function object")
    args <- names(formals(iso.equation))
    if (length(args) != length(iso.args) || any(args != iso.args)) {
      stop("GAM iso function must have signature (gamma, e, b, f1, f2, N)")
    }
    .ucs.ISO[[name]] <<- iso.equation
  }
}

register.gam("frequency",
             eq = function (o, e, b, f, f1, f2, N) { o },
             iso= function (gamma, e, b, f1, f2, N) { gamma })

register.gam("MI",
             eq = function (o, e, b, f, f1, f2, N) { log10(o / e) },
             iso= function (gamma, e, b, f1, f2, N) { 10^gamma * e })

register.gam("gmean",
             eq = function (o, e, b, f, f1, f2, N) { o / sqrt(N * e) },
             iso= function (gamma, e, b, f1, f2, N) { sqrt(N) * gamma * sqrt(e) })

register.gam("z.score",
             eq = function (o, e, b, f, f1, f2, N) { (o - e) / sqrt(e) },
             iso= function (gamma, e, b, f1, f2, N) { gamma * sqrt(e) + e })

register.gam("z.score.corr",
             eq = function (o, e, b, f, f1, f2, N) { gam.yates(o - e) / sqrt(e) },
             iso= function (gamma, e, b, f1, f2, N) { gam.yates.inv(gamma * sqrt(e)) + e })

register.gam("t.score",
             eq = function (o, e, b, f, f1, f2, N) { (o - e) / sqrt(o) },
             iso= function (gamma, e, b, f1, f2, N) { gamma^2 / 2 + gamma * sqrt(e + gamma^2 / 4) + e })

register.gam("Poisson.pv",
             eq = function (o, e, b, f, f1, f2, N) {
               -Rgamma(o, e, log=TRUE) # negative decadic logarithm of p-value!
             })

### Poisson.Stirling is two-sided => not a sound measure
### register.gam("Poisson.Stirling", eq = function (o, e, b, f, f1, f2, N) { o * (log(o) - log(e) - 1) / log(10) })

register.gam("MI2",
             eq = function (o, e, b, f, f1, f2, N) { log10(o^2 / e) },
             iso= function (gamma, e, b, f1, f2, N) { 10^(gamma / 2) * sqrt(e) })

register.gam("MI3",
             eq = function (o, e, b, f, f1, f2, N) { log10(o^3 / e) },
             iso= function (gamma, e, b, f1, f2, N) { 10^(gamma / 3) * e^(1/3) })

register.gam("MI4",
             eq = function (o, e, b, f, f1, f2, N) { log10(o^4 / e) },
             iso= function (gamma, e, b, f1, f2, N) { 10^(gamma / 4) * e^(1/4) })

### local.MI is an unsound measure (its minimum is reached for o = e/exp(1))
### register.gam("local.MI", eq = function (o, e, b, f, f1, f2, N) { o * log10(o / e) })

register.gam("Dice",
             eq = function (o, e, b, f, f1, f2, N) { (o / sqrt(N * e)) * (2 / b.star(b)) },
             iso= function (gamma, e, b, f1, f2, N) { sqrt(N) * gamma * (b.star(b) / 2) * sqrt(e) })

register.gam("Jaccard",
             eq = function (o, e, b, f, f1, f2, N) { o / (f1 + f2 - o) },
             iso= function (gamma, e, b, f1, f2, N) { sqrt(N) * (gamma / (1 + gamma)) * b.star(b) * sqrt(e) })

register.gam("MS",
             eq = function (o, e, b, f, f1, f2, N) { o / (sqrt(N * e) * b.norm(b)) },
             iso= function (gamma, e, b, f1, f2, N) { sqrt(N) * b.norm(b) * gamma * sqrt(e) })

register.gam("Liddell",
             eq = function (o, e, b, f, f1, f2, N) { N * (o - e) / (f2 * (N - f2)) },
             iso= function (gamma, e, b, f1, f2, N) { (gamma / N) * f2 * (N - f2) + e })

register.gam("chi.squared",
             eq = function (o, e, b, f, f1, f2, N) {
               N * (o - e) * abs(o - e) / (e * e.bar(e, b, N)) # (o-e) * abs(o-e) = signed square for one-sided measure
             },
             iso= function (gamma, e, b, f1, f2, N) {
               signed.sqrt(gamma / N) * sqrt(e.bar(e, b, N)) * sqrt(e) + e
             })

register.gam("chi.squared.corr",
             eq = function (o, e, b, f, f1, f2, N) {
               d.corr <- gam.yates(o - e)
               N * d.corr * abs(d.corr) / (e * e.bar(e, b, N))
             },
             iso= function (gamma, e, b, f1, f2, N) {
               gam.yates.inv( signed.sqrt(gamma / N) * sqrt(e.bar(e, b, N)) * sqrt(e) ) + e
             })

register.gam("odds.ratio.disc",
             eq = function (o, e, b, f, f1, f2, N) {
               O11 <- o
               O12 <- f1 - o
               O21 <- f2 - o
               O22 <- N - f1 - f2 + o
               log10( (O11 + .5) * (O22 + .5) / ((O12 + .5) * (O21 + .5)) )
             })

register.gam("log.likelihood",
             eq = function (o, e, b, f, f1, f2, N) {
               O11 <- o
               O12 <- f1 - o
               O21 <- f2 - o
               O22 <- N - f1 - f2 + o
               E11 <- e
               E12 <- f1 - e
               E21 <- f2 - e
               E22 <- N - f1 - f2 + e
               suppressWarnings( # logarithms may generate NaN's (which are then ignored by the ifelse)
                                ll  <- 2 * (ifelse(O11 > 0, O11 * log(O11 / E11), 0) +
                                            ifelse(O12 > 0, O12 * log(O12 / E12), 0) +
                                            ifelse(O21 > 0, O21 * log(O21 / E21), 0) +
                                            ifelse(O22 > 0, O22 * log(O22 / E22), 0))
                                )
               ifelse(o > e, ll, -ll)                # one-sided measure
             })

### average-MI is a two-sided measure => not sound
### register.gam("average.MI",
###              eq = function (o, e, b, f, f1, f2, N) {
###                O11 <- o
###                O12 <- f1 - o
###                O21 <- f2 - o
###                O22 <- N - f1 - f2 + o
###                E11 <- e
###                E12 <- f1 - e
###                E21 <- f2 - e
###                E22 <- N - f1 - f2 + e
###                (ifelse(O11 > 0, O11 * log10(O11 / E11), 0) +
###                 ifelse(O12 > 0, O12 * log10(O12 / E12), 0) +
###                 ifelse(O21 > 0, O21 * log10(O21 / E21), 0) +
###                 ifelse(O22 > 0, O22 * log10(O22 / E22), 0))
###              })

## use global variable MI.conf.alpha as parameter for MI.conf measure
MI.conf.alpha <- 2
register.gam("MI.conf",
             eq = function (o, e, b, f, f1, f2, N) { log10( Rgamma.inv(o, -MI.conf.alpha - log10(2), log=TRUE) / e ) })

## ebo.list <- fsig2ebo(f, f1, f2, N)        [ returns data frame with columns o, e, b, N ]
## fsig.list <- ebo2fsig(o, e, b=1, N=1e6)   [ returns data frame with columns f, f1, f2, N ]
##   convert between standard coordiantes and ebo-coordinates (ensures that result vectors have the same length)
fsig2ebo <- function (f, f1, f2, N) {
  l <- max(length(f), length(f1), length(f2), length(N))
  f <- as.double(f)                     # force frequency signature into floating-point
  f1 <- as.double(f1)
  f2 <- as.double(f2)
  data.frame(o=rep(f, length=l),
             e=rep(f1 * f2 / N, length=l),
             b=rep(f1 / f2, length=l),
             N=rep(N, length=l))
}

ebo2fsig <- function (o, e, b=1, N=1e6) {
  l <- max(length(o), length(e), length(b), length(N))
  data.frame(f=rep(o, length=l),
             f1=rep(sqrt(N * e * b), length=l),
             f2=rep(sqrt(N * e / b), length=l),
             N=rep(N, length=l))
}

## if (is.builtin.gam("chi.squared")) { ... }
##   check whether the specified GAM is supported by this module
is.builtin.gam <- function (key) {
  ok <- key %in% names(.ucs.GAM)
  return(ok)
}

## keys <- builtin.gams()
##   returns list of the names of GAMs supported by this module
builtin.gams <- function () {
  names(.ucs.GAM)
}

## scores <- gam.score(gam, f=f, f1=f1, f2=f2, N=N)
## scores <- gam.score(gam, o=o, e=e [, b=1] [, N=1e6])
##   compute GAM scores either from frequency signature or from ebo-coordinates;
##   in the latter case, b and N are optional; for a non-central measure, the
##   default b=1 yields a centralised version; for a non-size-invariant measure,
##   the default N=1e6 simulates data from a large corpus (where non-trivial effects
##   of the sample size are usually negligible);
##   note that the first version accepts non-integer arguments (i.e. jittered freq's)
gam.score <- function (key, o=NA, e=NA, b=1, N=1e6, f=NA, f1=NA, f2=NA) {
  if (missing(o) && missing(f)) stop("need either frequency signature or ebo coordinates")
  if (! missing(o)) {
    if (missing(e)) stop("ebo-coordinates incomplete (e is not speficied)")
    if ((! missing(f)) || (! missing(f1)) || (! missing(f2))) {
      warning("specified frequency signature ignored, using ebo-coordinates")
    }
    fsig <- ebo2fsig(o, e, b, N)
    l <- nrow(fsig)                     # automatically computes maximal common length
    o <- rep(o, length=l)
    e <- rep(e, length=l)
    b <- rep(b, length=l)
    N <- rep(N, length=l)
    f  <- fsig$f
    f1 <- fsig$f1
    f2 <- fsig$f2
  }
  else if (! missing(f)) {
    if (missing(f1) || missing(f2) || missing(N)) stop("frequency signature (f, f1, f2, N) incomplete")
    if ((! missing(o)) || (! missing(e)) || (! missing(b))) {
      warning("specified ebo-coordinates ignored, using frequency signature")
    }
    ebo <- fsig2ebo(f, f1, f2, N)
    l <- nrow(ebo)                      # automatically computes maximal common length
    f  <- as.double(rep(f, length=l))
    f1 <- as.double(rep(f1, length=l))
    f2 <- as.double(rep(f2, length=l))
    N  <- as.double(rep(N, length=l))
    o <- ebo$o
    e <- ebo$e
    b <- ebo$b
  }
  if (! is.builtin.gam(key)) stop(paste("GAM", key, "is not implemented"))
  gam.func <- .ucs.GAM[[ key ]]
  gam.func(o=o, e=e, b=b, f=f, f1=f1, f2=f2, N=N)
}


## ds <- add.jitter(ds [, amount = 0.5, overwrite=FALSE])
##   add variables x.jitter.f, x.jitter.f1, x.jitter.f2 with random jitter to data set
##   (required for all functions in the "gam" and "eo" modules that use jittering)
add.jitter <- function (ds, amount=0.5, overwrite=FALSE) {
  n.row <- nrow(ds)
  vars <- colnames(ds)
  jits <- paste("x.jitter", c("f", "f1", "f2"), sep=".")
  ## by default, don't overwrite existing jitter (if any of the jitter variables is missing, all three are recomputed)
  if (overwrite || !all("x.jitter.f" %in% vars)) {  
    for (v in jits) {
      ds[[ v ]] <- runif(n.row, min=-amount, max=amount)
    }
    O12 <- ds$f1 + ds$x.jitter.f1 - (ds$f + ds$x.jitter.f)
    err.f1 <- O12 < 0                   # invalid coordinates (O12 < 0)
    ds$x.jitter.f1[err.f1] <- ds$x.jitter.f1[err.f1] - O12[err.f1] # so that O12 == 0
    O21 <- ds$f2 + ds$x.jitter.f2 - (ds$f + ds$x.jitter.f)
    err.f2 <- O21 < 0                   # invalid coordinates (O21 < 0)
    ds$x.jitter.f2[err.f2] <- ds$x.jitter.f2[err.f2] - O21[err.f2] # so that O21 == 0
  }
  invisible(ds)
}

## if (! has.jitter(ds)) stop("jitter missing")
## has.jitter(ds, fail=TRUE)
has.jitter <- function (ds, fail=FALSE) {
  vars <- colnames(ds)
  for (v in paste("x.jitter", c("f", "f1", "f2"), sep=".")) {
    if (! (v %in% vars)) {
      if (fail) stop("you must add jitter vectors to the data set first (use add.jitter())")
      return(FALSE)
    }
  }
  return(TRUE)
}

## ds <- add.ebo(ds [, jitter=TRUE])
##   compute (or re-compute) ebo-coordinates for data set, optionally with jitter
add.ebo <- function (ds, jitter=FALSE) {
  f  <- ds$f
  f1 <- ds$f1
  f2 <- ds$f2
  N  <- ds$N
  if (jitter) {
    has.jitter(ds, fail=TRUE)
    f  <- f  + ds$x.jitter.f
    f1 <- f1 + ds$x.jitter.f1
    f2 <- f2 + ds$x.jitter.f2
  }
  ebo <- fsig2ebo(f, f1, f2, N)
  ds$o <- ebo$o
  ds$e <- ebo$e
  ds$b <- ebo$b
  invisible(ds)
}

## ds <- add.gams(ds, ds.find.am(ds) [, jitter=TRUE])
##   compute GAM scores for specified measures in data set (optionally taking jitter into account)
add.gams <- function (ds, keys, jitter=FALSE) {
  f  <- ds$f
  f1 <- ds$f1
  f2 <- ds$f2
  N  <- ds$N
  if (jitter) {
    has.jitter(ds, fail=TRUE)
    f  <- f  + ds$x.jitter.f
    f1 <- f1 + ds$x.jitter.f1
    f2 <- f2 + ds$x.jitter.f2
  }
  for (gam in keys) {
    if (is.builtin.gam(gam)) {
      cat(gam, " ", sep="")
      ds[[ am.key2var(gam) ]] <- gam.score(gam, f=f, f1=f1, f2=f2, N=N)
    } else {
      warning(paste("GAM", gam, "not implemented (skipped)"))
    }
  }
  cat("\n")
  invisible(ds)
}

## relative or absolute error for binary search (whichever is smaller)
gam.error <- function (a, b) {
  abs(a - b) / pmax(abs(a), abs(b), 1)
}

## f <- gam.iso(gam, gamma, f1=f1, f2=f2, N=N)
## o <- gam.iso(gam, gamma, e=e [, b=1] [, N=1e6])
##   compute iso-surface for GAM, i.e. find f- or o-coordinate such that g(x) = gamma
##   given f1, f2, N or e[, b, N]; in ebo-coordinates, defaults to b=1 (centralised
##   measure) and N=1e6 (data from large corpus);
##   in either case, the returned value will usually not be an integer
gam.iso <- function (key, gamma, e=NA, b=1, N=1e6, f1=NA, f2=NA, bsearch.min=NULL, bsearch.max=NULL) {
  if (missing(e) && missing(f1)) stop("need either f1, f2 or e to compute height of iso-surface")
  if (! missing(e)) {
    if ((! missing(f1)) || (! missing(f2))) {
      warning("specified frequency signature ignored, using ebo-coordinates")
    }
    l <- max(length(gamma), length(e), length(b), length(N))
    gamma <- rep(gamma, length=l)
    e <- rep(e, length=l)
    b <- rep(b, length=l)
    N <- rep(N, length=l)
    fsig <- ebo2fsig(1, e, b, N)
    f1 <- fsig$f1
    f2 <- fsig$f2
  }
  else if (! missing(f1)) {
    if (missing(f2) || missing(N)) stop("frequency signature (*, f1, f2, N) incomplete")
    if ((! missing(e)) || (! missing(b))) {
      warning("specified ebo-coordinates ignored, using frequency signature")
    }
    l <- max(length(gamma), length(f1), length(f2), length(N))
    gamma <- rep(gamma, length=l)
    f1 <- as.double(rep(f1, length=l))
    f2 <- as.double(rep(f2, length=l))
    N  <- as.double(rep(N, length=l))
    ebo <- fsig2ebo(1, f1, f2, N)
    e <- ebo$e
    b <- ebo$b
  }
  if (! is.builtin.gam(key)) stop(paste("GAM", key, "is not implemented"))
  if (key %in% names(.ucs.ISO)) {       # explicit formula for iso-surface is available
    iso.func <- .ucs.ISO[[ key ]]
    return( iso.func(gamma=gamma, e=e, b=b, f1=f1, f2=f2, N=N) )
  } else {                              # otherwise compute iso-surface by binary search
    gam.func <- .ucs.GAM[[ key ]]
    eps <- 10 * .Machine$double.eps     # require accuracy approx. 1 order of magnitue greater than machine eps
    ## set initial limits to plausible values, otherwise some AM equations may break down; note that for such
    ## AMs, the iso-surface may not exist for all values of f1 and f2 (and the specified gamma); for other AMs,
    ## this algorithm will not be able to determine the iso-surface correctly when it is outside the plausible range;
    ## use the bsearch.min and bsearch.max parameters to override automatically chosen limits in this case
    if (missing(bsearch.min)) {
      o.min <- rep(eps, length=l)         # smallest possible value for o is 1, with jittering approx 0.5
    } else {
      o.min <- rep(bsearch.min, length=l)
    }
    if (missing(bsearch.max)) {
      o.max <- rep(pmin(f1, f2), length=l) # realistic upper bound for o
    } else {
      o.max <- rep(bsearch.max, length=l)
    }
    while (max(gam.error(o.max, o.min)) > eps) { 
      o <- (o.min + o.max) / 2
      score <- gam.func(o=o, e=e, b=b, f=o, f1=f1, f2=f2, N=N)
      o.max <- ifelse( score >= gamma, o, o.max )
      o.min <- ifelse( score <= gamma, o, o.min )
    }
    o <- (o.min + o.max) / 2
    score <- gam.func(o=o, e=e, b=b, f=o, f1=f1, f2=f2, N=N)
    failures <- abs(score - gamma) / abs(gamma + eps) > .01 # return NA where iso-surface could not be determined (or is highly inaccurate)
    return( ifelse(failures, NA, o) )
  }
}

## gamma.n <- gamma.nbest(ds, gam, n [, jitter=TRUE])
##   compute n-best cutoff threshold for given GAM and data set (computes GAM scores and ranks on the fly)
gamma.nbest <- function (ds, gam, n, jitter=FALSE) {
  f  <- ds$f
  f1 <- ds$f1
  f2 <- ds$f2
  N  <- ds$N
  if (jitter) {
    has.jitter(ds, fail=TRUE)
    f  <- f  + ds$x.jitter.f
    f1 <- f1 + ds$x.jitter.f1
    f2 <- f2 + ds$x.jitter.f2
  }
  if (! is.builtin.gam(gam)) stop(paste("GAM", gam, "is not implemented"))
  scores <- sort(gam.score(gam, f=f, f1=f1, f2=f2, N=N), decreasing=TRUE)
  size <- length(scores)
  if (n > size) {
    return(min(scores))
  }
  else if (n < 1) {
    return(max(scores) + 1)
  }
  else {
    return(scores[n])
  }
}
