##
##  UCS / The R ToolBox
##

## Filename: iaa.R
## Modified: Wed May  5 14:01:03 2004 (evert)   
##   Author: Stefan Evert
##  Purpose: Measuring intercoder agreement on binary variables for two annotators

## All functions in this library accept either a contingency table in the form of
## a 2x2 matrix (with entries [1,1] and [2,2] corresponding to agreement), or two
## parallel Boolean vectors (of the same length) representing the annotators' choices.

## internal function to check arguments and construct contingency table if necessary
iaa.args <- function (x, y) {
  if (is.null(y)) {
    d <- dim(x)
    if (length(d) != 2 || any(d != 2)) stop("argument of iaa.*() must be 2x2 contingency table")
    return(x)
  } else {
    ok <- is.logical(x) && is.logical(y) && (length(x) == length(y))
    if (!ok) stop("arguments of iaa.*() must be Boolean vectors of the same length")
    return(table(x,y))
  }
}

## result <- iaa.kappa(tbl, conf.level=0.95)
## result <- iaa.kappa(x, y, conf.level=0.95)
##   computes Cohen's kappa statistic as a measure for intercoder agreement, with
##   approximate confidence interval at the specified level based on the formula
##   for sample variance from (Fleiss, Cohen & Everitt 1969, p. 325); the confidence
##   interval is two-sided and defaults to the 95% level;
##   returns list with elements kappa, sd, kappa.min, kappa.max
iaa.kappa <- function(x, y=NULL, conf.level=0.95) {
  ## notation follows Fleiss, Cohen & Everitt (1969)
  tbl <- iaa.args(x, y)                 # agreement contingency table (n[i,j] in reference)
  n <- sum(tbl)                         # sample size
  p <- tbl / n                          # contingency table of proportions
  px. <- rowSums(p)                     # row and column sums of the proportions
  p.x <- colSums(p)
  p11 <- p[1,1]; p12 <- p[1,2]; p21 <- p[2,1]; p22 <- p[2,2]
  p1. <- px.[1]; p2. <- px.[2]
  p.1 <- p.x[1]; p.2 <- p.x[2]
  p.o <- p11 + p22                      # observed proportion of agreement
  p.e <- p1. * p.1 + p2. * p.2          # expected proportion of chance agreement
  kappa <- (p.o - p.e) / (1 - p.e)      # Cohen's kappa (actually, the sample estimate \hat{\kappa})
  var <- (                              # sample estimate for variance
          p11 * ( (1 - p.e) - (p.1 + p1.)*(1 - p.o) )^2 + # i=1
          p22 * ( (1 - p.e) - (p.2 + p2.)*(1 - p.o) )^2 + # i=2
          (1 - p.o)^2 * p12 * (p.1 + p2.)^2 + # i=1, j=2
          (1 - p.o)^2 * p21 * (p.2 + p1.)^2   # i=2, j=1
          - (p.o * p.e - 2 * p.e + p.o)^2
          ) / (n * (1 - p.e)^4)
  sd <- sqrt(var)                       # standard deviation
  alpha <- (1 - conf.level) / 2         # compute z-score corresponding to two-sided confidence level
  z <- -qnorm(alpha)
  result <- data.frame(kappa=kappa, sd=sd, kappa.min=kappa-z*sd, kappa.max=kappa+z*sd, row.names="Result:")
  return(result)
}

## conf.int <- iaa.pta.conservative(tbl, conf.level=0.95, debug=FALSE)
## conf.int <- iaa.pta.conservative(x, y, conf.level=0.95, debug=FALSE)
##   computes a conservative estimate (i.e. a probably over-sized confidence interval)
##   for the proportion of true agreement between the coders, assuming that their
##   annotations are independent for the remaining candidates;
##   returns endpoints of the conf. interval as vector c(lower, upper)
iaa.pta.conservative <- function (x, y=NULL, conf.level=0.95, debug=FALSE) {
  n <- iaa.args(x,y)                    # agreement contingency table
  size <- sum(n)                        # sample size
  n.agree <- sum(diag(n))               # number of candidates on which coders agree
  ## internal function: p.value <- fisher.p.value(tbl)
  ## fisher.test() insists on computing a confidence interval for the odds ratio from the
  ## very expensive non-central hypergeometric distribution; this helper function computes
  ## only the p.value of Fisher's test; it cheats a little by returning two times a
  ## one-sided p-value (from the hypergeometric distribution) 
  fisher.p.value <- function (n) {
    k <- sum(n[1,])     # hypergeometric distribution for k balls (row 1) drawn from an urn
    a <- sum(n[,1])     # with a white balls (col 1) and b black balls (col 2)
    b <- sum(n[,2])
    x <- n[1,1]         # x is the number of white balls in the sample
    exp.x <- k * (a / (a+b))            # expectation of x
    if (x > exp.x) {
      p <- phyper(x-1, a, b, k, lower=FALSE) # upper tail probability if x > E[x]
    } else {
      p <- phyper(x, a, b, k)           # lower tail probability if x <= E[x]
    }
    return(if (2*p <= 1) 2*p else 1)    # bad cheat, but we're only interested in p < alpha anyway
  }
  ## internal function: p.value <- test.model(m)
  ## test whether dual model is consistent with observed data, when there are m candidates with true agreement;
  ## tries all possible divisions of test set into C1 (true agreement) and C2 (chance agreement) so that
  ## C1 contains exactly m candidates; divisions are parametrised by m.P, the number of TPs in C1;
  ## for each division, a two-sided p-value from Fisher's exact test applied to C2 is computed
  test.model <- function (m) {
    m.P.min <- max(0, m - n[2,2])
    m.P.max <- min(m, n[1,1])
    if (m.P.min > m.P.max) return(0)    # should not happen, really
    m.P.support <- m.P.min:m.P.max      # possible values for m.P
    p.value <- numeric(0)
    for (m.P in m.P.support) {          # number of TPs in C1
      m.M <- m - m.P                    # number of FPs in C1
      n. <- n                           # agreement contingency table for C2
      n.[1,1] <- n[1,1] - m.P
      n.[2,2] <- n[2,2] - m.M
      p <- fisher.p.value(n.)           # two-sided Fisher's test on n.[]
      p.value <- append(p.value, p)
    }
    p.max <- max(p.value)
    m.P.max <- m.P.support[which.max(p.value)]
    if (debug) cat("test.model(",m,") -> ",p.max," (for m.P=",m.P.max,")\n", sep="")
    return(p.max)                     # return largest p-value found
  }
  ## now run test.model() for all possible values of m and collect those which are accepted
  ## at the specified confidence level (i.e. where the maximal two-sided p-value is < alpha)
  alpha <- 1 - conf.level               # test.model() computes two-sided p-values
  m.support <- 0:n.agree                # possible values for m
  m.accept <- integer(0)
  for (m in m.support) {
    if (test.model(m) >= alpha) m.accept <- append(m.accept, m)
  }
  if (any(diff(m.accept,1) != 1)) warn("confidence set has unexpected gaps")
  result <- c(min(m.accept), max(m.accept)) / size
  names(result) <- c("prop.min", "prop.max")
  return(result)
}


## conf.int <- iaa.pta.homogeneous(tbl, conf.level=0.95, debug=FALSE)
## conf.int <- iaa.pta.homogeneous(x, y, conf.level=0.95, debug=FALSE)
##   computes a conservative estimate (i.e. a probably over-sized confidence interval)
##   for the proportion of true agreement between the coders, assuming that their
##   annotations are independent for the remaining candidates;
##   returns endpoints of the conf. interval as vector c(lower, upper)
iaa.pta.homogeneous <- function (x, y=NULL, conf.level=0.95, debug=FALSE) {
  n <- iaa.args(x,y)                    # agreement contingency table
  size <- sum(n)                        # sample size
  n.agree <- sum(diag(n))               # number of candidates on which coders agree
  p1. <- sum(n[1,]) / size              # overall proportions of TPs for coders
  p.1 <- sum(n[,1]) / size
  alpha <- (p1. + p.1) / 2  # MLEs for the population parameters (of the hom. dual model)
  del1 <- (p1. - p.1) / 2   # del1 = (1 - theta) * delta
  ## internal function: p.value <- binom.p.value(k, n, p)
  ## fast approximation to the p-value of a two-sided binomial test; returns appropriate
  ## one-sided p-value multiplied by two, but clamped to the interval [0,1]
  binom.p.value <- function (k, n, p) {
    if (k > n * p) {
      p.value <- pbinom(k - 1, n, p, lower.tail=FALSE) # upper tail if k > E[k]
    } else {
      p.value <- pbinom(k, n, p)        # lower tail otherwise
    }
    return (if (2*p.value <= 1) 2*p.value else 1) # clamp to [0,1]
  }
  ## internal function: p.value <- test.model(m)
  ## test whether dual model is consistent with observed data, when there are m candidates
  ## with true agreement; uses homogeneous assumption to estimate proportion of TPs in C2
  ## for each of the annotators
  test.model <- function (m) {
    ## since the homogeneous dual model makes a fairly strong assumption, which can
    ## lead to extreme (or impossible) values for the TP probabilities in C2 (beta1 for A,
    ## beta2 for B) and C1 (alpha), we begin with some sanity checks:
    ## the model is rejected when the hypothesised probabilities (beta1, beta2, alpha) are
    ## too unusual (for instance, when B agrees with all the TPs of A but also marks many other
    ## candidates as TPs, the model would postulate beta1 = 0 and accept all surface agreement
    ## as true agreement; the dual model is clearly not suitable for such a skewed situation)
    if (m >= n.agree) return(0)         # impossible (too much true agreement hypothesised)
    if (m >= size) return(1)            # means total agreement -> accept
    theta <- m / size
    ## check whether there is some division of the test set into C1 and C2 that
    ## gives a reasonable number of TPs in C1 (compared to the hypothesised probability alpha)
    mP.max <- min(m, n[1,1])            # limits on the number of TPs in C1 (cf. iaa.tpa.conservative())
    mP.min <- max(0, m - n[2,2])
    ## reasonable values are obtained by cutting off the upper and lower ALPHA quantiles
    ## from the theoretical distribution, where ALPHA = (1 - conf.level) / 2
    theoretical.max <- qbinom((1-conf.level)/2, m, alpha, lower.tail=FALSE)
    theoretical.min <- qbinom((1-conf.level)/2, m, alpha)
    if (debug) cat("mP = ",mP.min,"...",mP.max,
                   "  theoretical = ",theoretical.min,"...",theoretical.max,
                   "  [alpha=",alpha,"]\n", sep="")
    if (mP.max < theoretical.min || mP.min > theoretical.max) return(0) # compare limits to reasonable values
    ## check whether the hypothesised values for beta1 and beta2 are within the range [0,1]
    delta <- del1 / (1 - theta)         # C2 parameters of the population model (MLEs, I guess)
    beta1 <- alpha + delta
    beta2 <- alpha - delta
    if (any(beta1<=0, beta1>=1, beta2<=0, beta2>=1)) return(0) # impossible (-> m too large)
    q.e <- beta1*beta2 + (1 - beta1)*(1 - beta2) # expected rate of chance agreement in C2
    n.o <- n.agree - m                  # number of observed agreements (in C2)
    size.c2 <- size - m                 # sample size of C2
    p.value <- binom.p.value(n.o, size.c2, q.e)
    if (debug) cat("test.model(",m,") -> ",p.value,"  [exp=",size.c2*q.e,", obs=",n.o,"]\n", sep="")
    return(p.value)
  }
  ## now run test.model() for all possible values of m and collect those which are accepted
  ## at the specified confidence level (i.e. where the maximal two-sided p-value is < alpha)
  m.support <- 0:n.agree                # possible values for m
  m.accept <- integer(0)
  for (m in m.support) {
    if (test.model(m) >= 1 - conf.level) m.accept <- append(m.accept, m)
  }
  if (any(diff(m.accept,1) != 1)) warn("confidence set has unexpected gaps")
  result <- c(min(m.accept), max(m.accept)) / size
  names(result) <- c("prop.min", "prop.max")
  return(result)
}

