##
##  UCS / The R ToolBox
##

## Filename: lexstats.R
## Modified: Wed Jun 14 14:16:19 2006 (severt)   
##   Author: Stefan Evert
##  Purpose: Some utilities for lexical statistics, interfacing with the Lexstats software (Baayen 2001)

## References:
## Baayen, R. Harald (2001).  Word Frequency Distributions.  Kluwer, Dordrecht.


## plot log-transformed type density (so that area under log curve == number of types)
##   type.density.plot(g, plim=c(-20, 0), steps=10, add=FALSE, ...)
## g     = function that computes type density g(pi), must accept vector of type probabilities
## plim  = range of type probabilities, in log10 units (default corresponds to 1e-20 .. 1e0 = 1)
## steps = number of values computed in each log10 unit
## ...   = further arguments are passed to the plot() function
type.density.plot <- function (g, plim=c(-20, 0), steps=10, add=FALSE, ...) {
  lp <- seq(plim[1], plim[2], by=1/steps) # sequence of log10(pi) values
  p <- 10^lp
  g.log10 <- g(p) * p * log(10)         # log10-transformed type density
  param <- list(...) 
  pars <- names(param)                  # set default graphics parameters (e.g. plot type)
  if (! "type" %in% pars) param$type <- "l"
  if (! "lwd" %in% pars)  param$lwd <- 2
  if (! "xlab" %in% pars && ! add) param$xlab <- expression(log[10](pi))
  if (! "ylab" %in% pars && ! add) param$ylab <- expression(paste("type density ", g(pi)))
  if (! "xlim" %in% pars && ! add) param$xlim <- plim
  if (! "ylim" %in% pars && ! add) param$ylim <- c(0, max(g.log10))
  param$x <- lp
  param$y <- g.log10
  if (add) do.call("points", param) else do.call("plot", param)
}


## comparative plot of frequency spectra (barplot or points-and-lines on log scale)
##   spectrum.plot(spc=list(spc1, spc2, ...), m.max=Inf, log=FALSE, legend=NULL, y.min=100, y.max, xlab, ylab, ...)
## spc   = list of frequency spectrum vectors spc1=c(V1,V2,...), spc2=c(V1,V2,...), ...
## m.max = plot spectrum up to frequency rank m.max (defaults to length of spectrum vectors)
## log   = FALSE: barplot,  TRUE: points-and-lines on y-logarithmic scale (or specify explicitly: log="xy", etc.)
## legend = if specified, vector of labels for legend box
## y.min = minimum value shown on y-axis in logarithmic plot (defaults to 100)
## y.max = maximum value shown on y-axis (default: automatically computed from input data)
## xlab  = label for x-axis (defaults to "m")
## ylab  = label for y-axis (defaults to "Vm / E[Vm]")
## pch, lty, col, lwd = configure line styles in log plot, otherwise just the colours of the bars
##         (values are repeated if necessary, use pch=" " to disable points)
## ...   = further arguments are passed to the plot() function
spectrum.plot <- function (spc, m.max=Inf, log=FALSE, legend=NULL, y.min=100, y.max=0,
                           xlab=expression(m), ylab=expression(V[m] / paste(E, group("[", V[m], "]"))),
                           pch=c(1,3,15,2,20), lwd=1,
                           lty=c("solid", "dashed", "dotdash", "dotted", "twodash"),
                           col=if (log) c("black") else c("black", "grey50", "white", "grey70", "grey20"),
                           ...) {
  if (is.character(log)) { log.type <- log; log <- TRUE } else log.type <- "y"
  if (missing(m.max)) m.max <- min(sapply(spc, length)) # m.max defaults to length of shortest spectrum
  spc <- lapply(spc,                    # pad / trim shorter spectrum vectors to required length
                function (vec) { l <- length(vec); if (l < m.max) c(vec, rep(0, m.max-l)) else vec[1:m.max] })
  if (missing(y.max)) y.max <- max(sapply(spc, max))
  num.spc <- length(spc)
  pch <- rep(pch, length=num.spc)       # adjust length of style vectors (may repeat styles)
  lty <- rep(lty, length=num.spc)
  col <- rep(col, length=num.spc)
  lwd <- rep(lwd, length=num.spc)
  if (log) {                            # points-and-lines on logarithmic scale
    plot(1, 1, type="n", xlim=c(1,m.max), ylim=c(y.min,y.max), # set up plotting region
         log=log.type, xlab=xlab, ylab=ylab, ...)
    for (i in 1:num.spc) {
      points(1:m.max, spc[[i]], type="o", pch=pch[i], lty=lty[i], col=col[i], lwd=lwd[i])
    }
    if (!missing(legend)) {
      legend(m.max, y.max, xjust=1, yjust=1,
             legend=legend, pch=pch, lty=lty, col=col, lwd=lwd)
    }
  } else {                              # barplot on normal scale
    barplot(do.call("rbind", spc), beside=TRUE, ylim=c(0,y.max),
            col=col, names.arg=1:m.max, xlab=xlab, ylab=ylab,
            ...)
    if (!missing(legend)) {
      x.max <- par("usr")[2]
      legend(x.max, y.max, xjust=1, yjust=1,
             legend=legend, fill=col, col="black")
    }
  }
}

## read frequency spectrum file in lexstats format and convert to plain vector (with zero elements)
##   Vm <- read.spectrum(file, m.max=Inf, expected=FALSE)
## file     = filename of .spc file in lexstats format
## m.max    = maximum length of frequency spectrum (ignores any higher ranks in .spc file)
## expected = if TRUE, read expected class sizes (EVm column) instead of observed ones (Vm column)
## returns: list with N=sample_size, V=vocabulary_size, spc=spectrum_vector
read.spectrum <- function(file, m.max=Inf, expected=FALSE) {
  tmp <- read.table(file, header=TRUE)
  m.max <- min(m.max, max(tmp$m))
  m.idx <- tmp$m <= m.max
  result <- rep(0, m.max)
  if (expected) {
    if (! "EVm" %in% colnames(tmp)) stop("Sorry, .spc file doesn't contain expected frequencies [EVm].") 
    result[tmp$m[m.idx]] <- tmp$EVm[m.idx]
  }
  else {
    result[tmp$m[m.idx]] <- tmp$Vm[m.idx]
  }
  N <- sum(tmp$m * tmp$Vm)
  V <- sum(tmp$Vm)
  rm(tmp)
  return(list(N=N, V=V, spc=result))
}

## goodness-of-fit evaluation of an LNRE model (Baayen 2001, Sec. 3.3)
lnre.goodness.of.fit <- function(model, m.max=15) {
  ## compute multivariate chi-squared test statistic, following Baayen (2001, p. 119ff)
  n.param <- model$n.param              # number of parameters fitted to the data
  N <- model$N                          # observed N, V, Vm from model
  V <- model$V
  Vm <- model$spc
  if (m.max > length(Vm)) stop("not enough elements in frequency spectrum (try reducing m.max)")
  Vm <- Vm[1:m.max]                     # shorten frequency spectrum to desired number of elements
  df <- m.max + 1 - n.param             # degrees of freedom of the computed chi-squared statistic
  if (df < 1) stop("m.max too small (must be larger than number of fitted parameters)")
  E.Vm <- EVm(model, 1:m.max, N)         # E[Vm(N)]
  E.Vm.2N <- EVm(model, 1:(2*m.max), 2 * N)  # E[Vm(2N)]
  E.V <- EV(model, N)                   # E[V(N)]
  E.V.2N <- EV(model, 2 * N)            # E[V(2N)]
  err.vec <- c(V, Vm) - c(E.V, E.Vm)    # error vector (p. 119)
  cov.Vm.Vk <- diag(E.Vm) -             # "inner part" of covariance matrix (p. 120, (3.61))
    outer(1:m.max, 1:m.max, function (m,k) choose(m+k, m) * E.Vm.2N[m+k] / 2^(m+k))
  cov.Vm.V <- E.Vm.2N[1:m.max] / 2^(1:m.max) # (p. 121, (3.63))
  R <- rbind(c(VV(model, N), cov.Vm.V), # construct covariance matrix R (p. 119, (3.58))
             cbind(cov.Vm.V, cov.Vm.Vk))
  x2 <- err.vec %*% solve(R, err.vec)   # (p. 119, (3.59)) -- solve(R, e) returns R^(-1) %*% e, but should be more accurate and stable
  p <- pchisq(x2, df=df, lower=FALSE)   # x2 has this chi-squared distribution under H0 (p. 119f, following (3.59))
  return(data.frame(X2=x2, df=df, p=p))
}

## plot development of relative frequency spectrum (Vm/V) for increasing sample size N
##   relative.spectrum.plot(obs, m.max=5, log=FALSE, ...)
## obs   = data from lexstats .obs file (which lists V and V1,...,V5 for various sample sizes N)
## m.max = show relative spectrum elements up to m.max
## log   = if TRUE, use logarithmic scale for y-axis
## lty   = can be used to override default line styles (specify at least 5 entries)
## col   = can be used to override default line colours (specify at least 5 entries)
## ...   = other parameters are passed to plot() function 
relative.spectrum.plot <- function (obs, m.max=5, log=FALSE, lwd=2,
                                    lty=c("solid", "dashed", "12", "solid", "dotdash"),
                                    col=c("black", "black", "black", "grey60", "black"), ...) {
  alpha <- list(obs$V1/obs$V, obs$V2/obs$V, obs$V3/obs$V, obs$V4/obs$V, obs$V5/obs$V)
  if (log) {
    ## find suitable range [y.min .. 1] for logarithmic scale
    y.min <- min(sapply(alpha, function (v) { min(v[v > 0]) }), 1e-2)
    y.min <- 10^(floor(log10(y.min)))   # extend range to power of 10
  } else y.min <- 0
  N.max <- max(obs$N)
  if (N.max >= 1e6) {
    x <- obs$N / 1e6
    xlab <- "N (million tokens)"
  } else if (N.max >= 1e4) {
    x <- obs$N / 1e3
    xlab <- "N (thousand tokens)"
  } else {
    x <- obs$N
    xlab <- "N (tokens)"
  }
  ylab <- expression(paste("relative spectrum ", V[m]/V))
  if (log) {
    plot(1, 1, type="n", xlim=c(0,max(x)), ylim=c(y.min,1), log="y",
         xlab=xlab, ylab=ylab, ...)
  } else {
    plot(1, 1, type="n", xlim=c(0,max(x)), ylim=c(0,1), 
         xlab=xlab, ylab=ylab, ...)
  }
  for (i in 1:m.max) {
    lines(x, alpha[[i]], lty=lty[i], lwd=lwd, col=col[i])
  }
  legend(max(x), 1, xjust=1, yjust=1, bg="white",
         legend=paste("m =", 1:m.max),
         lty=lty[1:m.max], col=col[1:m.max], lwd=lwd+1)
}

## plot comparison of two vocabulary growth curves
##   vgc.comparison(N, V, EV, difference=FALSE, m=NULL, log=FALSE, lwd, ...)
## N   = vector of sample sizes N
## V   = vector of observed numbers of types V(N)
## EV  = vector of expected numbers of types E[V(N)]
## difference = if TRUE, plot relative difference (V(N)-E[V(N)]) / E[V(N)]
## m   = if m >= 1, V and EV represent spectrum elements V_m and E[V_m]
## log = if TRUE, use logarithmic scale (both axes, or x-axis when difference=TRUE)
## lwd, lty, col = control appearance of lines, must specify two values each
## pch = plot symbol for observed data points 
## ... = other parameters are passed to plot() function
vgc.comparison <- function (N, V, EV, difference=FALSE, m=NULL, log=FALSE,
                            lwd=c(1, 3), lty=c("solid", "dashed"), pch=20, col=c("black", "black"), ...) {
  N.max <- max(N)
  V.max <- max(V, EV)
  ## choose appropriate scale on x-axis 
  if (N.max >= 1e6) {
    x <- N / 1e6; xlab <- "N (million tokens)"
  } else if (N.max >= 1e4) {
    x <- N / 1e3; xlab <- "N (thousand tokens)"
  } else {
    x <- N ;      xlab <- "N (tokens)"
  }
  if (log) x.min <- 10^(floor(log10(min(x)))) else x.min <- 0
  x.max <- max(x)
  ## choose appropriate scale on y-axis
  if (difference) {
    y <- 100 * (V - EV) / EV
    y.range <- 10 * ceiling(max(abs(y)) / 10)
    y.min <- -y.range                   # show symmetric range around 0% on y-axis
    y.max <- +y.range
    ylab <- "relative difference (%)"
  } else {
    if (V.max >= 1e6) {
      y1 <- V / 1e6; y2 <- EV / 1e6
      ylab <- "million types"
    } else if (N.max >= 1e4) {
      y1 <- V / 1e3; y2 <- EV / 1e3
      ylab <- "thousand types"
    } else {
      y1 <- V ;      y2 <- EV
      ylab <- "types"
    }
    if (log) y.min <- 10^(floor(log10(min(y1, y2)))) else y.min <- 0
    y.max <- max(y1, y2)
  }
  ## plot vocabulary growth curves (or their relative difference)
  log.arg <- ""
  if (difference) {                     # difference=TRUE
    if (log) log.arg <- "x"
    plot(x, y, type="l", lwd=lwd[1], lty=lty[1], col=col[1],
         log=log.arg, xlim=c(x.min, x.max), ylim=c(y.min, y.max),
         xlab=xlab, ylab=ylab, ...)
    abline(h=0, lty="dotted", lwd=1)
    if (missing(m)) {
      legend.txt <- expression(frac(V(N) - paste(E, group("[",V(N),"]")), paste(E, group("[",V(N),"]"))))
    } else {
      legend.txt <- eval(substitute(expression(frac(V[m](N) - paste(E, group("[",V[m](N),"]")), paste(E, group("[",V[m](N),"]")))), list(m=m)))
    }
    legend(x.max, y.min, xjust=1, yjust=0, bg="white", y.intersp=1, adj=c(0,0.3),
           legend=legend.txt, lty=lty[1], lwd=lwd[1], col=col[1])
  } else {                              # difference = FALSE
    if (log) log.arg <- "xy"
    plot(x, y1, type="o", pch=pch, lty=lty[1], lwd=lwd[1], col=col[1],
         log=log.arg, xlim=c(x.min, x.max), ylim=c(y.min, y.max),
         xlab=xlab, ylab=ylab, ...)
    lines(x, y2, lty=lty[2], lwd=lwd[2], col=col[2])
    if (missing(m)) {
      legend.txt <- expression(V(N), paste(E, group("[",V(N),"]")))
    } else {
      legend.txt <- eval(substitute(expression(V[m](N), paste(E, group("[",V[m](N),"]"))), list(m=m)))
    }
    legend(x.max, y.min, xjust=1, yjust=0, bg="white",
           legend=legend.txt, lty=lty, lwd=lwd, col=col, pch=c(pch,-1))         
  }  
}



