##
##  UCS / The R ToolBox
##

## Filename: plots.R
## Modified: Sun May 15 13:56:16 2005 (evert)   
##   Author: Stefan Evert
##  Purpose: High-level commands for precision and recall plots


ucs.library("base")
ucs.library("sfunc")


## PAR$col     = vector of colours for precision graphs (etc.)
## PAR$lty     = vector of line types for precision graphs
## PAR$lwd     = vector of line widths of precision graphs
## PAR$bw.col  = b/w mode: vector of colours for precision graphs
## PAR$bw.lty  = b/w mode: vector of line types for precision graphs
## PAR$bw.lwd  = b/w mode: vector of line widths of precision graphs
## PAR$n.first = smallest n for which n-best list is computed (default: 100)
## PAR$n.step  = step width for n-best lists (default: 1)
## PAR$cex     = character expansion factor (passed to whatever output device a plot function uses)
## PAR$lex     = line width expansion (_added_ to all lwds in plot commands)
## PAR$do.file = if FALSE, evaluation.file() will produce on-screen versions only (for testing purposes) 
##   global configuration parameters for precision and recall plots
##   (can be modified directly by assigning to the parameters; reload "plots.R" to recover defaults)
.ucs.PAR <-
  list(col = rep(c("#000000", "#FF0000", "#0000FF", "#008800", "#FF6600"), 2),
       lty = c(rep("solid", 5), rep("dashed", 5)),
       lwd = rep(2, 10),
       bw.col = c(rep("black", 4), rep("#666666", 4), "black", "#666666"),
       bw.lty = c("solid","33","11","dotdash","solid","33","11","dotdash", "longdash","longdash"),
       bw.lwd = c(2,3,4,3, 3,3,4,3, 3,3),
       n.first = 100, n.step = 1, test.step = 10,
       cex = 1, lex = 0,
       do.file = TRUE)

## ucs.par("n.step")
## ucs.par(n.step=5, test.step=10)
##   functional interface to UCS default graphics parameters
ucs.par <- function (...) {
  args <- list(...)
  if (! length(args)) {                 # ucs.par()
    return(.ucs.PAR)
  }
  else if (is.null(names(args))) {      
    if (length(args) == 1 && is.list(args[[1]])) {  # ucs.par(list(n.first=50, n.step=2)), used to reset saved values
      args <- args[[1]]                 # then fall through to default case below
    } else {                            # ucs.par("n.first", "n.step") or ucs.par(c("n.first", "n.step"))
      pars <- as.character(unlist(args))
      return( if (length(pars) == 1) .ucs.PAR[[ pars[1] ]] else .ucs.PAR[pars] )
    }
  }
  known.pars <- names(.ucs.PAR)         # ucs.par(n.first=50, n.step=2, ...)
  invalid <- !(names(args) %in% known.pars)
  if (any(invalid)) {
    warning("invalid UCS graphics parameter(s) ", paste(names(args)[invalid], collapse=", "), " ignored")
    args <- args[! invalid]
  }
  old <- .ucs.PAR[names(args)]
  for (i in names(args)) { .ucs.PAR[[i]] <<- args[[i]] }
  return(invisible(old))
}


## check.pars()
##   used internally to ensure that .ucs.PAR settings are sensible
check.pars <- function ( ) {
  .ucs.PAR$col <<- rep(.ucs.PAR$col, length=10)
  .ucs.PAR$lty <<- rep(.ucs.PAR$lty, length=10)
  .ucs.PAR$lwd <<- rep(.ucs.PAR$lwd, length=10)
  .ucs.PAR$bw.col <<- rep(.ucs.PAR$bw.col, length=10)
  .ucs.PAR$bw.lty <<- rep(.ucs.PAR$bw.lty, length=10)
  .ucs.PAR$bw.lwd <<- rep(.ucs.PAR$bw.lwd, length=10)
  # might run some sanity checks here
}


## evaluation.plot(ds=glaw, keys=c("freq", "log.l", "t.score", ...) [, <options>])
##   plot combined precision, recall, or precision-by-recall graphs for up to 10 association measures
##   options:
##     ds         = data set for evaluation, must include rankings and b.TP variable
##     keys       = names of the AMs to be evaluated, may be abbreviated to unique prefix (in <ds>)
##     tp         = vector of TPs (TRUE, FALSE, NA), must match data set (default: ds$b.TP)
##     y.max      = max. precision value shown on y axis (determines aspect of view)
##     y.min, x.min, x.max = select plot region (for zooming in)
##     n.first    = overrides PAR$n.first (smallest n for which n-best list is computed)
##     n.step     = overrieds PAR$n.step (step width for n-best lists)
##     x.axis     = "n.best" (default), "proportion", "recall"
##     y.axis     = "precision" (default), "local.precision", "recall"
##     window     = approx. size of window from which local precision is estimated (default: 400)
##     show.baseline = show baseline precision? (default: TRUE)
##     show.nbest = mark these n-best lists with vertical lines
##     show.npair = show as number of pair types, or nothing for show.ncand=NULL (default: size of data set)
##     title      = main title of plot (optional)
##     test       = test significance of differences, argument specifies sig. level (default: 0.05 for test=TRUE)
##     test.am1   = first AM in test  (H1 = AM1 is better than AM2, i.e. two-sided test)
##     test.am2   = second AM in test
##     test.step  = step size for tests (multiples of n.step; default: 10)
##     test.relevance = (absolute) difference precision that is considered relevant (percentage points)
##     conf       = show confidence region, argument specifies sig. level (default: 0.05 = 95% confidence for conf=TRUE)
##     conf.am    = AM for which confidence region is shown
##     conf.am2   = optional second confidence region
##     cut        = consider first <cut> candidates from each ranking only (but show as 100%)
##     file       = save plot to EPS file (does not display on screen), argument specifies filename
##     aspect     = when saving to EPS file, enforce this aspect between y- and x-percentages (defaults to square plot)
##     plot.width = width of plot region in EPS file (defaults to 6 inches)
##     plot.height = height of plot region in EPS file (may be computed from aspect value)
##     cex        = character expansion factor for labels and legend box (default: 1.0)
##     lex        = line expansion, added to all line widths (default: 0)
##     bw         = display in b/w mode? (default: FALSE)
##     usercode   = callback function which can be used to add something to the plot before the legend box is drawn;
##                  invoked with parameters region=c(x.min, x.max, y.min, y.max) and
##                  pr=<list of precision.recall() data for selected AMs, indexed by AM key>
##     legend     = vector of strings or expressions to show in legend (parallel to keys);
##                  use legend=NULL to suppress legend box
##     bottom.legend = if TRUE, show legend box at bottom right rather than top right
##   [further named parameters are interpreted as local graphics settings and are passed to the par() function]
##   [names of AMs for significance test and confidence region may be abbreviated to unique prefixes]
evaluation.plot <- function (ds, keys, tp=NULL,
                             y.max=100, x.max=100, y.min=0, x.min=0,
                             n.first=ucs.par("n.first"), n.step=ucs.par("n.step"),
                             x.axis=c("n.best", "proportion", "recall"),
                             y.axis=c("precision", "local.precision", "recall"),
                             window=400,
                             show.baseline=TRUE,
                             show.nbest=NULL,
                             show.npair=NULL,
                             title=NULL,
                             test=FALSE, test.am1=NULL, test.am2=NULL, test.step=ucs.par("test.step"), test.relevant=0,
                             conf=FALSE, conf.am=NULL, conf.am2=NULL,
                             cut=NULL,
                             bw=FALSE,
                             usercode=NULL,
                             file=NULL, aspect=1, plot.width=6, plot.height=6,
                             cex=ucs.par("cex"), lex=ucs.par("lex"),
                             legend=NULL, bottom.legend=FALSE,
                             ...          # extra arguments are interpreted as graphics parameters
                             ) {
  
  x.axis <- match.arg(x.axis)
  x.percentage <- if (x.axis == "n.best") FALSE else TRUE # whether x-axis shows a percentage
  y.axis <- match.arg(y.axis)
  
  ## set default significance levels for difference test and confidence region
  if (test==TRUE) test <- 0.05
  if (conf==TRUE) conf <- 0.05

  ## check whether requested AMs are annotated and expand abbreviations
  keys <- ds.match.am(ds, keys, rank=TRUE) # stop()s unless all keys have unique matches
  n.keys <- length(keys)                # number of AMs in this plot

  ## check vector of TPs and set default if unspecified
  if (is.null(tp)) {                    # so one can also pass tp=NULL explicitly for default behavioue
    if (!("b.TP" %in% names(ds))) {
      stop('You must define TPs for your data set first! (variable named "b.TP")')
    }
    tp <- ds$b.TP
  }
  if (length(tp) != nrow(ds)) stop("vector of TPs doesn't match data set (different lengths)")

  ## if user has asked for significance tests or confidence region, check that
  ## AMs are specified and resolve them in <keys> (using partial matching if necessary)
  if (conf) {
    if (y.axis != "precision") stop ("confidence regions are only supported for precision graphs")
    if (missing(conf.am)) {
      stop("Missing argument conf.am for confidence region.")
    }
    idx <- pmatch(conf.am, keys)
    if (is.na(idx)) {
      stop(paste("Can't resolve conf.am=<",conf.am,"> in list of AM keys.", sep=""))
    }
    else {
      conf.am <- keys[idx]
    }
    if (! missing(conf.am2)) {
      idx <- pmatch(conf.am2, keys)
      if (is.na(idx)) {
        stop(paste("Can't resolve conf.am2=<",conf.am,"> in list of AM keys.", sep=""))
      }
      else {
        conf.am2 <- keys[idx]
      }
    }
  }
  if (test) {
    if (y.axis == "local.precision") stop ("significance tests cannot be combined with local precision estimates")
    if (missing(test.am1)) {
      stop("Missing argument test.am1 for significance tests.")
    }
    if (missing(test.am2)) {
      stop("Missing argument test.am2 for significance tests.")
    }
    idx.am1 <- pmatch(test.am1, keys)
    if (is.na(idx.am1)) {
      stop(paste("Can't resolve test.am1=<",test.am1,"> in list of AM keys.", sep=""))
    }
    else {
      test.am1 <- keys[idx.am1]
    }
    idx.am2 <- pmatch(test.am2, keys)
    if (is.na(idx.am2)) {
      stop(paste("Can't resolve test.am2=<",test.am2,"> in list of AM keys.", sep=""))
    }
    else {
      test.am2 <- keys[idx.am2]
    }
  }

  ## precompute and save precision and recall data for further tasks (incl. sig. tests and conf. region)
  pr.data <- list()
  for (i in 1:n.keys) {
    ## compute precision and recall values for current AM (with optional cut)
    cut.or.null <- if (missing(cut)) 0 else cut
    window.or.null <- if (y.axis == "local.precision") window else 0
    pr <- precision.recall(ds, keys[i], tp=tp, first=n.first, step=n.step, cut=cut.or.null, window=window.or.null)
    ## x-coordinates are n-best, proportion or recall values, depending on mode
    pr$x <- switch(x.axis, n.best=pr$n, proportion=pr$perc, recall=pr$recall)
    ## y-coordinates are precision or local precision values, depending on local.precision
    pr$y <- switch(y.axis, precision=pr$precision, local.precision=pr$lp, recall=pr$recall)
    pr.data[[ keys[i] ]] <- pr
  }

  ## set x.max default when showing n-best lists and generate label for x-axis
  n.pair <- max(pr.data[[1]]$n)         # not the most efficient way to find out ...
  if (x.axis == "proportion") {         # PROPORTION mode
    xlab.text <- "proportion of ranked data set (%)"
    if (! is.null(show.npair)) {
      n.pair <- show.npair                # override for evaluation of random samples
    }
    if (missing(show.npair) || (! is.null(show.npair))) {
      ## show number of pair types unless show.ncand=NULL was specified
      xlab.text <- paste(xlab.text, "[of", n.pair, "pair types]")
    }
  } else if (x.axis == "recall") {      # RECALL mode
    xlab.text <- "recall (%)"
  } else {
    xlab.text <- "n-best list"          # N-BEST mode
    if (missing(x.max)) x.max <- n.pair
  }

  ## determine appropriate y.max unless it has been specified explicitly
  if (missing(y.max)) {
    if (y.axis == "recall") {
      y.max <- 100
    } else {
      y.max <- max(sapply(pr.data, function (l) max(l$y) ))
      y.max <- 5 * ceiling((y.max + 1) / 5) # round up to multiple of 5%
    }
  }
  ## and generate label for y-axis
  ylab.text <- switch(y.axis,
                      precision="precision (%)",
                      local.precision=paste("local precision (%) [W=", window, "]", sep=""),
                      recall="recall (%)")

  ## screen coordinate units (so image is 1 unit wide and 1 unit high)
  delta.x <- x.max - x.min
  delta.y <- y.max - y.min

  ## localise changes to graphics parameters (e.g. cex)
  par.save  <- par(list(...))
  on.exit(par(par.save))                
  
  ## if parameter <file> is specified, write plot to EPS file 
  if (! missing(file)) {
    if (! missing(aspect)) {            # enforce specified aspect
      if (! missing(plot.height))
        warning("aspect parameter overrides explicit plot.height setting")
      if (x.percentage) {
        ## x-axis shows a percentage => aspect gives relation between percent points on x- and y-axis (normally 1:1)
        plot.height <- aspect * (delta.y / delta.x) * plot.width
      } else {
        ## x-axis shows n-best list => aspect gives relation between width and height of plot region
        plot.height <- aspect * plot.width
      }
    }
    ## now define standard margins around plot region, depending on character expansion factor
    margins <- c(0.88, 0.88, 0.88, 0.42) # bottom, left, top, right
    if (missing(title)) margins[3] <- 0.42
    margins <- cex * margins
    ## obtain size of device (i.e. paper size) by adding margins to plot region
    dev.width <- plot.width + margins[2] + margins[4]
    dev.height <- plot.height + margins[1] + margins[3]
    postscript(file=file, width=dev.width, height=dev.height,
               paper="special", horizontal=FALSE, onefile=FALSE)
    par(mai=margins)                    # set our standard margins
  }

  ## local settings for some graphics parameters
  par(cex=cex)
  
  ## read default line styles (colour or b/w)
  check.pars()
  if (bw) {
    col <- ucs.par("bw.col")
    lty <- ucs.par("bw.lty")
    lwd <- ucs.par("bw.lwd")
  } else {
    col <- ucs.par("col")
    lty <- ucs.par("lty")
    lwd <- ucs.par("lwd")
  }
  
  ## set up plot region and axes without plotting any curves
  plot(0,0, type="n", xlim=c(x.min, x.max), ylim=c(y.min, y.max),
       xlab=xlab.text, ylab=ylab.text,
       xaxs="i", yaxs="i",
       main=title)
  
  ## show confidence region as grey area in background
  if (conf) {
    pr <- pr.data[[conf.am]]            # retrieve precision/recall data for the requested AM
    upper <- 100 * binom.conf.interval(pr$tp, pr$tp+pr$fp, limit="upper", conf.level=conf)
    lower <- 100 * binom.conf.interval(pr$tp, pr$tp+pr$fp, limit="lower", conf.level=conf)
    if (bw) {
      polygon(c(pr$x, rev(pr$x)), c(upper, rev(lower)), # encircle conf. region outline in clockwise direction
              density=10, col="black", border="black") # shade region diagonally in b/w mode
    } else {
      polygon(c(pr$x, rev(pr$x)), c(upper, rev(lower)), 
              col="#AAAAAA", border="black") # draw region in light grey with black outline in colour mode
    }

    if (! missing(conf.am2)) {
      pr <- pr.data[[conf.am2]]
      upper <- 100 * binom.conf.interval(pr$tp, pr$tp+pr$fp, limit="upper", conf.level=conf)
      lower <- 100 * binom.conf.interval(pr$tp, pr$tp+pr$fp, limit="lower", conf.level=conf)
      if (bw) {
        polygon(c(pr$x, rev(pr$x)), c(upper, rev(lower)), # encircle conf. region outline in clockwise direction
                density=10, angle=-45, col="black", border="black") # shade second region orthogonal to first
      } else {
        polygon(c(pr$x, rev(pr$x)), c(upper, rev(lower)), 
                col="#FFFFAA", border="black") # draw second region in light yellow with black outline
      }
    }
  }

  if (show.baseline || (x.axis == "recall" && !is.null(show.nbest))) {
    ## baseline precision (percentage), also required for n-best markers in RECALL mode
    tp.na <- is.na(tp)             
    if (sum(tp.na) > 0) { # missing annotations indicate random sample evaluation
      tmp.tp <- sum(tp, na.rm=TRUE) # compute baseline from annotated candidates + conf. interval
      tmp.n <- sum(! tp.na)
      tmp.conf <- if (conf) conf else 0.05 # use same conf. level as for confidence intervals, if selected
      baseline <- 100 * tmp.tp / tmp.n
      baseline.min <- 100 * binom.conf.interval(tmp.tp, tmp.n, limit="lower", conf.level=tmp.conf)
      baseline.max <- 100 * binom.conf.interval(tmp.tp, tmp.n, limit="upper", conf.level=tmp.conf)
    } else {
      baseline <- 100 * sum(tp) / length(tp) # compute baseline from all candidates
      baseline.min <- 0
      baseline.max <- 0
    }
  }
  if (show.baseline && (y.axis != "recall")) {
    abline(h=baseline, lty="12", col="black", lwd=2+lex)
    if (baseline.max > 0) {
      abline(h=baseline.min, lty="12", col="black", lwd=1+lex)
      abline(h=baseline.max, lty="12", col="black", lwd=1+lex)
    }
    ## print baseline prec. value on the right (-5%) below the baseline (-3%)
    txt.x <- x.max - 0.05 * delta.x
    txt.y <- baseline - 0.02 * delta.y
    text(txt.x, txt.y, adj=c(1,1),
         labels=paste("baseline = ", round(baseline, dig=2), "%", sep=""))
  }
  
  ## plot precision graphs for all AMs
  for (i in 1:n.keys) {
    pr <- pr.data[[ keys[i] ]]          # retrieve precision/recall data for current AM
    ## plot precision values with corresponding lty and col
    points(pr$x, pr$y, type="l", lwd=lwd[i]+lex, lty=lty[i], col=col[i])
  }

  ## if show.nbest was specified, indicate the requested n-best lists by vertical or diagonal lines
  if (! is.null(show.nbest)) {
    for (n in show.nbest) {
      if (x.axis == "proportion") {     # PROPORTION mode
        pr <- pr.data[[1]] # get corresponding x-axis location from any one of the precision/recall tables
        idx <- sum(pr$n <= n) # this computes the largest row in pr for which the <n> column value is <= n
        abline(v=pr$x[idx], lwd=1, col="black")
      } else if (x.axis == "recall") {  # RECALL mode
        if (y.axis != "recall") { # skip in recall-by-recall plots :o)
          slope <- (baseline / 100) * n.pair / n # y * n = x * TPs, with TPs = baseline * n.pair
          abline(0, slope, lwd=1, col="black")
        }
      } else {                          # N-BEST mode
        abline(v=n, lwd=1, col="black")
      }
    }
  }
  
  ## perform tests for significance of differences between two measures
  yellow.triangle <- if (!bw) "yellow" else "#AAAAAA"  # define fill colour for yellow & red triangles
  yellow.arrow <- if (!bw) "#DDAA00" else "black"
  yellow.lwd <- if (!bw) 1+lex else 1
  red.triangle <- if (!bw) "red" else "black"
  red.arrow <- if (!bw) "#CC0000" else "black"
  red.lwd <- if (!bw) 1+lex else 2
  if (test) {
    steps <- length(pr.data[[1]]$n)     # perform tests for subset of n-best lists shown in graphs
    rank.am1 <- ds[[ am.key2rvar(test.am1) ]] # need rankings of the TPs for McNemar's test
    rank.am2 <- ds[[ am.key2rvar(test.am2) ]]
    if (x.axis == "recall") {           # in RECALL mode, ensure roughly equal angles between arrows
      tmp.k <- ceiling(steps / test.step) # number of arrows
      tmp.pr <- pr.data[[1]]
      tmp.sl1 <- tmp.pr$prec[1] / tmp.pr$recall[1] # slopes of first and last n-best line for AM1
      tmp.sl2 <- tmp.pr$prec[steps] / tmp.pr$recall[steps]
      tmp.a1 <- atan(tmp.sl1)           # corresponding angles
      tmp.a2 <- atan(tmp.sl2)
      tmp.angles <- rev(seq(tmp.a1,tmp.a2,length=tmp.k)) # compute slopes at roughly equal angles
      tmp.slopes <- tan(tmp.angles)     # ranges from tmp.sl1 .. tmp.sl2 -> scale its inverse to 2 .. steps
      tmp.x <- 1/tmp.slopes
      tmp.x <- tmp.x - min(tmp.x)
      i.vals <- floor(tmp.x * (steps - 2) / max(tmp.x)) + 2 
    } else {
      i.vals <- seq(1, steps, by=test.step)
    }
    for (i in i.vals) {
      n <- pr.data[[test.am1]]$n[i]   # use same n-best lists as for the curves
      x1 <- pr.data[[test.am1]]$x[i]  # x-coordinates for both AMs (identical except in RECALL mode)
      x2 <- pr.data[[test.am2]]$x[i]
      y1 <- pr.data[[test.am1]]$precision[i] # y-coordinates for both AMs (= precision values)
      y2 <- pr.data[[test.am2]]$precision[i]
      if (pmax(x1,x2) < x.min || pmin(x1,x2) > x.max) next # no need to run expensive tests off-screen
      ## compute difference sets: A = AM1 accepts, but AM2 doesn't; B = AM2 accepts, but AM1 doesn't
      idx.A <- (rank.am1 <= n & rank.am2 > n)
      idx.B <- (rank.am1 > n & rank.am2 <= n)
      set.A <- tp[idx.A]
      set.B <- tp[idx.B]
      ## now count TPs and FPs in difference sets (excluding NA's if necessary)
      tp.A <- sum(set.A, na.rm=TRUE)
      fp.A <- sum((! set.A), na.rm=TRUE)
      tp.B <- sum(set.B, na.rm=TRUE)
      fp.B <- sum((! set.B), na.rm=TRUE)
      ## apply Fisher's exact test to this 2x2 table (which is expensive, but gives accurate results)
      if (sum(tp.A+fp.A) > 0 && sum(tp.B+fp.B) > 0) {
        result <- fisher.test(cbind(c(tp.A,fp.A), c(tp.B,fp.B)))
        if (result$p.value < test) { # yellow triangle = significant evidence for some difference
          if (x.axis == "recall") {
            arrows(x1, y1, x2, y2, col=yellow.arrow, lwd=yellow.lwd, length=0.1, angle=15)
          } else {
            points(x1, y1, col="black", bg=yellow.triangle, pch=25) # down triangle for AM1
            points(x2, y2, col="black", bg=yellow.triangle, pch=24) # up triangle for AM2
          }
        }
        if (test.relevant > 0) { # red triangle = significant evidence for relevant difference 
          ## use a wild & complicated estimate for the difference between the number of TPs
          odds <- max(result$conf.int[1], 1/result$conf.int[2]) # exchange AM1 and AM2 -> odds = 1 / odds
          if (odds > 1) {               # should be the case iff there is a significant difference
            diff.N <- (sum(idx.A) + sum(idx.B)) / 2 # number of candidates in each difference set (should be identical)
            diff.C <- (tp.A + tp.B) / (2 * (tp.A + fp.A + tp.B + fp.B)) * diff.N # expected average number of TPs
            diff.coef <- (odds + 1) / (odds - 1) # diff.d = estimated difference in number of TPs
            diff.d <- diff.coef * diff.N - sqrt( diff.coef^2 * diff.N^2 - 4 * diff.C * (diff.N - diff.C) )
            diff.prec <- 100 * diff.d / n     # estimated difference in precision values (% points)
            if (diff.prec >= test.relevant) {
              ## this is for debugging; perhaps make it optional with verbose flag or so
              cat("RELEVANT: N =",diff.N,"  C =",diff.C,"  d =",diff.d,"  n =",n,"  D(p) =", diff.prec, "  odds =",odds,"\n")
              if (x.axis == "recall") {
                arrows(x1, y1, x2, y2, col=red.arrow, lwd=red.lwd, length=.1, angle=15)
              } else {
                points(x1, y1, col="black", bg=red.triangle, pch=25)
                points(x2, y2, col="black", bg=red.triangle, pch=24)
              }
            }
          }
        }
      }
    }
  }

  ## execute user callback function, which can add something to the plot before it is written to an EPS file
  if (! is.null(usercode)) {
    usercode(region=c(x.min, x.max, y.min, y.max), pr=pr.data)
  }

  ## default legend just uses the AM keys
  if (missing(legend)) {
    legend <- keys
  }

  if (length(legend) > 0) {
    ## check that number of legend strings is the same as the number of AM keys
    if (length(legend) > n.keys) warning("too many text strings in legend argument, extra ones will be ignored")
    if (length(legend) < n.keys) warning("not enough text strings in legend argument, legend box will be incomplete")
    length(legend) <- n.keys            # discard surplus entries, or pad with NA's 

    ## simple legend box (move in from upper right corner by 5%)
    leg.x <- x.max - 0.05 * delta.x
    leg.y <- if (bottom.legend) y.min + 0.05 * delta.y else y.max - 0.05 * delta.y
    leg.yjust <- if (bottom.legend) 0 else 1
    leg.pch  <- rep(-1, n.keys)           # if sig.test results are shown, display corresp. triangles in legend
    leg.bg <- rep("black", n.keys)        # background colours for triangle symbols
    if (test) {
      leg.pch[c(idx.am1, idx.am2)] <- c(25,24)
      leg.bg[c(idx.am1, idx.am2)] <- rep(yellow.triangle, 2)
    }
    legend(leg.x, leg.y, xjust=1, yjust=leg.yjust, # top or bottom right-hand corner
           legend=legend,               # show AM keys in legend
           pch=leg.pch, pt.bg=leg.bg,
           lwd=3+lex, lty=lty[1:n.keys], col=col[1:n.keys],
           bg="white")
  }
    
  ## if writing to EPS file, close file and associated device now
  if (! missing(file)) {
    dev.off()                           # close EPS file device
  }
}

## evaluation.file(ds=glaw, keys=c("freq", "log.l", "t.score", ...), file="plot.eps" [, bw=FALSE, <options>])
##  convenience function to avoid double function calls for screen and EPS versions;
##  calls evaluation.plot() twice, with and without file argument; if bw option is not specified,
##  it is automatically set to FALSE for screen version and TRUE for EPS version;
##  note that the EPS version is skipped if PAR$do.file is FALSE (for testing purposes)
evaluation.file <- function (ds, keys, file, bw=NULL, ...) {
  screen.bw <- if (missing(bw)) FALSE else bw
  file.bw <- if (missing(bw)) TRUE else bw
  evaluation.plot(ds, keys, bw=screen.bw, ...) # on-screen version
  if (ucs.par("do.file") && !missing(file)) evaluation.plot(ds, keys, bw=file.bw, file=file, ...) # EPS version
}


## precision.plot(ds=glaw, keys=c("freq", "log.l", "t.score", ...) [, <options>])
## recall.plot(ds=glaw, keys=c("freq", "logl", "tscore", ...) [, <options>])
## recall.precision.plot(ds=glaw, keys=c("freq", "logl", "tscore", ...) [, <options>])
##   these functions names are provided for backward compatibility as aliases of precision.plot() 
precision.plot <- function ( ds, keys, ... ) { evaluation.plot(ds, keys, ...) }
recall.plot <- function ( ds, keys, y.axis="recall", ... ) {
  if (! missing(y.axis)) stop ("parameter 'y.axis' not allowed in recall plot")
  evaluation.plot(ds, keys, y.axis="recall", ...)
}
recall.precision.plot <- function ( ds, keys, x.axis="recall", y.axis="precision", ... ) {
  if (! missing(x.axis)) stop ("parameter 'x.axis' not allowed in precision-by-recall plot")
  if (! missing(y.axis)) stop ("parameter 'y.axis' not allowed in precision-by-recall plot")
  evaluation.plot(ds, keys, x.axis="recall", y.axis="precision", ...)
}

