##
##  UCS / The R ToolBox
##

## Filename: zm.R
## Modified: Sun May 15 13:56:33 2005 (evert)   
##   Author: Stefan Evert
##  Purpose: The Zipf-Mandelbrot LNRE model described in (Evert 2004)

## References:
## Evert, Stefan (2004).  A Simple LNRE Model for Random Character Sequences.
##   In: Proceedings of the JADT 2004 Conference, Louvain-la-Neuve, Belgium.


ucs.library("sfunc")
ucs.library("lexstats")

## use logit transform from the "boot" library
## (could also easily implement logit() and inv.logit() in R)
library("boot")                         


## create object representing Zipf-Mandelbrot model;
## the constructor can be invoked in several different ways:
##   model <- zm(alpha=0.9, C=0.1)                 ...  specify parameters explicitly
##   model <- zm(alpha=0.8, N=1e6, V=77800)        ...  estimate C from N and V
##   model <- zm(spc=c(V1,V2,...), N=1e6, V=77800) ...  estimate both parameters from obs. spectrum
## (N, V, and spc can be specified in addition to explicit parameters and
## will be stored in the returne object for use by the write.lexstats() function)
zm <- function (alpha=NULL, C=NULL, N=NULL, V=NULL, spc=NULL, m.max=15, stepmax=10, debug=FALSE) {
  model <- list()
  obs.ok <- (!missing(N)) && (!missing(V))  # observed data available? (N and V)
  spc.ok <- !missing(spc)               # spectrum available (spc=c(V1,V2,...), for sample size N)
  if (obs.ok) {
    model$N <- N
    model$V <- V
    if (spc.ok) model$spc <- spc
  }
  if (missing(alpha)) {
    if (spc.ok && obs.ok) {
      alpha <- spc[1] / V               # direct estimate for alpha
      ## perform parameter optimisation, using up to m.max spectrum elements
      ## (but stop before the first rank m with Vm < 100)
      m.max <- min(length(spc), m.max)
      spc.exclude <- (1:m.max)[spc[1:m.max] < 100]
      if (length(spc.exclude) >= 1) m.max <- min(spc.exclude) - 1
      if (m.max < 1) m.max <- 1         # consider hapaxes at least
      ## cost function for non-linear minimsation (multinomial chi-squared statistic)
      cost <- function (alpha.logit, N, V, Vm, m.max, debug) {
        alpha <- inv.logit(alpha.logit)     # parameter alpha is logit-transformed to range [-inf, inf]
        if (alpha <= 0 || alpha >= 1) {     # return +infinity when parameter is out of range (should not happen with alpha.tan)
          if (debug) cat("NLM:  alpha =", alpha, "   (out of range)\n")
          return(Inf)
        }
        temp <- zm(alpha=alpha, N=N, V=V)   # ZM model with specified alpha
        E.Vm <- EVm(temp, 1:m.max, N )      # compute chi-squared statistic (Baayen 2001, 118)
        E.V <- EV(temp, N)
        Vplus <- V - sum(Vm)
        E.Vplus <- E.V - sum(E.Vm)
        X2 <- sum((Vm - E.Vm)^2 / E.Vm) + (Vplus - E.Vplus)^2 / E.Vplus
        if (debug) cat("NLM:  alpha =", alpha, "   X2 =", X2, "\n")
        return(X2)
      }
      ## perform minimisation of cost function, and use optimised parameter estimate if successful
      res <- nlm(cost, logit(alpha), print.level=debug, N=N, V=V, Vm=spc[1:m.max], stepmax=stepmax, m.max=m.max, debug=debug)
      if (res$code <= 3) {
        alpha <- inv.logit(res$estimate) # use estimate from non-linear minimisation
      } else {
        warning("Parameter optimisation failed. Using primary estimate (V1/V).")
        alpha <- spc[1] / V                 # just to be sure (should still be set to this value from initial estimate above)
      }
      if ((alpha <= 0) || (alpha >= 1)) stop("Sorry, automatic estimation of alpha parameter failed (out of valid range).")
    } else {
      stop("You have to specify N, V, and spc for automatic estimation of the alpha parameter.")
    }
  } else {
    if ((alpha <= 0) || (alpha >= 1)) stop("The parameter alpha must be in the range 0 < alpha < 1.")
  }
  model$alpha <- alpha
  if (missing(C)) {
    if (obs.ok) {
      C <- alpha * V / (gamma(1-alpha) * N^alpha)
    } else {
      stop("You have to specify N and V for automatic estimation of the C parameter.") 
    }
  }
  model$C <- C
  model$n.param <- 2                    # number of estimated parameters (for multivariate chi-squared test)
  class(model) <- "zm"                  # bless into "zm" class
  return(model)
}

# print summary of model parameters and observed data stored in the model
# with "print(model)" or simply by typing "model" in an R session 
print.zm <- function (model) {
  cat("\nZipf-Mandelbrot LNRE model.\n")
  B <- ((1 - model$alpha) / model$C)^(1 / (1 - model$alpha))
  cat("Parameters:  alpha =", model$alpha, " C =", model$C, " [ B =", B, "]\n")
  if ((!is.null(model$N)) && (!is.null(model$V))) {
    cat("Observed:  N =", model$N, "   V  =", model$V)
    if (!is.null(model$spc)) {
      cat("     Vm  = ")
      if (length(model$spc) > 5) {
        cat(model$spc[1:5], "...")
      } else {
        cat(model$spc)
      }
    }
    cat("\n")
    cat("Expected:  N =", model$N, " E[V] =", EV(model, model$N))
    cat("   E[Vm] =", EVm(model, 1:5, model$N), "...\n")
  } else {
    cat("(No observed data available.)\n")
  }
  cat("\n")
}

## expected spectrum and conditional parameter distribution of Zipf-Mandelbrot LNRE model, plus variances
##   E[V(N)]                      = EV(model, N)
##   E[V_m(N)]                    = EVm(model, m, N)
##   E[V_m(N)] / E[V(N)]          = EVm(model, m, N, relative=TRUE)
##   E[V_{m+1}(N)] / E[V_m(N)]    = EVm(model, m, N, ratio=TRUE)
##   E[V_{m, rho}(N)]             = EVm(model, m, N, rho=<rho>) 
##   E[V_{m, rho}(N)] / E[V_m(N)] = EVm(model, m, N, rho=<rho>, relative=TRUE) 
##   E[V_{m, >rho}(N)]            = EVm(model, m, N, rho=<rho>, lower=FALSE) 
##   E[V_{m, >rho}(N)] / E[V_m(N)]= EVm(model, m, N, rho=<rho>, relative=TRUE, lower=FALSE) 
EV.zm <- function (model, N) {
  alpha <- model$alpha
  C <- model$C
  return(C * N^alpha * gamma(1 - alpha) / alpha)
}

EVm.zm <- function (model, m, N, rho=1, relative=FALSE, ratio=FALSE, lower=TRUE) {
  m <- as.integer(m)
  if (any(m <= 0)) stop("Error: m must be an integer > 0.")
  alpha <- model$alpha
  C <- model$C
  if (missing(rho)) {
    if (!missing(lower)) warning("'lower' parameter can only be used together with 'rho' (ignored)")
    if (ratio) {
      return( (m - alpha) / (m + 1) )
    } else if (relative) {
      return( alpha * gamma(m - alpha) / (gamma(m + 1) * gamma(1 - alpha)) )
    } else {
      return( C * N^alpha * gamma(m - alpha) / gamma(m + 1) )
    }
  } else {
    if ((rho < 0) || (rho > 1)) stop("Error: rho must be in the range 0 <= rho <= 1.")
    if (relative || ratio) {
      return( Rgamma(m - alpha, N *  rho, lower=lower) )
    } else {
      return( Rgamma(m - alpha, N * rho, lower=lower) * EVm.zm(model, m, N) )
    }
  }
}

VV.zm <- function (model, N) {
  return( EV.zm(model, 2*N) - EV.zm(model, N) )
}

VVm.zm <- function (model, m, N, rho=1, relative=FALSE, lower=TRUE) {
  m <- as.integer(m)
  if (any(m <= 0)) stop("Error: m must be an integer > 0.")
  if (missing(rho)) {
    if (!missing(relative)) stop("Error: can't compute variances for relative spectrum")
    if (!missing(lower)) warning("'lower' parameter can only be used together with 'rho' (ignored)")
    return( EVm.zm(model, m, N) - choose(2*m, m) * 2^(-2*m) * EVm.zm(model, 2*m, 2*N) )
  } else {
    if ((rho < 0) || (rho > 1)) stop("Error: rho must be in the range 0 <= rho <= 1.")
    if (relative) {
      mu.1 <- EVm.zm(model, m, N, rho=rho, lower=TRUE) # (Evert 2004, Appendix A.3)
      mu.2 <- EVm.zm(model, m, N, rho=rho, lower=FALSE)
      var.1 <- VVm.zm(model, m, N, rho=rho, lower=TRUE)
      var.2 <- VVm.zm(model, m, N, rho=rho, lower=FALSE)
      mu <- mu.1 + mu.2
      var <- var.1 + var.2
      r <- if (lower) mu.1 / mu else mu.2 / mu # actually, variance is the same for lower=FALSE and lower=TRUE 
      s <- if (lower) var.1 / var else var.2 / var
      return( (var / mu^2) * ( (r-s)^2 + s*(1-s) ) )
    } else {
      return( EVm.zm(model, m, N, rho=rho, lower=lower) - choose(2*m, m) * 2^(-2*m) * EVm.zm(model, 2*m, 2*N, rho=rho, lower=lower) )
    }
  }
}
  
## generic functions for object-oriented calling syntax
EV <- function(model, N, ...) {
  UseMethod("EV")
}
EVm <- function(model, m, N, ...) {
  UseMethod("EVm")
}
VV <- function(model, N, ...) {
  UseMethod("VV")
}
VVm <- function(model, m, N, ...) {
  UseMethod("VVm")
}

## write necessary files to evaluate goodness-of-fit with lexstats tools
##   write.lexstats(model, file=<prefix>)
## model = ZM model object based on empirical data (N, V, and spc)
## file  = creates files: <prefix>_bZM.spc, <prefix>_bZM.sp2, <prefix>_bZM.ev2
write.lexstats.zm <- function (model, file) {
  N <- model$N
  V <- model$V
  Vm <- model$spc
  if (is.null(N) || is.null(V) || is.null(Vm) || length(Vm) < 15) {
    stop("ZM model must be based on empirical data (N, V, and spectrum of length >= 15 required).")
  }
  file <- paste(file, "_bZM", sep="")  # trick lnreChi2 into assuming two estimated parameters
  Vm <- Vm[1:15]
  ## write .spc file
  EVm <- EVm(model, 1:15, N)
  EalphaM <- EVm(model, 1:15, N, relative=TRUE)
  X <- data.frame(m=1:15, Vm=Vm, EVm=EVm, alphaM=Vm/V, EalphaM=EalphaM)
  write.table(X, file=paste(file, ".spc", sep=""), sep="\t", row.names=FALSE, quote=FALSE)
  ## write .sp2 file
  X <- data.frame(m=1:30, EVm2N=EVm(model, 1:30, 2*N))
  write.table(X, file=paste(file, ".sp2", sep=""), sep="\t", row.names=FALSE, quote=FALSE)
  ## write .ev2 file
  X <- data.frame(V=V, EV=EV(model, N), EV2N=EV(model, 2*N))
  write.table(X, file=paste(file, ".ev2", sep=""), sep="\t", row.names=FALSE, quote=FALSE)
  return(file)                          # returns basename of data files
}

## generic function for object-oriented calling syntax
write.lexstats <- function(model, file, ...) {
  UseMethod("write.lexstats")
}


## plot estimate for proportion of low-frequency types in frequency class m
##   zm.rho.plot(m, alpha=c(.5,.75,.9), xmin=-6, ymin=-6, xmax=2, ymax=1, filename="", ...)
## m        = frequency class
## alpha    = alpha parameter values (defaults to 0.5, 0.75, and 0.9)
## filename = if specified, save plot to EPS file
## xmin, xmax = range on x-axis (expected frequency) in log10 units (default: 1e-6 .. 100)
## ymin, ymax = range on y-axis (p-value / estimate for proportion) in log10 units (default: 1e-6 .. 10)
## lty, col, lwd = configure line styles (should have same length as parameter vector)
zm.rho.plot <- function (m, alpha=c(0.5, 0.75, 0.9),
                         xmin=-6, ymin=-6, xmax=2, ymax=1,
                         lty=c("solid","33","11"), col=c("grey50","black","black"), lwd=c(3,3,4),
                         filename="") {
  log.ef <- seq(xmin-1, xmax+1, by=0.1) # logarithmic expected frequency (base 10)
  ef <- 10 ^ log.ef
  pois.p <- ppois(m-1, lambda=ef, lower=F) # p-value from Poisson test, i.e. P(X >= m)
  ylab <- eval(substitute(expression(paste("p-value / ", E * group("[",R[list(m,rho)],"]"))), list(m=m)))
  par.save <- par(mar=c(5,5,2,2)+.1)
  plot(ef, pois.p, log="xy", lwd=lwd[1], type="l",
       xlim=c(10^xmin, 10^xmax), ylim=c(10^ymin, 10^ymax),
       xlab="expected frequency", ylab=ylab)
  for (i in ymin:ymax) abline(h=10^i, col="black", lty="22")
  for (i in xmin:xmax) abline(v=10^i, col="black", lty="22")
  abline(h=1, col="black", lty="solid", lwd=2)
  abline(v=m, col="black", lty="solid", lwd=2)
  text(m, 10^-1, pos=4, offset=0.5, labels=paste("m =",m), srt=-90)
  for (i in 1:length(alpha)) {
    model <- zm(alpha=alpha[i], C=1)    # expected proportions do _not_ depend on C and N
    prop <- EVm(model, m, 1, rho=ef, relative=TRUE) # with N=1, expected frequency equals prob. parameter pi
    lines(ef, prop, lwd=lwd[i], col=col[i], lty=lty[i])
  }
  legtext <- expression("Poisson test")
  for (i in 1:length(alpha)) {
    legtext <- append(legtext, parse(text=paste("alpha ==",alpha[i])))
  }
  legend(x=10^(xmax-1/4), y=10^(ymin+1/4), xjust=1, yjust=0, legend=legtext,
         lty=c("solid",lty), col=c("black",col), lwd=4, bg="white")
  par(par.save)
  if (filename != "") {
    dev.copy2eps(file=filename)
  }
}
