# -*-cperl-*-
## Filename: adj-n-from-cwb.perl
## Modified: Sun May  8 20:10:31 2005 (evert)   
##   Author: Stefan Evert
##  Purpose: Extract relational Adj+N cooccurrences from a CWB-encoded corpus 
STDERR->autoflush(1);
use sigtrap qw<die normal-signals>, 'handler' => \&broken_pipe, 'PIPE';	# so program exits e.g. when output pipe is broken
$Abort = 0;					# will be set by broken_pipe() handler

use CWB;					# CWB/Perl modules needed
use CQP;

use Getopt::Long;
use locale;

$Help = 0;					# -h = help page
$LexemeAtt = "lemma";				# -l = p-attribute for lexemes
$PosAtt = "pos";				# -p = p-attribute for POS tags
$Registry = undef;				# -r = registry directory 
$Adjacent = 0;					# -a = only adjacent pairs
$Verbose = 0;					# -v = print progress information

$ok = GetOptions(
		 "lexemes|l=s" => \$LexemeAtt,
		 "pos|p=s" => \$PosAtt,
		 "adjacent|a" => \$Adjacent,
		 "registry|r=s" => \$Registry,
		 "verbose|v" => \$Verbose,
		 "help|h" => \$Help,
		 );

die
  "\nUsage:  ucs-tool adj-n-from-cwb <tagset> <CORPUS>\n\n",
  "Extract adjective + noun cooccurrences from a CWB-encoded corpus, where the\n",
  "adjective is an attributive modifier of the noun (rather than a predicate).\n",
  "The corpus must be registered in the IMS Corpus Workbench under the name\n",
  "<CORPUS> (all uppercase), and <tagset> identifies the POS tagset to use.\n",
  "The extraction results are written to STDOUT as a sequence of pair tokens.\n\n",
  "Options:\n",
  "  -l <att>   use p-attribute <att> for lexemes (default: lemma)\n",
  "     --lexemes=<att>\n",
  "  -p <att>   use p-attribute <att> for POS tags (default: pos)\n",
  "     --pos=<att>\n",
  "  -a         extract adjacent adjective-noun pairs only\n",
  "     --adjacent\n",
  "  -r <dir>   set CWB registry directory (default: system setting)\n",
  "     --registry=<dir>\n",
  "  -v         verbose mode (show progress information)\n",
  "     --verbose\n",
  "  -h         show this help page\n",
  "     --help\n\n",
  "Tagsets:\n",
  "  penn    English   Penn treebank\n",
  "  claws   English   CLAWS tagset (BNC)\n",
  "  stts    German    STTS tagset (TreeTagger)\n",
  "\n"
  unless $ok and (not $Help) and @ARGV == 2;

($Tagset, $Corpus) = @ARGV;

# CQP queries used for extraction (depending on tagset)
# $Query{$Tagset} = [ $adj_pattern, $skip_pattern, $noun_pattern ];  ($skip_pattern is excluded in --adjacent mode)
%Query = (
	  "penn" => ['[pos = "JJ.*" & word="[a-z-]+" %cd]', '([pos = ",|CC|JJ.*|RB"]* [pos = "JJ.*"])?', '[pos = "NN.*" & word="[a-z-]+" %cd]'],
	  "claws" => ['[pos = "AJ." & word="[a-z-]+" %cd]', '([word = "," | pos = "CJC|AJ.|AV0"]* [pos = "AJ."])?', '[pos = "NN." & word="[a-z-]+" %cd]'],
	  "stts" => ['[pos = "ADJA" & word="[a-z-]+" %cd]', '[pos != "NN|NE|\$.|VV.*"]{0,9}', '[pos = "NN" & word="[a-z-]+" %cd]'],
	 );

$Tagset = lc $Tagset;
@query = @{$Query{$Tagset}};
$query = ($Adjacent) ? "$query[0] $query[2]" : "@query";

die 
  "Error:  tagset '$Tagset' not supported.\n",
  "        [supported tagsets: ".join(", ", sort keys %Query)."]\n"
  unless defined $query;

# substitute p-attribute for POS tags (if -p switch was used)
$query =~ s/pos\s*=/$PosAtt =/g
  if $PosAtt ne "pos";

# try to be helpful if user forgot to spell corpus name in uppercase
$Corpus = uc $Corpus;

# now spawn CQP process and set up configuration
@flags = (defined $Registry) ? ("-r $Registry") : ();
$cqp = new CQP @flags;
$cqp->set_error_handler(sub {die map {"$_\n"} (@_, "Program aborted.");});
$cqp->exec("set MatchingStrategy traditional");	# allow multiple nested matches
$cqp->exec($Corpus);				# activate source corpus
$cqp->exec("show -word +$LexemeAtt");		# change p-attribute for kwic output (-l flag)
$cqp->exec("set Context 0");			# configure kwic output
$cqp->exec("show -cpos");
$cqp->exec("set LeftKWICDelim ''");
$cqp->exec("set RightKWICDelim ''");

exit 1 if $Abort;

# run CQP query (from %Query)
print STDERR "Running CQP query ...";
$cqp->exec("A = $query");
($N) = $cqp->exec("size A");
print STDERR " $N matches\n";

exit 1 if $Abort;

# download matches (in batches of 10000 each) and print pair types on STDOUT 
print STDERR "Processing "
  if $Verbose;
for ($start = 0; $start < $N; $start += 10000) {
  print STDERR "."
    if $Verbose;
  $end = $start + 9999;
  $end = $N - 1
    if $end >= $N;
  @lines = $cqp->exec("cat A $start $end");
  foreach $line (@lines) {
    @words = split " ", $line;
    die "KWIC format error [$line].\n"
      unless @words >= 2;
    $l1 = shift @words;				# l1 = adjective = first token
    $l2 = pop @words;				# l2 = noun = last token
    print "$l1\t$l2\n";
    last if $Abort;
  }
  last if $Abort;
}
print STDERR " done\n"
  if $Verbose;



##
##  subroutines
##

sub broken_pipe {
  print STDERR "[output pipe broken, exiting prematurely]\n"
    if $Verbose and not $Abort;
  $Abort = 1;
}
