# -*-cperl-*-
## Filename: segment-from-cwb.perl
## Modified: Sun May  8 20:10:39 2005 (evert)   
##   Author: Stefan Evert
##  Purpose: Extract segment-based cooccurrences from a CWB-encoded corpus 
STDERR->autoflush(1);
use sigtrap qw<die normal-signals>, 'handler' => \&broken_pipe, 'PIPE';	# so program exits e.g. when output pipe is broken

use CL;						# CWB/Perl modules needed
use CQP;

use Getopt::Long;
use locale;

$Help = 0;					# -h = help page
$LexemeAtt = "lemma";				# -l = p-attribute for lexemes
$PosAtt = "pos";				# -p = p-attribute for POS tags
$Tag1 = ".*";					# -t1 = POS tag for first components (CWB regexp)
$Tag2 = ".*";					# -t2 = POS tag for second components (CWB regexp)
$Clean = "";					# -c = clean up data by matching against Perl regexp
$Frequency = 0;					# -f = cull data for frequency threshold
$Registry = undef;				# -r = registry directory 
$Verbose = 0;					# -v = print progress information

$ok = GetOptions(
		 "lexemes|l=s" => \$LexemeAtt,
		 "pos|p=s" => \$PosAtt,
		 "tag1|t1=s" => \$Tag1,
		 "tag2|t2=s" => \$Tag2,
		 "cleanup|c=s" => \$Clean,
		 "registry|r=s" => \$Registry,
		 "frequency|f=i" => \$Frequency,
		 "verbose|v" => \$Verbose,
		 "help|h" => \$Help,
		 );

die
  "\nUsage:  ucs-tool segment-from-cwb [options] <CORPUS> <segments>\n\n",
  "Extract segment-based cooccurrence data from a CWB-encoded corpus.\n",
  "The corpus must be registered in the IMS Corpus Workbench under the name\n",
  "<CORPUS> (all uppercase), and <segments> is the name of a s-attribute\n",
  "that encodes the segments to use.  The extraction results are written to\n",
  "STDOUT in the format required by the ucs-make-tables program.\n\n",
  "Options:\n",
  "  -l <att>   use p-attribute <att> for lexemes (default: lemma)\n",
  "     --lexemes=<att>\n",
  "  -p <att>   use p-attribute <att> for POS tags (default: pos)\n",
  "     --pos=<att>\n",
  "  -t1 <rx>   POS tags of first components must match CWB regexp <rx>\n",
  "     --tag1=<rx>\n",
  "  -t2 <rx>   POS tags of second components must match CWB regexp <rx>\n",
  "     --tag2=<rx>\n",
  "  -c <rx>    clean up data by filtering lexemes with Perl regexp <rx>\n",
  "     --cleanup=<rx>   (automatically anchored at start and end of string)\n",
  "  -r <dir>   set CWB registry directory (default: system setting)\n",
  "     --registry=<dir>\n",
  "  -f <n>     when a frequency threshold f >= <n> will later be imposed\n",
  "     --frequency=<n>  on the data set, this option can be used to reduce\n",
  "             the amount of data that has to be processed significantly\n",
  "  -v         verbose mode (show progress information)\n",
  "     --verbose\n",
  "  -h         show this help page\n",
  "     --help\n\n",
  unless $ok and (not $Help) and @ARGV == 2;

($Corpus, $SegmentAtt) = @ARGV;

$Clean = qr/^(?:$Clean)$/			# compile --cleanup regular expression 
  if $Clean;

## set up source corpus and attributes
$CL::Registry = $Registry			
  if defined $Registry;
$C = new CL::Corpus $Corpus
  or die "Can't open CWB corpus $Corpus - aborted\n";
$L = $C->attribute($LexemeAtt, "p")
  or die "Can't open p-attribute $Corpus.$L - aborted\n";
$S = $C->attribute($SegmentAtt, "s")
  or die "Can't open s-attribute $Corpus.$SegmentAtt - aborted\n";

if ($Tag1 ne ".*" or $Tag2 ne ".*") {
  $CheckTags = 1;
  $P = $C->attribute($PosAtt, "p")
    or die "Can't open p-attribute $Corpus.$PosAtt - aborted\n";
  %ID1 = map {$_ => 1} $P->regex2id($Tag1);	# lookup hash for lexicon IDs of valid POS tags (first components)
  %ID2 = map {$_ => 1} $P->regex2id($Tag2);	# lookup hash for lexicon IDs of valid POS tags (second components)
}
else {
  $CheckTags = 0;
}

## then process all segments and print records on STDOUT
$segments = $S->max_struc;
foreach $nr (0 .. $segments-1) {
  printf STDERR "Processing segment %7d/%d   \r", $nr+1, $segments
    if $Verbose;
  ($start, $end) = $S->struc2cpos($nr);
  %l1 = ();					# collect valid lexemes (without duplicates)
  %l2 = ();
  for $cpos ($start .. $end) {
    if ($CheckTags) {
      $id = $P->cpos2id($cpos);
      $valid1 = $ID1{$id};
      $valid2 = $ID2{$id};
      next unless $valid1 or $valid2;
    }
    else {
      $valid1 = $valid2 = 1;
    }

    $id = $L->cpos2id($cpos);
    $str = $L->id2str($id);

    next
      if $Clean and not $str =~ $Clean;

    if ($Frequency > 0) {
      $valid1 = $valid2 = 0
	unless $L->id2freq($id) >= $Frequency;
      # rationale behind the frequency threshold: the cooccurrence frequency of a pair type (u,v)
      # must be less than <n> if the token frequency (and hence the corresponding marginal frequency,
      # which is less than or equal to the token frequency) of either u or v is less than <n>
    }

    $l1{$str}++ 
      if $valid1;
    $l2{$str}++
      if $valid2;
  }
  print "$nr\n";				# segment ID = number of region
  print join("\t", keys %l1), "\n";		# labels of first components
  print join("\t", keys %l2), "\n";		# labels of second components
  print "\n";					# record separator
}
print STDERR "$segments record printed on STDOUT            \n"
  if $Verbose;



##
##  subroutines
##

sub broken_pipe {
  print STDERR "[output pipe broken, exiting prematurely]\n";
  exit 1;
}
