# -*-cperl-*-
## Filename: count-collocates.perl
## Modified: Sun May  8 20:12:04 2005 (evert)   
##   Author: Stefan Evert
##  Purpose: count number of collocates for each component type and write results to text files
$| = 1;

use UCS;
use UCS::File;
use UCS::DS::Stream;

use Getopt::Long;
use locale;

$Frequency = 0;					# -f = frequency threshold
$Threshold = 0;					# -t = threshold (number of collocates)
$Verbose = 0;					# -v = verbose mode
$Help = 0;					# -h = help page

$ok = GetOptions(
		 "frequency-threshold|f=i" => \$Frequency,
		 "threshold|t=i" => \$Threshold,
		 "verbose|v" => \$Verbose,
		 "help|h" => \$Help,
		 );

die 
  "\nUsage:  ucs-tool count-collocates [-f <n>] [-t <n>] <data.ds> <coll-l1.txt> <coll-l2.txt>\n\n",
  "Reads data set file <data.ds> and computes number of collocates for each\n",
  "component type (both l1 and l2).  Results for the first components are written\n",
  "to the file <coll-l1.txt> (not in data set format), results for the second\n",
  "components to <coll-l2.txt>.  Each line of these files has the format\n",
  "  <component-type> TAB <number-of-collocates> TAB <examples>\n",
  "where <examples> lists up to 10 of the most frequent collocates.  The types\n",
  "with the largest number of collocates are shown first.\n\n",
  "Options:\n",
  "  -f <n>   only count collocates with cooccurrence frequency >= <n>\n",
  "     --frequency-threshold=<n>\n",
  "  -t <n>   only report component types with at least <n> collocates\n",
  "     --threshold=<n>\n",
  "  -v       verbose mode (show progress)\n",
  "     --verbose\n",
  "\n"
  unless $ok and (not $Help) and @ARGV == 3;

($dsfile, $l1file, $l2file) = @ARGV;

$ds = new UCS::DS::Stream::Read $dsfile;
$size = $ds->global("size");
$size = "??????" unless defined $size;

%L1 = ();					# collocates of l1 components
%count1 = ();					# collocate counts
%L2 = ();					# collocates of l2 components
%count2 = ();					# collocate counts

while ($ds->read) {
  printf STDERR "Processing row %6d / %s   \r", $ds->row, $size
    if $Verbose and ($ds->row & 0xff) == 0;
  $f = $ds->value("f");
  next 						# skip pair types below frequency threshold (with -f option)
    if $f < $Frequency;
  $l1 = $ds->value("l1");
  $l2 = $ds->value("l2");
  $count1{$l1}++
    unless exists $L1{$l1}{$l2};
  $L1{$l1}{$l2} += $f;
  $count2{$l2}++
    unless exists $L2{$l2}{$l1};
  $L2{$l2}{$l1} += $f;
}
$size = $ds->row;
$ds->close;

print STDERR "Processing complete ($size rows).               \n"
  if $Verbose;

write_report($l1file, \%L1, \%count1);
write_report($l2file, \%L2, \%count2);


# write_report($filename, \%L1, \%count1);
sub write_report {
  my ($filename, $L, $count) = @_;
  my $fh = UCS::File::Open "> $filename";
  print STDERR "Creating report $filename ... [sorting] "
    if $Verbose;
  my @types = 
    sort { $count->{$b} <=> $count->{$a} or $a cmp $b } 
      grep { $count->{$_} > $Threshold }	# always ok without -t option
	keys %$count;
  print STDERR "[saving] "
    if $Verbose;
  foreach my $t (@types) {
    print $fh "$t\t", $count->{$t};
    my $data = $L->{$t};
    my @collocates = sort { $data->{$b} <=> $data->{$a} or $a cmp $b } keys %$data;
    $#collocates = 9				# truncate to 10 most frequent collocates
      if @collocates > 10;
    foreach my $c (@collocates) {
      print $fh "\t$c (", $data->{$c}, ")";
    }
    print $fh "\n";
  }
  $fh->close;
  print STDERR "ok\n"
    if $Verbose;
}
