# -*-cperl-*-
## Filename: dispersion-test.perl
## Modified: Sun May  8 20:12:34 2005 (evert)   
##   Author: Stefan Evert
##  Purpose: perform dispersion test on data set with dispersion counts
$| = 1;

use UCS;
use UCS::DS::Stream;
use UCS::SFunc;

use Math::BigInt lib => "GMP";			# need high precision math for distribution of dispersion
use Math::BigFloat;

use Getopt::Long;
use locale;

$Opt_M = 0;					# -m = (maximal) frequency rank
$Opt_N = 0;					# -N = sample size
$Opt_K = 0;					# -k = number of chunks
$Opt_S = 0;					# -S = chunk size (alternative for -k)
$Opt_V = 0;					# -V = number of types in frequency class m (for prob. table)
$Opt_alpha = 0;					# -a = significance level (alpha)
$Verbose = 0;					# -v = verbose mode
$Help = 0;					# -h = help page

@significance = qw<.05 .01 .001>;		# show critical values for these significance levels

Getopt::Long::Configure("no_ignore_case");
$ok = GetOptions(
		 "m=i" => \$Opt_M,
		 "N=i" => \$Opt_N,
		 "k=i" => \$Opt_K,
		 "S=i" => \$Opt_S,
		 "V=i" => \$Opt_V,
		 "alpha|a=f" => \$Opt_alpha,
		 "verbose|v" => \$Verbose,
		 "help|h" => \$Help,
		 );

if (@ARGV == 0) {
  $ok = 0
    unless $Opt_M > 1 and $Opt_N > 0 and ($Opt_K > 0 or $Opt_S > 0) and not $Opt_alpha;
  $dsfile = undef;
}
else {
  $dsfile = shift @ARGV;
  $ok = 0
    unless @ARGV == 0 and not $Opt_V;
  $Opt_M = 10 					# default in mode 2: -m 10
    unless $Opt_M;
}

die 
  "\nUsage:  ucs-tool dispersion-test [-V <types>] -m <rank> -N <size> -k <chunks>\n\n",
  "Prints table of probabilities P(d|m) and P(<=d|m) from dispersion test, for\n",
  "sample size N=<size>, k=<chunks> chunks, and frequency class m=<rank>.  If\n",
  "the number of types in frequency class m is specified with the -V option, the\n",
  "expected number of underdispersed types and some critical values are also shown.\n",
  "\n",
  "Usage:  ucs-tool dispersion-test [-v] [-a <alpha>] [-m <max.m>] [-N <size>] [-k <chunks>] <data.ds>\n\n",
  "Perform a dispersion test on the data set <data.ds> (which must contain dispersion\n",
  "counts from the MakeTables utility).  The sample size and number of chunks are read\n",
  "from the data set file and can be overridden with the -N and -k options.  If the -m\n",
  "option is specified, dispersion tests will only be carried out for the frequency\n",
  "up to <m.max> (default: 10).  The -v option prints some progress information.\n",
  "With the -a (or --alpha) option, only results that are significant at the specified\n",
  "level will be printed.\n",
  "\n",
  unless $ok and (not $Help);

# open data set file to initialise k and S parameters (in mode 2)
if ($dsfile) {
  $ds = new UCS::DS::Stream::Read $dsfile;
  die "Error:  no dispersion counts (n.disp) in data set $dsfile\n"
    unless $ds->var("n.disp");
  $k = $ds->global("chunks");
  if ($k) {
    $Opt_K = $k
      unless $Opt_K;
    print STDERR "Warning:  -k $Opt_K overrides data set value k=$k\n"
      if $Opt_K != $k;
  }
  $S = $ds->global("chunk.size");
  if ($S) {
    $Opt_S = $S
      unless $Opt_S;
    print STDERR "Warning:  -S $Opt_S overrides data set value S=$S\n"
      if $Opt_S != $S;
  }
  $ds->read;					# preload the first line to determine N
  $N = $ds->value("N");
  $Opt_N = $N
    unless $Opt_N;
  print STDERR "Warning:  -N $Opt_N overrides data set value N=$N\n"
    if $Opt_N != $N;
}

# auto-compute k from S or vice versa (either of the two must be non-zero)
if ($Opt_K == 0) {
  $Opt_K = int(($Opt_N - 1) / $Opt_S) + 1;	# emulates ceil() function
}
elsif ($Opt_S == 0) {
  $Opt_S = int(($Opt_N - 1) / $Opt_K) + 1;
}
print STDERR "Warning: k=$Opt_K * S=$Opt_S does not match sample size N=$Opt_N!\n"
  unless $Opt_N == $Opt_K * $Opt_S;

if (not $dsfile) {
  ##
  ## mode 1: print probability table for dispersion test
  ##
  print "\n";
  print "Probability Table for Dispersion Test\n";
  print "-------------------------------------\n";
  print "\n";
  print "Sample size:      N = $Opt_N\n";
  print "Number of chunks: k = $Opt_K\n";
  print "Chunk size:       S = $Opt_S\n";
  print "Frequency class:  m = $Opt_M\n";
  print "Types in class:   V_m = $Opt_V  (for critical values)\n"
    if $Opt_V;
  print "\n";
  $m = $Opt_M;
  $f_int = '%'.length("$m").'d';		# number of digits for d and m
  $f_V = '%'.length("$Opt_V").'d';		# number of digits for V_m
  foreach $d (1 .. $m) {
    $p = ddisp($d, $m);
    $f_float = ($p < 1e-4) ? '%e' : '%12.10f';	# appropriate 12-character format for p
    printf "P(d = $f_int  | m = $f_int) = $f_float\n", $d, $m, $p;
    $p = pdisp($d, $m);
    $f_float = ($p < 1e-4) ? '%e' : '%12.10f';	# appropriate 12-character format for p
    printf "P(d <= $f_int | m = $f_int) = $f_float\n", $d, $m, $p;
    if ($Opt_V) {
      printf "Expected number of underdispersed types: $f_V out of $f_V\n", int($p * $Opt_V + 0.5), $Opt_V;
      print "Critical values: ";
      foreach $alpha (@significance) {
	$n_critical = qbinom($alpha, $Opt_V, $p, 'upper') + 1;
	if ($p >= 1 or $n_critical > $Opt_V) {
	  print "  NA ($alpha) ";		# test is never rejected
	}
	else {
	  printf "$f_V (%s) ", $n_critical, $alpha;
	}
      }
      print "\n";
    }
    print "\n";
  }
}
else {
  ##
  ## mode 2: apply dispersion test to data set file
  ##
  %Vm = ();					# $Vm{$m} = number of types in frequency class $m
  %VmD = ();					# $VmD{$m}{$d} = number of types with dispersion $d in class $m

  # recall that data set stream has already been opened (as $ds) and first row has been read
  $size = $ds->global("size");
  $size = "??????" unless defined $size;
  while (not $ds->eof) {
    printf STDERR "Processing row %6d / %s   \r", $ds->row, $size
      if $Verbose and ($ds->row & 0xff) == 0;
    $m = $ds->value("f");
    if ($m > 1 and not ($Opt_M and $m > $Opt_M)) { # ignore hapaxes and frequency ranks above -m threshold
      $d = $ds->value("n.disp");
      $Vm{$m}++;
      $VmD{$m}{$d}++;
    }
    $ds->read;
  }
  $size = $ds->row;
  $ds->close;

  print STDERR "Processing complete ($size rows).               \n"
    if $Verbose;

  print "\n";
  $header = "Dispersion Test for data set $dsfile";
  print $header, "\n";
  print "-" x length($header), "\n";
  print "\n";
  print "Sample size:      N = $Opt_N\n";
  print "Number of chunks: k = $Opt_K\n";
  print "Chunk size:       S = $Opt_S\n";
  print "\n";
  @ranks = sort {$a <=> $b} keys %Vm;
  foreach $m (@ranks) {
    $V = $Vm{$m};
    $f_int = '%'.length("$m").'d';		# number of digits for d and m
    $f_V = '%'.length("$V").'d';		# number of digits for number of underdispersed types
    printf "    m = $f_int, Vm = $f_V\n", $m, $V;
    print "\n";
    @d_vals = sort {$a <=> $b} grep {$_ < $m} keys %{$VmD{$m}};
    $Vd = 0;					# number of types with dispersion <= $d
    foreach $d (@d_vals) {
      $Vd += $VmD{$m}{$d};
      $prob = pdisp($d, $m);			# probability P(<=d|m) for individual pair type
      $p_value = pbinom($Vd - 1, $V, $prob, 'upper'); # p-value for observed number of underdispersed types
      if (not $Opt_alpha or $p_value < $Opt_alpha) {
	printf "d = $f_int: $f_V expected, $f_V observed", $d, int($prob * $V + 0.5), $Vd;
	$sigcode = "";				# compute significance code: * (.05), ** (.01), *** (.001)
	grep { $sigcode .= "*" if $p_value < $_ } @significance;
	print "  $sigcode\n";
	print "[p-value = $p_value]\n\n";
      }
    }
  }
}




exit 1;




##
## subroutines: compute density and probabilities for dispersion test
##

# probability density P(d|m) for sample size N and k chunks of size S
# $p = ddisp($d, $m);  [implicit: $Opt_N, $Opt_K, $Opt_S]
BEGIN { %cacheDDISP = (); }			# memoize probability density
sub ddisp {
  my ($d, $m) = @_;
  my ($N, $K, $S) = ($Opt_N, $Opt_K, $Opt_S);
  return 0 if $d <= 0 or $d > $m or $d > $K;
  my $key = "$d\t$m";
  return $cacheDDISP{$key}
    if exists $cacheDDISP{$key};
  my $result = new Math::BigFloat binom($K, $d) * A($d, $m); # convert numerator to BigFloat ...
  $result /= binom($N, $m);			# ... so we can divide the denominator into it
  $result = 0.0 + $result->bstr;		# convert back to ordinary float (need to go through string rep.)
  $cacheDDISP{$key} = $result;
  return $result;
}

# probability distribution P(<=d|m) for sample size N and k chunks of size S
# $p = pdisp($d, $m);  [implicit: $Opt_N, $Opt_K, $Opt_S]
BEGIN { %cachePDISP = (); }			# memoize distribution function
sub pdisp {
  my ($d, $m) = @_;
  my ($N, $K, $S) = ($Opt_N, $Opt_K, $Opt_S);
  return 0 if $d <= 0;
  return 1 if $d > $m or $d > $K;
  my $key = "$d\t$m";
  return $cachePDISP{$key}
    if exists $cachePDISP{$key};
  my $result = 0;
  for my $i (1 .. $d) {
    $result += ddisp($i, $m);
  }
  $cachePDISP{$key} = $result;
  return $result;
}

# recursive function A(d; m, S) used to compute probabilities P(d|m), returns BigInt
# $val = A($d, $m);    [implicit: $Opt_K, $Opt_S]
BEGIN { %cacheA = (); }				# memoize function values
sub A {
  my ($d, $m) = @_;
  my ($S, $K) = ($Opt_S, $Opt_K);
  return Math::BigInt->bzero() if $d <= 0 or $d > $m or $d > $K;
  my $key = "$d\t$m";
  return $cacheA{$key}
    if exists $cacheA{$key};
  my $result = binom($d * $S, $m);
  foreach my $j (1 .. $d-1) {
    $result -= binom($d, $j) * A($j, $m);
  }
  $cacheA{$key} = $result;
  return $result;
}

# naive multiplicative algorithm for binomial coefficient "n choose k"; returns BigInt
# $coef = binom($n, $k);
sub binom {
  my ($n, $k) = @_;
  return Math::BigInt->bzero() if $k < 0 or $k > $n;
  return Math::BigInt->bone() if $k == 0 or $k == $n;
  my $result = Math::BigInt->bone();
  foreach my $i (1 .. $k) {
    $result *= ($n - $i + 1);
    $result /= $i;	# intermediate result is integer, viz. binom($n, $i)
  }
  return $result;
}

