# -*-cperl-*-
## Filename: print-documentation.perl
## Modified: Fri Aug 26 02:41:06 2005 (severt)   
##   Author: Stefan Evert
##  Purpose: print all UCS/Perl PODs in PostScript format
$| = 1;

use UCS;
use UCS::File;

use File::Find;
use Getopt::Long;
use Pod::LaTeX;

use sigtrap qw<die normal-signals>;

# sort order for man pages (specify filenames with (partial) paths)
@List = split /\n/, <<'EOT';
== General Documentation
pod/ucsintro.pod
pod/ucsfile.pod
pod/ucsexp.pod
pod/ucsam.pod
== UCS/Perl Programs
bin/ucsdoc
bin/ucs-config
bin/ucs-tool
bin/ucs-list-am
bin/ucs-make-tables
bin/ucs-summarize
bin/ucs-select
bin/ucs-add
bin/ucs-join
bin/ucs-sort
bin/ucs-info
bin/ucs-print
== UCS/Perl Modules
lib/UCS.pm
lib/UCS/File.pm
lib/UCS/R.pm
lib/UCS/R/Expect.pm
lib/UCS/R/RSPerl.pm
lib/UCS/SFunc.pm
lib/UCS/Expression.pm
lib/UCS/Expression/Func.pm
lib/UCS/AM.pm
lib/UCS/AM/HTest.pm
lib/UCS/AM/Parametric.pm
lib/UCS/DS.pm
lib/UCS/DS/Stream.pm
lib/UCS/DS/Memory.pm
lib/UCS/DS/Format.pm
EOT

@SecStart = ();		             # indicates where sections begin (with name)
$N_add = 1;
%N = map { 
  if (s/^==+\s*//) {
    $SecStart[$N_add] = $_;	     # mark beginning of this section
    ();				     # skip entry
  } else {
    $_ => $N_add++; 	             # insert into hash
  }
} @List;
@Unlisted = ();

$Verbose = 0;			     # -v  show filenames of unlisted manpages
$Collate = 0;			     # -c  combine all pages into single file
$Latex = 0;			     # -l  convert into LaTeX format 
$Narrow = undef;		     # -n  assume narrow pages when resizing verbatim text
$Small = 0;			     # -s  use default font size "small" for verbatim text
$Help = 0;			     # -h  help message

$ok = GetOptions("v|verbose" => \$Verbose,
		 "c|collate" => \$Collate,
		 "l|latex" => \$Latex,
		 "n|narrow:i" => \$Narrow,
		 "s|small" => \$Small,
		 "h|help" => $Help,
		);

die "\nUsage:  ucs-tool print-documentation [-v] [-c] [-l [-n] [-s]] <prefix>\n\n",
  "Converts the manual pages of all UCS/Perl modules and programs\n",
  "into PostScript format for printing.  The manual pages will be\n",
  "written into a series of files named '<prefix>-NNN.ps', ordered\n",
  "as specified by an internal list.  Unlisted pages are appended\n",
  "at the end (use -v or --verbose to show their filenames).  With\n",
  "the -c or --collate option, all pages are combined into a single\n",
  "PostScript file.\n\n",
  "The -l or --latex option translates the manual pages into a LaTeX\n",
  "document. It is usually paired with --collate, which will add a\n",
  "and footer around the combined document and resize verbatim blocks\n",
  "automatically when they exceed a width of 75 characters.  When the\n",
  "document is to be included in a document with smaller page width,\n",
  "use --narrow or -n to reduce this limit by 10 characters.  It is\n",
  "also possible to specify the desired reduction after the -n flag.\n",
  "When the --small or -s option is specified, the default font size\n",
  "for verbatim blocks is set to \\small (but the calculation of font\n",
  "sizes is not changed).\n\n",
  unless @ARGV == 1 and $ok; 
$Prefix = shift @ARGV;

if (defined $Narrow) {				# reduce assumed page width by $Narrow characters
  $Narrow = 10 if $Narrow == 0;
}
else {
  $Narrow = 0;
}

$BinDir = $UCS::PerlDir."/bin";
$LibDir = $UCS::PerlDir."/lib";

## print general (introductory) PODs
print "Printing general documents: ";
@pods = glob "$LibDir/pod/*.pod";
foreach $pod (@pods) { print_pod($pod) }
print "\n";

## print documentation for UCS programs
print "Printing docs of programs:  ";
@progs = grep {not /~$/} glob "$BinDir/*";
@progs = grep {not /test/i} @progs;		# skip test programs without documentation
foreach $prog (@progs) { print_pod($prog) }
print "\n";

## finally, scan UCS/Perl lib dir for Perl modules and print their PODs
print "Printing docs of modules:   ";
find({
      wanted => sub { print_pod($File::Find::name) if /\.pm$/ }, 
      no_chdir => 1,
     },
     $LibDir);
print "\n";

## with --collate, manpages are now in nroff format -> use groff to create single PostScript file
if ($Collate) {
  print "Collating ... ";
  if ($Latex) {					# LATEX mode
    $fh = UCS::File::Open "> $Prefix.tex";
    print $fh <<'EOT';
\documentclass[a4paper]{article}

\usepackage{vmargin}
\setpapersize{A4}
\setmarginsrb{30mm}{20mm}{30mm}{20mm}{0mm}{0mm}{10pt}{20mm}

\usepackage{makeidx}
\makeindex

\begin{document}

\title{UCS/Perl Documentation}
\author{Stefan Evert}
\date{printed on \today}
\maketitle

\tableofcontents{}
\newpage
\sloppy
EOT
    foreach $i (1 .. $N_add - 1) {
      $texfile = sprintf "%s-%03d.tex", $Prefix, $i;
      if ($SecStart[$i]) {
	print $fh "\n\\section{$SecStart[$i]}\n\n";
      }
      $ifh = UCS::File::Open $texfile;
      transform_append($ifh, $fh);
      print $fh "\\newpage\n";
      $ifh->close;
      unlink($texfile);
    }
    print $fh <<'EOT';
\printindex
\end{document}
EOT
    $fh->close;
  }
  else {					# POSTSCRIPT mode
    @manfiles = map { sprintf "%s-%03d.man", $Prefix, $_ } 1 .. $N_add-1;
    UCS::File::ShellCmd(["groff", "-man", "-Tps", @manfiles], "$Prefix.ps");
    unlink($_) for @manfiles;
  }
 print "done\n";
}

## in --verbose mode, show filenames of unlisted manpages
if ($Verbose) {
  if (@Unlisted) {
    print "\nFilenames of unlisted manpages:\n";
    print "\t$_\n" for @Unlisted;
  }
  else {
    print "\nNo unlisted manpages. Good.\n";
  }
}

##
##  That's it.
##

# print one POD document (output filename will be chosen automatically)
sub print_pod {
  my $name = shift;

  my @keys = grep { $name =~ /\Q$_\E$/ } keys %N; # check whether file is listed
  die "Internal error: file $name matches multiple list entries (@keys)\n"
    if @keys > 1;
  my $Number = 0;
  if (@keys) { 
    $Number = $N{$keys[0]};
    die "Internal error: list entry @keys matches multiple files\n"
      unless $Number > 0;
    $N{$keys[0]} = 0;
  }
  else {
    $Number = $N_add++;
    push @Unlisted, $name;
  }
  
  my $outname = sprintf "%s-%03d", $Prefix, $Number;
  if ($Latex) {
    $Parser = new Pod::LaTeX;			# it seems necessary to re-initialise the parser for every translation
    if ($Collate) {
      $Parser->Head1Level(2);			# each manpage becomes a (numbered) subsection;
      $Parser->ReplaceNAMEwithSection(1);       # further parts of manpage become subsubsections etc.
      $Parser->LevelNoNum(3);			# but they will not be numbered
      $Parser->AddPreamble(0);			# to allow collation
      $Parser->AddPostamble(0);
    }
    ##  $Parser->StartWithNewPage(1);             # make sure each POD starts on a new page
    $Parser->parse_from_file($name, "$outname.tex");
  }
  elsif ($Collate) {
    UCS::File::ShellCmd("$BinDir/ucsdoc -onroff $name > $outname.man");
  }
  else {
    UCS::File::ShellCmd("$BinDir/ucsdoc -ps $name > $outname.ps");
  }
  print "[$Number] ";
}

# append latex file $in to file $out, cleaning up the latex code in the process
sub transform_append {
  my ($in, $out) = @_;
  while (<$in>) {
    if (/^\s*\\begin\{verbatim\}\s*$/) {	# dynamically adjust size of verbatim blocks
      my @verbatim = ();			# collect verbatim text
      my $width = 0;
      while (<$in>) {
	if (/^\s*\\end\{verbatim\}\s*$/) {
	  $_ = <$in>;
	  last unless /^\s*\\begin\{verbatim\}\s*$/; # combine adjacent verbatim blocks
	  push @verbatim, "";			# (separated by blank line)
	}
	else {
	  chomp;
	  s/\s+$//;
	  push @verbatim, $_;
	  $width = length($_)
	    if length($_) >= $width;
	}
      }
      my $next_line = $_;			# first line after verbatim block
      my $size = ($Small) ? "small" : "normalsize";
      $size = "small" if $width > 75 - $Narrow;
      $size = "footnotesize" if $width > 90 - $Narrow;
      $size = "scriptsize" if $width > 105 - $Narrow;
      $size = "tiny" if $width > 120 - $Narrow;
      print $out "\\begin{$size}\n\\begin{verbatim}\n";
      print $out "$_\n" for @verbatim;
      print $out "\\end{verbatim}\n\\end{$size}\n";
      $_ = $next_line;				# fall through to normal processing (can't be \begin{verbatim}!)
    }

    s/\$>\$/\\textgreater{}/g;			# use latex2e special text characters <, >, \, ~, and ^
    s/\$<\$/\\textless{}/g;
    s/\$\\backslash\$/\\textbackslash{}/g;
    s/\\texttt\{\\~\{\}\}/\\textasciitilde{}/g;
    s/\\\^\{\}/\\textasciicircum{}/g;
    s/(\\item)(\s+)/$1\{\}$2/g;			# so that \item [remark] ... works
    print $out $_;
  }
}
